/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.dtypes;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;

import javax.swing.JFrame;
import javax.swing.WindowConstants;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import org.apache.commons.math3.exception.MathParseException;
import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.DecompositionSolver;
import org.apache.commons.math3.linear.LUDecomposition;
import org.apache.commons.math3.linear.RealMatrix;

import at.lcm.bu21.general.dtypes.preview.TPeriodicPreview;
import at.lcm.bu21.general.plot.Plot;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Representation of a array-type
 * 
 * @author Gnther Weidenholzer
 */
public class TPeriodicMatrix extends TData {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Data variable */
	protected ArrayList<ArrayList<TPeriodic>> data = new ArrayList<ArrayList<TPeriodic>>();

	/**
	 * Create empty object
	 */
	public TPeriodicMatrix() {

	}

	/**
	 * Create object with given size
	 * 
	 * @param size
	 *            Size of Array
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TPeriodicMatrix(int[] size) throws TDataException {
		if (size.length == 2) {
			this.expandArray(size[0], size[1]);
		} else {
			throw new TDataException("Only 2 dimensions allowed!");
		}
	}

	/**
	 * Create object with given size
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of column
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TPeriodicMatrix(int row, int col) throws TDataException {
		this(new int[] { row, col });
	}

	/**
	 * Create object with given data (ArrayList)
	 * 
	 * @param row
	 *            Row of data to insert (ArrayList)
	 */
	public TPeriodicMatrix(ArrayList<TPeriodic> row) {
		this.data = new ArrayList<ArrayList<TPeriodic>>();
		this.data.add(row);
	}

	/**
	 * Create object with array of dc-vaules
	 * 
	 * @param array_dc
	 *            Array of dc-values
	 * @throws Exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TPeriodicMatrix(double[][] array_dc) throws Exception {

		int rows = array_dc.length;
		int cols = array_dc[0].length;

		TPeriodic zero = new TPeriodic();

		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				this.set(r, c, zero.offset(array_dc[r][c]));
			}
		}

	}

	/**
	 * Create object with given data (TArray of TNumeric(1x1) and/or TPeriodic)
	 * 
	 * @param array
	 *            Array of data to insert (TArray of TNumeric(1x1) and/or
	 *            TPeriodic)
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TPeriodicMatrix(TArray array) throws TDataException, CloneException {
		this(array.size());

		int[] sz = array.size();

		for (int row = 0; row < sz[0]; row++) {
			for (int col = 0; col < sz[1]; col++) {
				TData d = array.get(row, col);
				if (d instanceof TNumeric) {
					if (((TNumeric) d).isScalar()) {
						this.set(row, col, new TPeriodic().offset(((TNumeric) d).getReal(0, 0)));
					} else {
						throw new TDataException("TNumeric item must be scalar!");
					}
				} else if (d instanceof TPeriodic) {
					this.set(row, col, (TPeriodic) d);
				} else {
					throw new TDataException("Array must not contain other types than TPeriodic, TNumeric(1x1)!");
				}
			}
		}
	}

	/**
	 * Returns TPeriodicMatrix filled with random signals for testing purposes
	 * 
	 * @param rows
	 *            Number of rows
	 * @param cols
	 *            Number of columns
	 * @return TPeriodicMatrix
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public static TPeriodicMatrix getRandom(int rows, int cols) throws TDataException, CloneException {

		TPeriodicMatrix data = new TPeriodicMatrix(rows, cols);

		for (int row = 0; row < rows; row++) {
			for (int col = 0; col < cols; col++) {
				data.set(row, col, TPeriodic.getRandom());
			}
		}

		return data;

	}

	/**
	 * 3-phase Park Transformation
	 * 
	 * @return TPeriodicMatrix
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public static TPeriodicMatrix parkTransform() throws TDataException, CloneException {

		return TPeriodicMatrix.parkTransform(3);
	}

	/**
	 * Park Transformation
	 * 
	 * @param m
	 *            Number of phases
	 * 
	 * @return TPeriodicMatrix
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public static TPeriodicMatrix parkTransform(int m) throws TDataException, CloneException {
		TPeriodicMatrix p = new TPeriodicMatrix(3, m);

		for (int k = 0; k < m; k++) {

			p.set(0, k, new TPeriodic(1, 2. / m, (-2.0 * k * Math.PI) / m));
			p.set(1, k, new TPeriodic(1, 2. / m, (Math.PI / 2.0) - ((2.0 * k * Math.PI) / m)));
			p.set(2, k, new TPeriodic().offset(1.0 / m));

		}

		return (p);

	}

	/**
	 * inverse Park Transformation
	 * 
	 * @return TPeriodicMatrix
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public static TPeriodicMatrix inverseParkTransform() throws CloneException, TDataException {

		return TPeriodicMatrix.inverseParkTransform(3);
	}

	/**
	 * inverse Park Transformation
	 * 
	 * @param m
	 *            Number of phases
	 * 
	 * @return TPeriodicMatrix
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public static TPeriodicMatrix inverseParkTransform(int m) throws CloneException, TDataException {
		TPeriodicMatrix p = new TPeriodicMatrix(m, 3);

		for (int k = 0; k < m; k++) {

			p.set(k, 0, new TPeriodic(1, 1., (-2.0 * k * Math.PI) / m));
			p.set(k, 1, new TPeriodic(1, 1., (Math.PI / 2.) - ((2.0 * k * Math.PI) / m)));
			p.set(k, 2, new TPeriodic().offset(1.0));

		}

		return (p);

	}

	/**
	 * Creates Identity Matrix
	 * 
	 * @param dimension
	 *            Dimension
	 * @return IdentityMatrix
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public static TPeriodicMatrix identityMatrix(int dimension) throws CloneException, TDataException {
		TPeriodicMatrix m1 = new TPeriodicMatrix(dimension, dimension);

		TPeriodic identity = new TPeriodic().offset(1.0);
		for (int k = 0; k < dimension; k++) {
			m1.set(k, k, identity);
		}
		return m1;
	}

	/**
	 * Elementwise operation without arguments
	 * 
	 * @param op
	 *            TPeriodic method name
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix elemOp(String op) throws TDataException, CloneException, Exception {

		int[] size = this.size();
		TPeriodicMatrix data = new TPeriodicMatrix(size[0], size[1]);

		Method m = new TPeriodic().getClass().getMethod(op, new Class[0]);

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic thisobj = this.get(row, col);
				TPeriodic result = (TPeriodic) m.invoke(thisobj);
				if (result != null) {
					data.set(row, col, result);
				} else {
					data.set(row, col, thisobj);
				}

			}
		}

		return data;
	}

	/**
	 * Elementwise operation with double argument
	 * 
	 * @param op
	 *            TPeriodic method name
	 * @param val
	 *            Operand
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix elemOp(String op, double val) throws TDataException, CloneException, Exception {

		int[] size = this.size();
		TPeriodicMatrix data = new TPeriodicMatrix(size[0], size[1]);

		Method m = new TPeriodic().getClass().getMethod(op, new Class[] { double.class });

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic thisobj = this.get(row, col);
				TPeriodic result = (TPeriodic) m.invoke(thisobj, val);
				data.set(row, col, result);
			}
		}

		return data;
	}

	/**
	 * Elementwise operation without arguments, TNumeric result
	 * 
	 * @param op
	 *            TPeriodic method name
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws Exception
	 *             Exception by operation
	 */
	public TNumeric elemOpNum(String op) throws TDataException, Exception {

		int[] size = this.size();
		double[][] data = new double[size[0]][size[1]];

		Method m = new TPeriodic().getClass().getMethod(op, new Class[0]);

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic thisobj = this.get(row, col);
				data[row][col] = (Double) m.invoke(thisobj);
			}
		}

		return new TNumeric(data);
	}

	/**
	 * Elementwise operation with int[] argument
	 * 
	 * @param op
	 *            TPeriodic method name
	 * @param val
	 *            Operand
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix elemOp(String op, int[] val) throws TDataException, CloneException, Exception {

		int[] size = this.size();
		TPeriodicMatrix data = new TPeriodicMatrix(size[0], size[1]);

		Method m = new TPeriodic().getClass().getMethod(op, new Class[] { int[].class });

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic thisobj = this.get(row, col);
				TPeriodic result = (TPeriodic) m.invoke(thisobj, val);
				data.set(row, col, result);
			}
		}

		return data;
	}

	/**
	 * Elementwise operation with double argument
	 * 
	 * @param op
	 *            TPeriodic method name
	 * @param arg
	 *            Operand
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix elemOp(String op, TPeriodicMatrix arg) throws TDataException, CloneException, Exception {

		int[] size = this.size();
		TPeriodicMatrix data = new TPeriodicMatrix(size[0], size[1]);

		Method m = new TPeriodic().getClass().getMethod(op, new Class[] { TPeriodic.class });

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {

				TPeriodic thisobj = this.get(row, col);
				TPeriodic val = arg.get(row, col);
				TPeriodic result = (TPeriodic) m.invoke(thisobj, val);
				data.set(row, col, result);
			}
		}

		return data;

	}

	/**
	 * Matrix addition, source matrix remains unchanged
	 * 
	 * @param other
	 *            Second matrix
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix add(TPeriodicMatrix other) throws TDataException, CloneException, Exception {

		TPeriodicMatrix C;

		C = this.elemOp("add", other);
		return C;
	}

	/**
	 * Matrix subtraction, source matrix remains unchanged
	 * 
	 * @param other
	 *            Second matrix
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix subtract(TPeriodicMatrix other) throws TDataException, CloneException, Exception {

		TPeriodicMatrix C;

		C = this.elemOp("subtract", other);
		return C;
	}

	/**
	 * Matrix multiplication, source matrix remains unchanged!
	 * 
	 * @param other
	 *            Second matrix
	 * @return Result
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix multiply(TPeriodicMatrix other) throws TDataException, CloneException {

		// result (m x n) = this (m x s) * other (s x n)
		int[] size1 = this.size();
		int[] size2 = other.size();

		int m = size1[0];
		int n = size2[1];
		int s = size1[1];

		if (s != size2[0]) {
			throw new TDataException("Inner matrix dimension must agree");
		}

		TPeriodicMatrix result = new TPeriodicMatrix(m, n);

		for (int i = 0; i < m; i++) {
			for (int j = 0; j < n; j++) {

				TPeriodic sum = new TPeriodic();

				for (int k = 0; k < s; k++) {
					sum = sum.add(this.get(i, k).multiply(other.get(k, j)));
				}
				sum.compress();
				result.set(i, j, sum);
			}
		}

		return result;

	}

	/**
	 * Pointwise matrix inversion
	 * 
	 * If matrix is badly conditioned at more than 5% of the period, an
	 * Exception is thrown
	 * 
	 * @return Result
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix inverse() throws TDataException, CloneException {
		int[] size = this.size();

		if (size[0] != size[1]) {
			throw new TDataException("Matrix must be square!");
		}

		int n = 128;

		RealMatrix[] data = new RealMatrix[n];

		for (int k = 0; k < n; k++) {
			double tmp = (2.0 * Math.PI * k) / n;

			RealMatrix m = this._valueAt(tmp);

			DecompositionSolver luSolver = new LUDecomposition(m).getSolver();

			if (luSolver.isNonSingular()) {
				RealMatrix inv = luSolver.getInverse();
				data[k] = inv;
			}
		}

		TPeriodicMatrix result = new TPeriodicMatrix(size[0], size[1]);
		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {

				double[] value = new double[n];

				for (int k = 0; k < n; k++) {
					value[k] = data[k].getEntry(row, col);
				}

				TPeriodic tp = new TPeriodic();
				tp.setFFT(value);
				result.set(row, col, tp);
			}
		}
		return result;
	}

	/**
	 * Transpose matrix
	 * 
	 * @return transposed matrix
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix transpose() throws CloneException, TDataException {

		int[] size = this.size();
		TPeriodicMatrix result = new TPeriodicMatrix();

		for (int row = 0; row < size[1]; row++) {
			for (int col = 0; col < size[0]; col++) {
				result.set(row, col, this.get(col, row));
			}
		}
		return result;
	}

	/**
	 * Returns slice matrix at given angle
	 * 
	 * @param angle
	 *            Angle
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric valueAt(double angle) throws TDataException {

		return new TNumeric(this._valueAt(angle));
	}

	/**
	 * Returns slice matrix at given angle
	 * 
	 * @param angle
	 * @return Result
	 * @throws MathParseException
	 *             Mathematics exception
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	private BlockRealMatrix _valueAt(double angle) throws MathParseException, TDataException {

		int[] size = this.size();

		BlockRealMatrix matrix = new BlockRealMatrix(size[0], size[1]);

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				matrix.setEntry(row, col, this.get(row, col).valueAt(angle));
			}
		}
		return matrix;
	}

	/**
	 * Multiplication of the TPeriodicMatrix with the given TNumeric. source
	 * matrix remains unchanged!
	 * 
	 * @param other
	 *            Numeric matrix
	 * @return Result
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix multiply(RealMatrix other) throws TDataException, CloneException {

		// result (m x n) = this (m x s) * other (s x n)
		int[] sizeThis = this.size();
		int[] sizeOther = new int[] { other.getRowDimension(), other.getColumnDimension() };

		if (sizeThis[1] != sizeOther[0]) {
			throw new TDataException("Inner matrix dimension must agree");
		}

		TPeriodicMatrix result = new TPeriodicMatrix(sizeThis[0], sizeOther[1]);

		for (int i = 0; i < sizeThis[0]; i++) {
			for (int j = 0; j < sizeOther[1]; j++) {

				TPeriodic sum = new TPeriodic();

				for (int k = 0; k < sizeThis[1]; k++) {
					sum = sum.add(this.get(i, k).scale(other.getEntry(k, j)));
				}
				result.set(i, j, sum);
			}
		}
		return result;
	}

	/**
	 * Matrix multiplication (other * this), source matrix remains unchanged!
	 * 
	 * @param other
	 *            first matrix
	 * @return result Result matrix
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix r_multiply(RealMatrix other) throws TDataException, CloneException {

		// result (m x n) = this (m x s) * other (s x n)
		int[] sizeThis = this.size();
		int[] sizeOther = new int[] { other.getRowDimension(), other.getColumnDimension() };

		if (sizeOther[1] != sizeThis[0]) {
			throw new TDataException("Inner matrix dimension must agree");
		}

		TPeriodicMatrix result = new TPeriodicMatrix(sizeOther[0], sizeThis[1]);

		for (int i = 0; i < sizeOther[0]; i++) {
			for (int j = 0; j < sizeThis[1]; j++) {

				TPeriodic sum = new TPeriodic();

				for (int k = 0; k < sizeOther[1]; k++) {
					sum = sum.add(this.get(k, j).scale(other.getEntry(i, k)));
				}
				result.set(i, j, sum);
			}
		}
		return result;
	}

	/**
	 * Multiplication of the TPeriodicMatrix with the given TNumeric. source
	 * matrix remains unchanged!
	 * 
	 * @param other
	 *            Numeric matrix
	 * @return Result
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix multiply(TNumeric other) throws TDataException, CloneException {

		return this.multiply(other.getRealMatrix());
	}

	/**
	 * Matrix multiplication (other * this), source matrix remains unchanged!
	 * 
	 * @param other
	 *            Numeric matrix
	 * @return Result
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix r_multiply(TNumeric other) throws TDataException, CloneException {

		return this.r_multiply(other.getRealMatrix());
	}

	/**
	 * Scales a given matrix. Source remains unchanged.
	 * 
	 * @param scale
	 *            scale factor
	 * @return Scaled matrix
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix scale(double scale) throws TDataException, CloneException, Exception {
		return this.elemOp("scale", scale);
	}

	/**
	 * Phase shift the given matrix. Source object remains unchanged!
	 * 
	 * @param shift
	 *            shift angle in radiants
	 * @return Result
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws Exception
	 *             Exception by operation
	 */
	public TPeriodicMatrix shift(double shift) throws TDataException, CloneException, Exception {
		return this.elemOp("shift", shift);
	}

	/**
	 * Creates TArray
	 * 
	 * @return Array
	 * @throws TDataException
	 *             Wrong index
	 */
	public TArray toTArray() throws TDataException {

		TArray tmp = new TArray(this.size());

		for (int p = 0; p < this.size()[0]; p++) {
			for (int q = 0; q < this.size()[1]; q++) {
				tmp.set(p, q, this.get(p, q));
			}
		}

		return tmp;

	}

	/**
	 * Preview!
	 * 
	 * @return Plot of preview
	 * @throws Exception
	 *             Unable to shoe plot
	 */
	public Plot showPreview() throws Exception {
		return this.showPreview(null);
	}

	/**
	 * Preview!
	 * 
	 * @param title
	 *            Title of view
	 * @return Plot of preview
	 * @throws Exception
	 *             Unable to shoe plot
	 */
	public Plot showPreview(String title) throws Exception {
		return this.showPreview(title, 1, false);
	}

	/**
	 * Preview!
	 * 
	 * @param title
	 *            Title of view
	 * @param nPeriods
	 *            Number of periods displayed
	 * @param showFundamental
	 *            boolean.
	 * @return Plot of preview
	 * @throws Exception
	 *             Unable to shoe plot
	 */
	public Plot showPreview(String title, int nPeriods, boolean showFundamental) throws Exception {

		int[] size = this.size();

		JFrame f = new JFrame("Periodic signal matrix preview");
		f.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		SubPlot plot = new SubPlot(size[0], size[1]);
		plot.setTitle(title);

		ClassLoader classLoader = TPeriodic.class.getClassLoader();
		@SuppressWarnings("unchecked")
		Class<TPeriodicPreview> c = (Class<TPeriodicPreview>) classLoader
				.loadClass("at.lcm.bu21.general.dtypes.preview.TPeriodicPreview");
		Method mpreviewWave = c.getMethod("previewWave",
				new Class[] { TPeriodic.class, String.class, int.class, boolean.class });

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic sig = this.get(row, col);
				plot.setPlot(row, col, (Plot) mpreviewWave.invoke(null, sig, null, nPeriods, showFundamental));
			}
		}

		Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
		f.setContentPane(plot.getPanel());
		f.pack();
		f.setSize((int) (dim.width * 0.7), (int) (dim.height * 0.7));
		f.setVisible(true);

		return plot;
	}

	/**
	 * Show spectrum preview
	 * 
	 * @param title
	 *            Title of preview
	 * @return Plot of preview
	 * @throws Exception
	 *             Unable to show plot
	 */
	public Plot showSpectrum(String title) throws Exception {
		return this.showSpectrum(title, false);
	}

	/**
	 * Show spectrum preview
	 * 
	 * @param title
	 *            Title of preview
	 * @param rms
	 *            true: rms, false: magnitude
	 * @return Plot of preview
	 * @throws Exception
	 *             Unable to show plot
	 */
	public Plot showSpectrum(String title, boolean rms) throws Exception {
		int[] size = this.size();

		JFrame f = new JFrame("Periodic signal matrix spectrum");
		f.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		SubPlot plot = new SubPlot(size[0], size[1]);
		plot.setTitle(title);

		ClassLoader classLoader = TPeriodic.class.getClassLoader();
		@SuppressWarnings("unchecked")
		Class<TPeriodicPreview> c = (Class<TPeriodicPreview>) classLoader
				.loadClass("at.lcm.bu21.general.dtypes.preview.TPeriodicPreview");
		Method mpreviewSpectrum = c.getMethod("previewSpectrum",
				new Class[] { TPeriodic.class, String.class, int.class, boolean.class });

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				TPeriodic sig = this.get(row, col);
				plot.setPlot(row, col, (Plot) mpreviewSpectrum.invoke(null, sig, title, sig.getMaxOrder(), rms));
			}
		}

		Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
		f.setContentPane(plot.getPanel());
		f.pack();
		f.setSize((int) (dim.width * 0.7), (int) (dim.height * 0.7));
		f.setVisible(true);

		return plot;
	}

	/**
	 * Get selected data
	 * 
	 * @param ident
	 *            List of identifier of data to select
	 * @return Data
	 */
	@Override
	protected TPeriodic getData(String[] ident) {
		TPeriodic data, tmp;
		String[] items;
		int row, col;

		data = (TPeriodic) super.getData(ident);
		if (data == null) {
			items = ident[0].split("[,\\]]");
			if (items.length == 2) {
				try {
					row = Integer.parseInt(items[0]);
					col = Integer.parseInt(items[1]);
					tmp = this.get(row, col);
					if (tmp != null) {
						data = (TPeriodic) tmp.getData(Arrays.copyOfRange(ident, 1, ident.length));
					}
				} catch (Exception ex) {
				}
			}
		}

		return data;
	}

	/**
	 * Create a deep clone of given data-structure
	 * 
	 * @return Deep clone of structure
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	@Override
	public TPeriodicMatrix deepClone() throws CloneException {
		return (TPeriodicMatrix) Clone.deepClone(this);
	}

	/**
	 * Expand size of array
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of columns
	 */
	protected void expandArray(int row, int col) {
		ArrayList<TPeriodic> r;

		// add absent rows
		while (row > this.data.size()) {
			this.data.add(new ArrayList<TPeriodic>());
		}

		// add absent columns
		for (int i = 0; i < this.data.size(); i++) {
			r = this.data.get(i);
			while (col > r.size()) {
				r.add(new TPeriodic());
			}
		}
	}

	/**
	 * Set value of array
	 * 
	 * @param row
	 *            Index of row of array
	 * @param col
	 *            Index of column of array
	 * @param value
	 *            Value to insert
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void set(int row, int col, TPeriodic value) throws CloneException {

		if (value != null) {
			// expand array
			if ((row >= this.size()[0]) || (col >= this.size()[1])) {
				this.expandArray(Math.max(row + 1, this.size()[0]), Math.max(col + 1, this.size()[1]));
			}

			// replace value on give position
			this.data.get(row).set(col, value.deepClone());
		} else {
			throw new NullPointerException("Null is not allowed to add to TPeridic");
		}
	}

	/**
	 * Get value of array
	 * 
	 * @param row
	 *            Index of row of array
	 * @param col
	 *            Index of column of array
	 * @return Value of array
	 * @throws TDataException
	 *             Wrong index
	 */
	public TPeriodic get(int row, int col) throws TDataException {
		if ((row < this.data.size()) && (col < this.data.get(0).size())) {
			return this.data.get(row).get(col);
		} else {
			throw new TDataException("Index out of bounds!");
		}
	}

	/**
	 * Get a row of a array
	 * 
	 * @param row
	 *            Index of row of array
	 * @return Row of array
	 * @throws TDataException
	 *             Wrong index
	 */
	public TPeriodicMatrix getRow(int row) throws TDataException {
		if (row < this.data.size()) {
			return new TPeriodicMatrix(this.data.get(row));
		} else {
			throw new TDataException("Index out of bounds!");
		}
	}

	/**
	 * Get a array of the size (two dimension)
	 * 
	 * @return Size of array
	 */
	public int[] size() {
		int[] size = { 0, 0 };
		size[0] = this.data.size();
		if (size[0] != 0) {
			size[1] = this.data.get(0).size();
		}
		return size;
	}

	@Override
	public void xml2data(XMLStreamReader xmlReader) throws XMLStreamException, TDataException {
		ArrayList<TPeriodic> row;
		int nrows, ncolumns;

		nrows = Integer.parseInt(xmlReader.getAttributeValue(null, "nrows"));
		ncolumns = Integer.parseInt(xmlReader.getAttributeValue(null, "ncolumns"));
		this.data = new ArrayList<ArrayList<TPeriodic>>(nrows);

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
			row = new ArrayList<TPeriodic>(ncolumns);
			this.data.add(row);

			while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
				row.add((TPeriodic) TData.xml2obj(xmlReader));
			}
			if (row.size() != ncolumns) {
				throw new TDataException("Wrong number of elements!");
			}
			xmlReader.next();
		}
		if (this.data.size() != nrows) {
			throw new TDataException("Wrong number of elements!");
		}
		xmlReader.next();
	}

	@Override
	protected void data2xml(XMLStreamWriter xmlWriter) throws XMLStreamException {
		int rows = 0, cols = 0;
		rows = this.data.size();
		if (rows > 0) {
			cols = this.data.get(0).size();
		}

		xmlWriter.writeStartElement("PeriodicMatrix");
		xmlWriter.writeAttribute("nrows", String.valueOf(rows));
		xmlWriter.writeAttribute("ncolumns", String.valueOf(cols));
		for (int r = 0; r < this.data.size(); r++) {
			xmlWriter.writeStartElement("row");
			for (int c = 0; c < this.data.get(r).size(); c++) {
				this.data.get(r).get(c).data2xml(xmlWriter);
			}
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		TPeriodicMatrix other = (TPeriodicMatrix) obj;
		if (this.data == null) {
			if (other.data != null) {
				return false;
			}
		} else if (!this.data.equals(other.data)) {
			return false;
		}
		return true;
	}

	/**
	 * Test if equal with tolerance
	 * 
	 * @param obj
	 *            matrix to compare with
	 * @param tol
	 *            tolerance of signal difference (rms)
	 * @return true if equal
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public boolean equals(Object obj, double tol) throws CloneException, TDataException {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		TPeriodicMatrix other = (TPeriodicMatrix) obj;

		if (this.equals(obj)) {
			return true;
		}

		int[] sz = this.size();
		if (!Arrays.equals(sz, other.size())) {
			return false;
		}

		for (int p = 0; p < sz[0]; p++) {
			for (int q = 0; q < sz[1]; q++) {
				if (!this.get(p, q).equals(other.get(p, q), tol)) {
					return false;
				}
			}
		}

		return true;
	}

	@Override
	public String toString() {
		int nrow = 0, ncol = 0;
		nrow = this.data.size();
		if (nrow != 0) {
			ncol = this.data.get(0).size();
		}
		return "[" + nrow + "x" + ncol + " PeriodicMatrix]";
	}

}
