/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor;

import at.lcm.bu21.general.dtypes.TPeriodic;

/**
 * Ground Singal Generator
 * 
 * @author Christoph Halbmayr
 */
class TPeriodicGroundSignGen {

	/**
	 * maximal order
	 */
	private int maxOrder = 20;

	/**
	 * Creates an empty Object
	 */
	TPeriodicGroundSignGen() {

	}

	/**
	 * creates an Rectangular TPeriodic
	 * 
	 * @param offs
	 *            Offset
	 * @param ampl
	 *            Amplitude
	 * @param phas
	 *            Phase
	 * @param phi
	 *            Active Time
	 * @return rectangular TPeriodic
	 */
	public TPeriodic genRect(final double offs, final double ampl, double phas, double phi) {
		TPeriodic rect = new TPeriodic();

		phas *= Math.PI / 180;
		phi *= Math.PI / 180;
		phi /= 2;
		phas -= phi;

		final double[] coeffsA = new double[this.maxOrder];
		final double[] coeffsB = new double[this.maxOrder];

		final double mean = (offs + (((4 / Math.PI) * phi) / 2)) - 1.0;

		for (int i = 0; i < this.maxOrder; i++) {

			coeffsA[i] = ((4 / Math.PI) * Math.sin((i + 1) * phi)) / (i + 1);
			coeffsB[i] = 0;

		}

		rect.setCoeffs(mean, coeffsA, coeffsB);
		rect = rect.scale(ampl);
		rect = rect.shift(phas);

		return rect;
	}

	/**
	 * creates an Sinus TPeriodic
	 * 
	 * @param offs
	 *            Offset
	 * @param ampl
	 *            Amplitude
	 * @param phas
	 *            Phase
	 * @param order
	 * @return Sinus TPeriodic
	 */
	public TPeriodic genSin(final double offs, final double ampl, final double phas, final int order) {
		final TPeriodic sinus = new TPeriodic(order, ampl, (phas - 90) * (Math.PI / 180));

		return sinus;
	}

	/**
	 * creates an Triangular TPeriodic
	 * 
	 * @param offs
	 *            Offset
	 * @param ampl
	 *            Amplitude
	 * @param phas
	 *            Phase Active Time
	 * @param phi
	 * @return triangular TPeriodic
	 */
	public TPeriodic genTriangle(final double offs, final double ampl, double phas, double phi) {

		TPeriodic triangle = new TPeriodic();

		phas *= Math.PI / 180;
		phi *= Math.PI / 180;
		phi /= 2;
		phas -= phi;

		final double[] coeffsA = new double[this.maxOrder];
		final double[] coeffsB = new double[this.maxOrder];

		final double mean = (phi / (2 * Math.PI)) + offs;
		for (int i = 0; i < this.maxOrder; i++) {

			coeffsA[i] =

					((1 - Math.cos(phi * (i + 1))) / ((i + 1) * (i + 1))) * (2 / (Math.PI * phi)

					)

			;
			coeffsB[i] = 0.0;

		}

		triangle.setCoeffs(mean, coeffsA, coeffsB);
		triangle = triangle.scale(ampl);
		triangle = triangle.shift(phas);

		return triangle;
	}

	/**
	 * @param maxOrder
	 *            the maxOrder to set
	 */
	public void setMaxOrder(final int maxOrder) {
		this.maxOrder = maxOrder;
	}

}
