/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.math.interpolator2d;

import java.util.Arrays;

import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.DecompositionSolver;
import org.apache.commons.math3.linear.LUDecomposition;
import org.apache.commons.math3.linear.RealMatrix;

import at.lcm.bu21.general.dtypes.TNumeric;

/**
 * 
 * 
 * @author Gnther Weidenholzer
 */

public class RbfInt2d extends Interpolator2d {

	private double[] w;
	private double rho;
	private ConvexHull convexhull;

	/**
	 * 
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @throws Exception
	 */
	public RbfInt2d(double[] xdata, double[] ydata, double[] zdata)
			throws Exception {

		this(xdata, ydata, zdata, 0.0);
	}

	/**
	 * 
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @param rho
	 *            smoothing parameter for rbf approximation. rho=0 means exact
	 *            interpolation
	 * @throws Exception
	 */
	public RbfInt2d(double[] xdata, double[] ydata, double[] zdata, double rho)
			throws Exception {

		super(xdata, ydata, zdata);
		this.convexhull = new ConvexHull(this.xnorm, this.ynorm);
		this.rho = rho;

		this.init();

	}

	/**
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @throws Exception
	 */
	public RbfInt2d(TNumeric xdata, TNumeric ydata, TNumeric zdata)
			throws Exception {
		this(Interpolator2d.dataVectors(xdata, ydata, zdata));
	}

	/**
	 * xyz[0]...xdata xyz[1]...ydata xyz[2]...zdata
	 * 
	 * @param xyz
	 * @throws Exception
	 */
	protected RbfInt2d(double[][] xyz) throws Exception {
		this(xyz[0], xyz[1], xyz[2]);
	}

	/**
	 * 
	 * @param xdata
	 * @param ydata
	 * @param zdata
	 * @param x
	 * @param y
	 * @throws Exception
	 */
	private void init() throws Exception {

		int npoints = this.xnorm.length;

		RealMatrix l = new BlockRealMatrix(npoints + 3, npoints + 3);
		ArrayRealVector rhs = new ArrayRealVector(npoints + 3);

		// Create L-Matrix
		for (int row = 0; row < npoints; row++) {

			double xr = this.xnorm[row];
			double yr = this.ynorm[row];

			// K
			for (int col = 0; col < row; col++) {

				double xc = this.xnorm[col];
				double yc = this.ynorm[col];

				double dx = xr - xc;
				double dy = yr - yc;

				double r = Math.sqrt((dx * dx) + (dy * dy));

				double tmp = Math.pow(r, 3.0);
				l.setEntry(row, col, tmp);
				l.setEntry(col, row, tmp);

			}

			// approximation!
			l.setEntry(row, row, this.rho);

			// P
			l.setEntry(row, npoints, 1.0);
			l.setEntry(row, npoints + 1, xr);
			l.setEntry(row, npoints + 2, yr);

			// P^T
			l.setEntry(npoints, row, 1.0);
			l.setEntry(npoints + 1, row, xr);
			l.setEntry(npoints + 2, row, yr);

			// right hand side
			rhs.setEntry(row, this.zdata[row]);

		}

		rhs.setEntry(npoints, 0.0);
		rhs.setEntry(npoints + 1, 0.0);
		rhs.setEntry(npoints + 2, 0.0);

		// solve equation system
		DecompositionSolver solver = new LUDecomposition(l).getSolver();
		this.w = solver.solve(rhs).toArray();

	}

	public String hotIntCode(String name) {

		String str = name + "\n{\n";

		String str2 = String.format("scaling_vector = [%1.8g,%1.8g]\n",
				this.xrange, this.yrange);

		str2 += "centers = [";
		for (int k = 0; k < this.xnorm.length; k++) {
			if (k > 0) {
				str2 += "; ";
			}
			str2 += String.format("%1.8g,%1.8g", this.xnorm[k], this.ynorm[k]);
		}
		str2 += "]\n";

		str2 += String.format("weights_RBF = %s\n",
				Arrays.toString(Arrays.copyOf(this.w, this.xnorm.length)));

		str2 += String.format("weights_poly = %s\n", Arrays.toString(Arrays
				.copyOfRange(this.w, this.xnorm.length, this.w.length)));

		str2 += "RBF_kernel = 1\n";

		// indent
		str += str2.replaceAll("(?m)^", "\t");
		str += "}\n";

		return str;

	}

	@Override
	public double value_(double x, double y) {

		int npoints = this.xnorm.length;

		double a1 = this.w[npoints];
		double ax = this.w[npoints + 1];
		double ay = this.w[npoints + 2];

		double sum = a1 + (ax * x) + (ay * y);

		for (int k = 0; k < npoints; k++) {

			double r2 = ((x - this.xnorm[k]) * (x - this.xnorm[k]))
					+ ((y - this.ynorm[k]) * (y - this.ynorm[k]));
			sum += this.w[k] * Math.pow(r2, 3. / 2.);
		}

		return sum;
	}

	@Override
	public double dz_dx_(double x, double y) {

		int npoints = this.xnorm.length;

		double sum = this.w[npoints + 1]; // ax

		for (int k = 0; k < npoints; k++) {

			double dx = x - this.xnorm[k];
			double dy = y - this.ynorm[k];
			double r = Math.hypot(dx, dy);
			sum += this.w[k] * 3. * r * dx;
		}

		return sum;
	}

	@Override
	public double dz_dy_(double x, double y) {

		int npoints = this.xnorm.length;

		double sum = this.w[npoints + 2]; // ay

		for (int k = 0; k < npoints; k++) {

			double dx = x - this.xnorm[k];
			double dy = y - this.ynorm[k];
			double r = Math.hypot(dx, dy);
			sum += this.w[k] * 3. * r * dy;
		}

		return sum;
	}

	@Override
	public double dz_dx2_(double x, double y) {

		int npoints = this.xnorm.length;

		double sum = 0;

		for (int k = 0; k < npoints; k++) {

			double dx = x - this.xnorm[k];
			double dy = y - this.ynorm[k];
			double r = Math.hypot(dx, dy);
			sum += this.w[k] * 3. * (r + ((dx * dx) / r));
		}

		return sum;
	}

	@Override
	public double dz_dy2_(double x, double y) {

		int npoints = this.xnorm.length;

		double sum = 0;

		for (int k = 0; k < npoints; k++) {

			double dx = x - this.xnorm[k];
			double dy = y - this.ynorm[k];
			double r = Math.hypot(dx, dy);
			sum += this.w[k] * 3. * (r + ((dy * dy) / r));
		}

		return sum;
	}

	@Override
	public double dz_dxy_(double x, double y) {

		int npoints = this.xnorm.length;

		double sum = 0;

		for (int k = 0; k < npoints; k++) {

			double dx = x - this.xnorm[k];
			double dy = y - this.ynorm[k];
			double r = Math.hypot(dx, dy);
			sum += (this.w[k] * 3. * dx * dy) / r;
		}

		return sum;
	}

	@Override
	boolean contains_(double x, double y) {
		return this.convexhull.contains(x, y);
	}

}
