/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import java.util.LinkedList;

/**
 * Link selection of different plots
 * 
 * @author Werner Koppelsttter
 */
public class LinkPlot {

	/** List of plot witch are linked with this object */
	private LinkedList<Plot> list = new LinkedList<Plot>();

	/**
	 * Create LinkPlot object
	 */
	public LinkPlot() {
	}

	/**
	 * Create standard link-object for internal use
	 * 
	 * @param plot
	 *            Plot to link
	 */
	LinkPlot(Plot plot) {
		this.list.add(plot);
	}

	/**
	 * Add plot to link-object
	 * 
	 * @param plot
	 *            Plot to link
	 */
	public void addPlot(Plot plot) {
		SubPlot sub;

		if (plot != null) {
			if (plot instanceof SubPlot) {
				sub = (SubPlot) plot;
				for (int row = 0; row < sub.getSize()[0]; row++) {
					for (int col = 0; col < sub.getSize()[1]; col++) {
						this.addPlot(sub.getPlot(row, col));
					}
				}
			} else {
				if (!this.list.contains(plot)) {
					this.list.add(plot);
					plot.getLink().rmPlot(plot);
					plot.setLink(this);
					this.clearSelect();
				}
			}

		}
	}

	/**
	 * Remove all plots form link-object (unlink)
	 */
	public void rmAllPlots() {
		while (!this.list.isEmpty()) {
			this.rmPlot(this.list.getFirst());
		}
	}

	/**
	 * Remove plot from link-object (unlink)
	 * 
	 * @param plot
	 *            Plot to unlink
	 */
	public void rmPlot(Plot plot) {
		SubPlot sub;

		if (plot instanceof SubPlot) {
			sub = (SubPlot) plot;
			for (int row = 0; row < sub.getSize()[0]; row++) {
				for (int col = 0; col < sub.getSize()[1]; col++) {
					this.rmPlot(sub.getPlot(row, col));
				}
			}
		} else {
			this.clearSelect();
			this.clearSuppress();
			this.list.remove(plot);
			plot.setLink(new LinkPlot(plot));
		}
	}

	/**
	 * Suppress value of signal
	 * 
	 * @param signal
	 *            Index of signal
	 * @param index
	 *            Index of value
	 * @param suppress
	 *            Suppress state
	 */
	public void suppress(int signal, int index, boolean suppress) {
		for (Plot plot : this.list) {
			plot.suppress(signal, index, suppress);
		}
	}

	/**
	 * Suppress selected/non-selected values of signal
	 * 
	 * @param signal
	 *            Index of signal
	 * @param inverse
	 *            Selected or non selected values
	 */
	public void suppressSelected(int signal, boolean inverse) {
		for (Plot plot : this.list) {
			plot.suppressSelected(signal, inverse);
		}
	}

	/**
	 * Clear suppress of all plots
	 */
	public void clearSuppress() {
		for (Plot plot : this.list) {
			plot.clearSuppress();
		}
	}

	/**
	 * Set selection of index to signal of all plots
	 * 
	 * @param signal
	 *            Index signal
	 * @param index
	 *            Index of data-point
	 * @param select
	 *            Selection state
	 */
	public void select(int signal, int index, boolean select) {
		for (Plot plot : this.list) {
			plot.select(signal, index, select);
		}
	}

	/**
	 * Clear selection of all plots
	 */
	public void clearSelect() {
		for (Plot plot : this.list) {
			plot.clearSelect();
		}
	}

	/**
	 * Redraw all plots
	 */
	public void redraw() {
		for (Plot plot : this.list) {
			plot.redraw();
		}
	}

	/**
	 * Reset all plots
	 */
	public void reset() {
		for (Plot plot : this.list) {
			plot.reset();
		}
	}
}
