/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;

import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Class for plotting singals
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public abstract class Plot2XY extends Plot2D {

	/**
	 * Create empty object
	 */
	protected Plot2XY() {
		super();
	}

	/**
	 * Adds a signal to the plot
	 * 
	 * @param name
	 *            Name of signal
	 * @param ySignal
	 *            y-vector
	 * @throws PlotException
	 *             Exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[] ySignal)
			throws PlotException, TDataException, CloneException {
		this.addSignal(new Signal(name, ySignal));
	}

	/**
	 * Adds a signal to the plot
	 * 
	 * @param name
	 *            Name of signal
	 * @param ySignal
	 *            y-vector
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric ySignal)
			throws PlotException, TDataException, CloneException {
		if (ySignal.isReal()) {
			this.addSignal(new Signal(name, ySignal));
		} else {
			this.addSignal(
					new Signal(name, new TNumeric(ySignal.getRealMatrix()),
							new TNumeric(ySignal.getImagMatrix())));
		}
	}

	/**
	 * Add a signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param xSignal
	 *            x-vector
	 * @param ySignal
	 *            y-vector
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, double[] xSignal, double[] ySignal)
			throws PlotException, TDataException, CloneException {
		this.addSignal(new Signal(name, xSignal, ySignal));
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param xSignal
	 *            x-vector
	 * @param ySignal
	 *            y-vector
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TNumeric xSignal, TNumeric ySignal)
			throws PlotException, TDataException, CloneException {
		this.addSignal(new Signal(name, xSignal, ySignal));
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param per
	 *            TPeriodic signal
	 * @throws PlotException
	 *             Plot exception
	 */
	public void addSignal(String name, TPeriodic per) throws PlotException {
		if (this.plot.getDomainAxis() instanceof NumberAxis) {
			((NumberAxis) this.plot.getDomainAxis())
					.setTickUnit(new NumberTickUnit(30.0D));
		}

		this.addSignal(name, per, 360);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param per
	 *            TPeriodic signal
	 * @param nPoints
	 *            Numer of point to plot
	 * @throws PlotException
	 *             Plot exception
	 */
	public void addSignal(String name, TPeriodic per, int nPoints)
			throws PlotException {
		Signal sig;
		double[] phi;

		sig = new Signal(name, per, nPoints);
		phi = sig.getXRef();
		for (int k = 0; k < phi.length; k++) {
			phi[k] *= 180.0 / Math.PI;
		}
		this.addSignal(sig);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param x
	 *            TPeriodic x-signal
	 * @param y
	 *            TPeriodic y-signal
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TPeriodic x, TPeriodic y)
			throws PlotException, TDataException, CloneException {
		this.addSignal(name, x, y, 360);
	}

	/**
	 * Add signal to the plot
	 * 
	 * @param name
	 *            Signal name
	 * @param x
	 *            TPeriodic x-signal
	 * @param y
	 *            TPeriodic y-signal
	 * @param nPoints
	 *            Number of point to plot
	 * @throws PlotException
	 *             Plot exception
	 * @throws TDataException
	 *             Unable to create data-object
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public void addSignal(String name, TPeriodic x, TPeriodic y, int nPoints)
			throws PlotException, TDataException, CloneException {
		double[] phi;

		phi = new double[nPoints + 1];
		for (int k = 0; k <= nPoints; k++) {
			phi[k] = ((double) k / nPoints) * 2 * Math.PI;
		}

		this.addSignal(name, x.valueAt(phi), y.valueAt(phi));
	}

	// /**
	// * Sets the x-axis log scale
	// *
	// * @param log
	// * Logarithmic x-axis
	// */
	// public void setXLog(boolean log) {
	// if (log == true) {
	// try {
	// this.plot.setDomainAxis(new LogAxis(this.getXLabel()));
	// // this.plot.setDomainAxis(new
	// // LogarithmicAxis(this.getXLabel()));
	// } catch (Exception e) {
	// }
	// } else {
	// this.plot.setDomainAxis(new NumberAxis(this.getXLabel()));
	// }
	// }

	// /**
	// * Sets the y-axis log scale
	// *
	// * @param log
	// * Logarithmic y-axis
	// */
	// public void setYLog(boolean log) {
	// if (log == true) {
	// try {
	//
	// this.plot.setRangeAxis(new LogarithmicAxis(this.getYLabel()));
	// } catch (Exception e) {
	// }
	// } else {
	// this.plot.setRangeAxis(new NumberAxis(this.getXLabel()));
	// }
	//
	// }

}
