/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.MathUtils;

/**
 * Class with all information of a signal to plot
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */
public class Signal {
	/** Signal name */
	private String name;

	/** Drawing Style */
	private Style style;

	/** Drawing Style */
	private Style selectStyle;

	/** x-values to plot */
	private double[] xValues;

	/** y-values to plot */
	private double[] yValues;

	/** z-values to plot */
	private double[] zValues;

	/** Is value is selected */
	private boolean[] selected;

	/** Is value suppressed */
	private boolean[] suppressed;

	/** Is signal enabled */
	private boolean enable = true;

	/**
	 * Create signal of vector
	 * 
	 * @param name
	 *            Name of signal
	 * @param yValues
	 *            Vector of y-values
	 * @throws PlotException
	 *             Plot exception
	 */
	public Signal(String name, double[] yValues) throws PlotException {
		double[] xValues;

		xValues = new double[yValues.length];
		for (int k = 0; k < xValues.length; k++) {
			xValues[k] = k + 1;
		}
		this.setData(name, xValues, yValues.clone(),
				new double[yValues.length]);
	}

	/**
	 * Create signal of vector
	 * 
	 * @param name
	 *            Name of signal
	 * @param xValues
	 *            Vector of x-values
	 * @param yValues
	 *            Vector of y-values
	 * @throws PlotException
	 *             Plot exception
	 */
	public Signal(String name, double[] xValues, double[] yValues)
			throws PlotException {
		this.setData(name, xValues.clone(), yValues.clone(),
				new double[xValues.length]);
	}

	/**
	 * Create signal of vector
	 * 
	 * @param name
	 *            Name of signal
	 * @param xValues
	 *            Vector of x-values
	 * @param yValues
	 *            Vector of y-values
	 * @param zValues
	 *            Vector of z-values
	 * @throws PlotException
	 *             Plot exception
	 */
	public Signal(String name, double[] xValues, double[] yValues,
			double[] zValues) throws PlotException {
		this.setData(name, xValues.clone(), yValues.clone(), zValues.clone());
	}

	/**
	 * Create signal of vector
	 * 
	 * @param name
	 *            Name of signal
	 * @param xValues
	 *            Vector of x-values
	 * @param yValues
	 *            Vector of y-values
	 * @param zValues
	 *            Vector of z-values
	 * @throws PlotException
	 *             Plot exception
	 */
	public Signal(String name, double[] xValues, double[] yValues,
			double[][] zValues) throws PlotException {
		this.setData(name, xValues, yValues, zValues);
	}

	/**
	 * Create signal of a TNumeric. Only vector is allowed. If TNumeric is
	 * complex then x-axis is real and y-axis is imaginary
	 * 
	 * @param name
	 *            Name of signal
	 * @param yData
	 *            Data of y-signal
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public Signal(String name, TNumeric yData)
			throws PlotException, CloneException {
		double[] xVal, yVal;

		if (!yData.isReal()) {
			throw new PlotException("Data must be real!");
		}

		// y-vector
		if ((yData.size()[0] == 1) || (yData.size()[1] == 1)) {
			if (yData.size()[0] == 1) {
				yVal = yData.getRealMatrix().getRow(0);
			} else {
				yVal = yData.getRealMatrix().getColumn(0);
			}
		} else {
			throw new PlotException("y-data must be a vector!");
		}

		xVal = new double[yVal.length];
		for (int k = 0; k < xVal.length; k++) {
			xVal[k] = k + 1;
		}

		this.setData(name, xVal, yVal, new double[yVal.length]);
	}

	/**
	 * Create signal of TNumerics. Only real vectors are allowed.
	 * 
	 * @param name
	 *            Name of signal
	 * @param xData
	 *            Data of x-signal
	 * @param yData
	 *            Data of y-signal
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public Signal(String name, TNumeric xData, TNumeric yData)
			throws PlotException, CloneException {
		double[] xVal, yVal;

		if (!xData.isReal() || !yData.isReal()) {
			throw new PlotException("Data must be real!");
		}

		// x-vector
		if ((xData.size()[0] == 1) || (xData.size()[1] == 1)) {
			if (xData.size()[0] == 1) {
				xVal = xData.getRealMatrix().getRow(0);
			} else {
				xVal = xData.getRealMatrix().getColumn(0);
			}
		} else {
			throw new PlotException("x-data must be a vector!");
		}

		// y-vector
		if ((yData.size()[0] == 1) || (yData.size()[1] == 1)) {
			if (yData.size()[0] == 1) {
				yVal = yData.getRealMatrix().getRow(0);
			} else {
				yVal = yData.getRealMatrix().getColumn(0);
			}
		} else {
			throw new PlotException("y-data must be a vector!");
		}

		this.setData(name, xVal, yVal, new double[yVal.length]);
	}

	/**
	 * Create signal of TNumerics. x,y,z must have the same length or z is a
	 * matrix (nxm), x(n), y(m)
	 * 
	 * @param name
	 *            Name of signal
	 * @param xData
	 *            Data of x-signal
	 * @param yData
	 *            Data of y-signal
	 * @param zData
	 *            Data of z-signal
	 * @throws PlotException
	 *             Plot exception
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public Signal(String name, TNumeric xData, TNumeric yData, TNumeric zData)
			throws PlotException, CloneException {
		double[] xVal, yVal;

		if (!xData.isReal() || !yData.isReal() || !zData.isReal()) {
			throw new PlotException("Data must be real!");
		}

		// x-vector
		if ((xData.size()[0] == 1) || (xData.size()[1] == 1)) {
			if (xData.size()[0] == 1) {
				xVal = xData.getRealMatrix().getRow(0);
			} else {
				xVal = xData.getRealMatrix().getColumn(0);
			}
		} else {
			throw new PlotException("x-data must be a vector!");
		}

		// y-vector
		if ((yData.size()[0] == 1) || (yData.size()[1] == 1)) {
			if (yData.size()[0] == 1) {
				yVal = yData.getRealMatrix().getRow(0);
			} else {
				yVal = yData.getRealMatrix().getColumn(0);
			}
		} else {
			throw new PlotException("y-data must be a vector!");
		}

		// z-data
		if ((xVal.length == yVal.length)
				&& ((zData.size()[0] == 1) || (zData.size()[1] == 1))) {
			if (xData.size()[0] == 1) {
				this.setData(name, xVal, yVal, zData.getRealMatrix().getRow(0));
			} else {
				this.setData(name, xVal, yVal,
						zData.getRealMatrix().getColumn(0));
			}
		} else {
			this.setData(name, xVal, yVal, zData.getRealMatrix().getData());
		}

	}

	/**
	 * Create signal of TPeriodic
	 * 
	 * @param name
	 *            Name of signal
	 * @param per
	 *            Periodic signal
	 * @param nPoints
	 *            Number of used points
	 * @throws PlotException
	 *             Plot exception
	 */
	public Signal(String name, TPeriodic per, int nPoints)
			throws PlotException {
		double tmp;
		double[] xValues;

		tmp = (2.0 * Math.PI) / nPoints;
		xValues = new double[nPoints + 1];
		for (int k = 0; k <= nPoints; k++) {
			xValues[k] = k * tmp;
		}
		this.setData(name, xValues, per.valueAt(xValues),
				new double[xValues.length]);

	}

	/**
	 * Set data of signal
	 * 
	 * @param name
	 *            Name
	 * @param xValues
	 *            x-vector
	 * @param yValues
	 *            y-vector
	 * @param zValues
	 *            z-matrix
	 * @throws PlotException
	 *             Plot exception
	 */
	private void setData(String name, double[] xValues, double[] yValues,
			double[][] zValues) throws PlotException {
		int len, index;
		double[] xVal, yVal, zVal;

		if ((yValues.length != zValues.length)
				|| (xValues.length != zValues[0].length)) {
			throw new PlotException("Data dimension must agree!");
		}

		len = xValues.length * yValues.length;
		xVal = new double[len];
		yVal = new double[len];
		zVal = new double[len];
		for (int y = 0; y < yValues.length; y++) {
			for (int x = 0; x < xValues.length; x++) {
				index = x + (y * xValues.length);
				xVal[index] = xValues[x];
				yVal[index] = yValues[y];
				zVal[index] = zValues[y][x];
			}
		}

		this.setData(name, xVal, yVal, zVal);
	}

	/**
	 * Set data of signal
	 * 
	 * @param name
	 *            Name
	 * @param xValues
	 *            x-vector
	 * @param yValues
	 *            y-vector
	 * @param zValues
	 *            z-vector
	 * @throws PlotException
	 *             Plot exception
	 */
	private void setData(String name, double[] xValues, double[] yValues,
			double[] zValues) throws PlotException {
		this.setName(name);
		if ((xValues.length == yValues.length)
				&& (xValues.length == zValues.length)) {
			this.xValues = xValues;
			this.yValues = yValues;
			this.zValues = zValues;
			this.selected = new boolean[xValues.length];
			this.suppressed = new boolean[xValues.length];
		} else {
			throw new PlotException("Data dimension must agree!");
		}
	}

	/**
	 * Get non suppressed values of signal. Order x,y,z
	 * 
	 * @return Matrix with signal values
	 */
	double[][] getValues() {
		double[][] data;
		int index = 0;

		data = new double[3][this.sizeNonSuppress()];
		for (int k = 0; k < this.suppressed.length; k++) {
			if (!this.suppressed[k]) {
				data[0][index] = this.xValues[k];
				data[1][index] = this.yValues[k];
				data[2][index] = this.zValues[k];
				index++;
			}
		}

		return data;
	}

	/**
	 * Get selected values of signal. Order x,y,z
	 * 
	 * @return Matrix with selected signal values
	 */
	double[][] getSelectedValues() {
		double[][] data;
		int index = 0;

		data = new double[3][this.sizeSelect()];
		for (int k = 0; k < this.suppressed.length; k++) {
			if (!this.suppressed[k] && this.selected[k]) {
				data[0][index] = this.xValues[k];
				data[1][index] = this.yValues[k];
				data[2][index] = this.zValues[k];
				index++;
			}
		}

		return data;
	}

	/**
	 * Get selected indices of signal.
	 * 
	 * @return Matrix with selected signal indices
	 */
	public int[] getSelectedIndices() {
		int[] indices;
		int index = 0;

		indices = new int[this.sizeSelect()];
		for (int k = 0; k < this.suppressed.length; k++) {
			if (!this.suppressed[k] && this.selected[k]) {
				indices[index] = k;
				index++;
			}
		}

		return indices;
	}

	/**
	 * Get suppressed indices of signal.
	 * 
	 * @return Matrix with suppressed signal indices
	 */
	public int[] getSuppressedIndices() {
		int[] indices;
		int index = 0;

		indices = new int[this.size() - this.sizeNonSuppress()];
		for (int k = 0; k < this.suppressed.length; k++) {
			if (this.suppressed[k]) {
				indices[index] = k;
				index++;
			}
		}

		return indices;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * @param name
	 *            the name to set
	 */
	public void setName(String name) {
		if (name == null) {
			this.name = "";
		} else {
			this.name = name;
		}
	}

	/**
	 * Get style of signal
	 * 
	 * @return the style
	 */
	public Style getStyle() {
		return this.style;
	}

	/**
	 * Set style of signal
	 * 
	 * @param style
	 *            the style to set
	 */
	public void setStyle(Style style) {
		this.style = style;
	}

	/**
	 * Get select style of signal
	 * 
	 * @return the style
	 */
	public Style getSelectStyle() {
		return this.selectStyle;
	}

	/**
	 * Set select style of signal
	 * 
	 * @param style
	 *            the style to set
	 */
	public void setSelectStyle(Style style) {
		this.selectStyle = style;
	}

	/**
	 * Get non suppressed x-Values
	 * 
	 * @return x-values
	 */
	public double[] getX() {
		double[] values;
		int index = 0;

		values = new double[this.sizeNonSuppress()];
		for (int k = 0; k < this.xValues.length; k++) {
			if (!this.suppressed[k]) {
				values[index++] = this.xValues[k];
			}
		}
		return values;
	}

	/**
	 * Get reference of all x-values
	 * 
	 * @return Reference to x-values
	 */
	double[] getXRef() {
		return this.xValues;
	}

	/**
	 * Get min of non suppressed x-value
	 * 
	 * @return Min x
	 */
	public double getMinX() {
		double min;

		min = MathUtils.min(this.getX());
		if (this.getStyle().getSignalStyle() == SignalStyle.Bar) {
			min -= 0.5;
		}

		return min;
	}

	/**
	 * Get max of non suppressed x-value
	 * 
	 * @return Max x
	 */
	public double getMaxX() {
		double max;

		max = MathUtils.max(this.getX());
		if (this.getStyle().getSignalStyle() == SignalStyle.Bar) {
			max += 0.5;
		}
		return max;
	}

	/**
	 * Get non suppressed y-Values
	 * 
	 * @return y-values
	 */
	public double[] getY() {
		double[] values;
		int index = 0;

		values = new double[this.sizeNonSuppress()];
		for (int k = 0; k < this.yValues.length; k++) {
			if (!this.suppressed[k]) {
				values[index++] = this.yValues[k];
			}
		}
		return values;
	}

	/**
	 * Get reference of all y-values
	 * 
	 * @return Reference to y-values
	 */
	double[] getYRef() {
		return this.yValues;
	}

	/**
	 * Get min of non suppressed y-value
	 * 
	 * @return Min y
	 */
	public double getMinY() {
		return MathUtils.min(this.getY());
	}

	/**
	 * Get max of non suppressed y-value
	 * 
	 * @return Max y
	 */
	public double getMaxY() {
		return MathUtils.max(this.getY());
	}

	/**
	 * Get non suppressed z-Values
	 * 
	 * @return z-values
	 */
	public double[] getZ() {
		double[] values;
		int index = 0;

		values = new double[this.sizeNonSuppress()];
		for (int k = 0; k < this.zValues.length; k++) {
			if (!this.suppressed[k]) {
				values[index++] = this.zValues[k];
			}
		}
		return values;
	}

	/**
	 * Get reference of all z-values
	 * 
	 * @return Reference to z-values
	 */
	double[] getZRef() {
		return this.zValues;
	}

	/**
	 * Get min of non suppressed z-value
	 * 
	 * @return Min z
	 */
	public double getMinZ() {
		return MathUtils.min(this.getZ());
	}

	/**
	 * Get max of non suppressed z-value
	 * 
	 * @return Max z
	 */
	public double getMaxZ() {
		return MathUtils.max(this.getZ());
	}

	/**
	 * Clear selected indices
	 */
	void clearSelect() {
		this.selected = new boolean[this.selected.length];
	}

	/**
	 * Is one index selected
	 * 
	 * @return Is selected
	 */
	boolean isSelected() {
		boolean select = false;

		for (boolean sel : this.selected) {
			select |= sel;
		}

		return select;
	}

	/**
	 * Is index selected
	 * 
	 * @param index
	 *            Index to check
	 * @return Is selected
	 */
	public boolean isSelected(int index) {
		boolean select = false;

		if ((0 <= index) && (index < this.suppressed.length)) {
			if (!this.suppressed[index]) {
				select = this.selected[index];
			}
		}

		return select;
	}

	/**
	 * Set selection of index
	 * 
	 * @param index
	 *            Index to change
	 * @param select
	 *            Selection state
	 */
	void select(int index, boolean select) {

		if ((0 <= index) && (index < this.suppressed.length)) {
			if (!this.suppressed[index]) {
				this.selected[index] = select;
			}
		}
	}

	/**
	 * Clear selected indices
	 */
	void clearSuppress() {
		this.suppressed = new boolean[this.suppressed.length];
	}

	/**
	 * Is a index suppressed
	 * 
	 * @return Suppressed
	 */
	boolean isSuppressed() {
		boolean suppress = false;

		for (boolean s : this.suppressed) {
			suppress |= s;
		}

		return suppress;
	}

	/**
	 * Is index selected
	 * 
	 * @param index
	 *            Index to check
	 * @return Is selected
	 */
	public boolean isSuppressed(int index) {
		boolean suppress = false;

		if ((0 <= index) && (index < this.suppressed.length)) {
			suppress = this.suppressed[index];
		}

		return suppress;
	}

	/**
	 * Suppress index
	 * 
	 * @param index
	 *            Index of not suppressed to change
	 * @param suppress
	 *            Suppress state
	 */
	void suppress(int index, boolean suppress) {

		if ((0 <= index) && (index < this.suppressed.length)) {
			if (!this.suppressed[index]) {
				this.suppressed[index] = suppress;
				this.selected[index] = false;
			}
		}
	}

	/**
	 * Suppress selected
	 * 
	 * @param inverse
	 *            Suppress selected or not selected
	 */
	void suppressSelected(boolean inverse) {
		for (int k = 0; k < this.suppressed.length; k++) {
			if (!this.suppressed[k]) {
				if (inverse) {
					this.suppressed[k] = !this.selected[k];
				} else {
					this.suppressed[k] = this.selected[k];
				}
			}
		}
		this.clearSelect();
	}

	/**
	 * Get number of selected points in signal
	 * 
	 * @return Size
	 */
	public int sizeSelect() {
		int count = 0;

		for (int k = 0; k < this.selected.length; k++) {
			if (!this.suppressed[k] && this.selected[k]) {
				count++;
			}
		}

		return count;
	}

	/**
	 * Get number of non suppressed points in signal
	 * 
	 * @return Size
	 */
	public int sizeNonSuppress() {
		int count = 0;

		for (boolean b : this.suppressed) {
			if (!b) {
				count++;
			}
		}

		return count;
	}

	/**
	 * Get number of points in signal
	 * 
	 * @return Size
	 */
	public int size() {
		return this.suppressed.length;
	}

	/**
	 * Is signal enable to select
	 * 
	 * @return Is signal enable
	 */
	public boolean isEnable() {
		return this.enable;
	}

	/**
	 * Set signal enable
	 * 
	 * @param enable
	 *            Enable of signal
	 */
	public void setEnable(boolean enable) {
		this.enable = enable;
	}
}
