/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.utils;

import java.io.File;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

import javax.swing.JPanel;

import com.sun.pdfview.PDFFile;
import com.sun.pdfview.PagePanel;

/**
 * Pdf-viewer
 * 
 * @author Werner Koppelsttter
 */
public class PDFViewer {

	/** Number of page */
	private int page = 1;

	/** Pdf data */
	private PDFFile pdfFile = null;

	/** Panel of viewer */
	private PagePanel pdfPanel = new PagePanel();

	/** Reference to refresh-thread */
	private transient Thread refreshThread = null;

	/**
	 * Get Panel of pdf-viewer
	 * 
	 * @return Panel of viewer
	 */
	public JPanel getPanel() {
		return this.pdfPanel;
	}

	/**
	 * Set file to display
	 * 
	 * @param file
	 *            Pdf-file
	 */
	public void setFile(String file) {
		this.setFile(new File(file));
	}

	/**
	 * Set file to display
	 * 
	 * @param file
	 *            Pdf-file
	 */
	public void setFile(File file) {
		FileChannel fc;
		ByteBuffer buf;

		try {
			fc = new RandomAccessFile(file, "r").getChannel();
			buf = fc.map(FileChannel.MapMode.READ_ONLY, 0, fc.size());
			this.pdfFile = new PDFFile(buf);
			this.page = 1;
		} catch (Exception e) {
			this.pdfFile = null;
		}

		this.refresh();
	}

	/**
	 * Get number of pages
	 * 
	 * @return Number of pages
	 */
	public int getNumPages() {
		int pages = 0;

		if (this.pdfFile != null) {
			pages = this.pdfFile.getNumPages();
		}

		return pages;
	}

	/**
	 * Get number of page displayed
	 * 
	 * @return Page number
	 */
	public int getPage() {
		return this.page;
	}

	/**
	 * Set page to display
	 * 
	 * @param page
	 *            Page number
	 */
	public void setPage(int page) {
		if ((0 < page) && (page <= this.getNumPages())) {
			this.page = page;
		}
		this.refresh();
	}

	/**
	 * Refresh display
	 */
	private void refresh() {
		if (this.pdfFile == null) {
			this.pdfPanel.showPage(null);
		} else {
			if (this.refreshThread == null) {
				// refresh-thread needed because show page works only with
				// showing panel
				this.refreshThread = new Thread() {
					@Override
					public void run() {
						while (!PDFViewer.this.pdfPanel.isShowing()) {
							try {
								Thread.sleep(50);
							} catch (InterruptedException e) {
							}
						}
						PDFViewer.this.pdfPanel.showPage(PDFViewer.this.pdfFile.getPage(PDFViewer.this.page));
						PDFViewer.this.refreshThread = null;
					}
				};
				this.refreshThread.start();
			}
		}
	}

}
