/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.utils;

import java.util.ArrayList;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

/**
 * Reads a xml-file in a inner structure for changes
 * 
 * @author Werner Koppelsttter
 */
public class XMLNode {

	/**
	 * Name of XML Node
	 */
	private String name = null;

	/**
	 * List of attributes
	 */
	private ArrayList<Pair<String, String>> attributes = null;

	/**
	 * List of sub-nodes
	 */
	private ArrayList<XMLNode> nodes = null;

	/**
	 * Text content of a node
	 */
	private String text = null;

	/**
	 * Create a node
	 * 
	 * @param name
	 *            Name of node
	 */
	public XMLNode(String name) {
		this.name = name;
	}

	/**
	 * Create node from xml-file with given xmlReader
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws XMLStreamException
	 *             Unable to read from XML
	 */
	public XMLNode(XMLStreamReader xmlReader) throws XMLStreamException {

		this.name = xmlReader.getLocalName();
		// Attributes of nodes
		for (int i = 0; i < xmlReader.getAttributeCount(); i++) {
			if (this.attributes == null) {
				this.attributes = new ArrayList<Pair<String, String>>(
						xmlReader.getAttributeCount());
			}
			this.attributes.add(
					new Pair<String, String>(xmlReader.getAttributeLocalName(i),
							xmlReader.getAttributeValue(i)));
		}
		xmlReader.next();

		// Other nodes/text
		if (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
			switch (xmlReader.getEventType()) {
			case XMLStreamConstants.CHARACTERS:
				// character element
				this.text = new String();
				while (xmlReader
						.getEventType() == XMLStreamConstants.CHARACTERS) {
					this.text += xmlReader.getText();
					xmlReader.next();
				}
				break;
			case XMLStreamConstants.START_ELEMENT:
				this.nodes = new ArrayList<XMLNode>(1);
				while (xmlReader
						.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.nodes.add(new XMLNode(xmlReader));
				}
				break;
			default:
				throw new XMLStreamException("Invalid tag by reading File");
			}
		}

		xmlReader.next();
	}

	@Override
	public XMLNode clone() {
		XMLNode clone;

		clone = new XMLNode(this.getName());
		clone.text = this.text;
		if (this.attributes != null) {
			clone.attributes = new ArrayList<Pair<String, String>>(
					this.attributes.size());
			for (Pair<String, String> p : this.attributes) {
				clone.attributes.add(
						new Pair<String, String>(p.getFirst(), p.getSecond()));
			}
		}

		return clone;
	}

	/**
	 * Create a deep clone of object
	 * 
	 * @return Cloned node
	 */
	public XMLNode deepClone() {
		XMLNode clone;

		clone = this.clone();

		if (this.nodes != null) {
			clone.nodes = new ArrayList<XMLNode>(this.nodes.size());
			for (XMLNode n : this.nodes) {
				clone.nodes.add(n.deepClone());
			}
		}

		return clone;
	}

	/**
	 * Write node to xml-file with given xmlWriter
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws XMLStreamException
	 *             Unable to write to XML
	 */
	public void writeXML(XMLStreamWriter xmlWriter) throws XMLStreamException {

		xmlWriter.writeStartElement(this.name);
		// Attributes
		if (this.attributes != null) {
			for (Pair<String, String> attr : this.attributes) {
				xmlWriter.writeAttribute(attr.getFirst(), attr.getSecond());
			}
		}

		// Elements
		if (this.nodes != null) {
			// sub-nodes
			for (int i = 0; i < this.nodes.size(); i++) {
				this.nodes.get(i).writeXML(xmlWriter);
			}
			if (this.text != null) {
				// TODO Execption
			}
		} else {
			if (this.text != null) {
				// character
				xmlWriter.writeCharacters(this.text);
			}
		}
		xmlWriter.writeEndElement();
	}

	/**
	 * Get name of node
	 * 
	 * @return Name of node
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Set name of node
	 * 
	 * @param name
	 *            Nam of node
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 * Get text of node
	 * 
	 * @return Text of node
	 */
	public String getText() {
		return this.text;
	}

	/**
	 * Set text of node
	 * 
	 * @param text
	 *            Text of node
	 */
	public void setText(String text) {
		this.text = text;
	}

	/**
	 * Get array of attributes
	 * 
	 * @return List of attributes
	 */
	public String[][] getAttributes() {
		String[][] attr;
		Pair<String, String> a;

		if (this.attributes != null) {
			attr = new String[this.attributes.size()][];
			for (int k = 0; k < attr.length; k++) {
				a = this.attributes.get(k);
				attr[k] = new String[] { a.getFirst(), a.getSecond() };
			}
		} else {
			attr = new String[0][];
		}

		return attr;
	}

	/**
	 * Get pair of attribute with name
	 * 
	 * @param key
	 *            Name of attribute
	 * @return Pair of attribute
	 */
	private Pair<String, String> getAttr(String key) {
		Pair<String, String> attr = null;

		if (this.attributes != null) {
			for (Pair<String, String> a : this.attributes) {
				if (a.getFirst().compareTo(key) == 0) {
					attr = a;
				}
			}
		}

		return attr;
	}

	/**
	 * Get attribute with name
	 * 
	 * @param key
	 *            Name of attribute
	 * @return Value of attribute
	 */
	public String getAttribute(String key) {
		String val = null;
		Pair<String, String> attr = null;

		attr = this.getAttr(key);
		if (attr != null) {
			val = attr.getSecond();
		}

		return val;
	}

	/**
	 * Set attribute of name with value
	 * 
	 * @param key
	 *            Name of attribute
	 * @param value
	 *            Value of attribute
	 */
	public void setAttribute(String key, String value) {
		Pair<String, String> attr = null;

		attr = this.getAttr(key);
		if (attr != null) {
			attr.setSecond(value);
		} else {
			if (this.attributes == null) {
				this.attributes = new ArrayList<Pair<String, String>>(1);
			}
			this.attributes.add(new Pair<String, String>(key, value));
		}
	}

	/**
	 * Remove attribute with name
	 * 
	 * @param key
	 *            Name of attribute
	 * @return Value of removed attribute
	 */
	public String rmAttribute(String key) {
		String val = null;

		if (this.attributes != null) {
			for (Pair<String, String> a : this.attributes) {
				if (a.getFirst().compareTo(key) == 0) {
					this.attributes.remove(a);
					val = a.getSecond();
					break;
				}
			}
			if (this.attributes.size() == 0) {
				this.attributes = null;
			}
		}

		return val;
	}

	/**
	 * Clear list of attributes
	 */
	public void clearAttributes() {
		this.attributes = null;
	}

	/**
	 * Get number of nodes
	 * 
	 * @return Number of nodes
	 */
	public int getNodeSize() {
		int size = 0;

		if (this.nodes != null) {
			size = this.nodes.size();
		}

		return size;
	}

	/**
	 * Get array of nodes
	 * 
	 * @return List of nodes
	 */
	public XMLNode[] getNodes() {
		XMLNode[] n = new XMLNode[0];

		if (this.nodes != null) {
			n = this.nodes.toArray(n);
		}

		return n;
	}

	/**
	 * Get node at index
	 * 
	 * @param index
	 *            Index of node
	 * @return Node at index
	 */
	public XMLNode getNode(int index) {
		XMLNode node = null;

		if (this.nodes != null) {
			node = this.nodes.get(index);
		}

		return node;
	}

	/**
	 * Add node to list
	 * 
	 * @param node
	 *            Node to add
	 */
	public void addNode(XMLNode node) {
		if (this.nodes == null) {
			this.nodes = new ArrayList<XMLNode>(1);
		}
		this.nodes.add(node);
	}

	/**
	 * Add node at index to list
	 * 
	 * @param index
	 *            Index of node
	 * @param node
	 *            Node to add
	 */
	public void addNode(int index, XMLNode node) {
		if (this.nodes == null) {
			this.nodes = new ArrayList<XMLNode>(1);
		}
		this.nodes.add(index, node);
	}

	/**
	 * Set node at index (replace node)
	 * 
	 * @param index
	 *            Index of node
	 * @param node
	 *            Node to set
	 */
	public void setNode(int index, XMLNode node) {
		if (this.nodes == null) {
			this.nodes = new ArrayList<XMLNode>(1);
		}
		this.nodes.set(index, node);
	}

	/**
	 * Remove node at index
	 * 
	 * @param index
	 *            Index of node
	 * @return Removed node
	 */
	public XMLNode rmNode(int index) {
		XMLNode node = null;

		if (this.nodes != null) {
			node = this.nodes.remove(index);
			if (this.nodes.size() == 0) {
				this.nodes = null;
			}
		}

		return node;
	}

	/**
	 * Clear list of nodes
	 */
	public void clearNodes() {
		this.nodes = null;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		XMLNode other = (XMLNode) obj;
		if (this.name == null) {
			if (other.name != null) {
				return false;
			}
		} else if (!this.name.equals(other.name)) {
			return false;
		}
		if (this.text == null) {
			if (other.text != null) {
				return false;
			}
		} else if (!this.text.equals(other.text)) {
			if (this.getName().compareTo(other.getName()) != 0) {
				return false;
			}
		}
		if (this.attributes == null) {
			if (other.attributes != null) {
				return false;
			}
		} else {
			if (this.attributes.size() != other.attributes.size()) {
				return false;
			} else {
				for (Pair<String, String> p : this.attributes) {
					if (!other.attributes.contains(p)) {
						return false;
					}
				}
			}
		}
		if (this.nodes == null) {
			if (other.nodes != null) {
				return false;
			}
		} else if (!this.nodes.equals(other.nodes)) {
			return false;
		}
		return true;
	}
}
