import os
from os import walk
from lxml import etree
from abc import ABCMeta, abstractmethod


class AbstractFmi2ModelType(metaclass=ABCMeta):
    def __init__(self, root, model_identifier, src_dir):
        """

        Parameters
        ----------
        root : etree.Element
            xml-root-element
        model_identifier: string
        src_dir : string
            fmu data directory
        """
        self.source_dir = os.path.join(src_dir, "sources")
        self.element = self.getElement(root, model_identifier)
        self.addSourceFiles()

    @abstractmethod
    def getElement(self, root, model_identifier):
        pass

    def getSourceFiles(self):
        """
        Get the source files
        Returns
        -------
        resource_files : list
            List of files in the fmu source directory
        """
        source_files = []
        for (dirpath, dirnames, filenames) in walk(self.source_dir):
            source_files.extend(filenames)
        return source_files

    def addSourceFiles(self):
        """
        Add the <SourceFiles> xml-tag to <ModelExchange>

        XML Schema Description
        ----------------------
        List of source file names that are present in the "sources" directory of the FMU
        and need to be compiled in order to generate the binary of the FMU (only meaningful for source code FMUs).

        File-Tag
        --------
        Name of the file including the path relative to the sources directory, using the forward slash as separator
        (for example:
        name = "myFMU.c";
        name = "modelExchange/solve.c")


        Returns
        -------

        """
        source_files = self.getSourceFiles()
        if not len(source_files) == 0:
            source_file = etree.SubElement(self.element, 'SourceFiles')
            for filename in source_files:
                etree.SubElement(source_file, 'File', name=filename)


class Fmi2ModelExchange(AbstractFmi2ModelType):
    """
    Class for the ModelExchange xml-tag

    XML Schema Description
    ----------------------
    The FMU includes a model or the communication to a tool that provides a model.
    The environment provides the simulation engine for the model.

    modelIdentifier
    ---------------
    Short class name according to C-syntax, e.g. "A_B_C".
    Used as prefix for FMI2 functions if the functions are provided in C source code or in static libraries,
    but not if the functions are provided by a DLL/SharedObject.
    modelIdentifier is also used as name of the static library or DLL/SharedObject.

    needsExecutionTool (Optional)
    -----------------------------
    If true, a tool is needed to execute the model and the FMU just contains the communication to this tool.
    Type: Boolean, Default: false

    completedIntegratorStepNotNeeded (Optional)
    -------------------------------------------
    Type: Boolean, Default: false

    canBeInstantiatedOnlyOncePerProcess (Optional)
    -------------------------------------------
    Type: Boolean, Default: false

    canNotUseMemoryManagementFunctions (Optional)
    ---------------------------------------------
    Type: Boolean, Default: false

    canGetAndSetFMUstate (Optional)
    -------------------------------
    Type: Boolean, Default: false

    canSerializeFMUstate (Optional)
    --------------------------------
    Type: Boolean, Default: false

    providesDirectionalDerivative (Optional)
    ----------------------------------------
    Type: Boolean, Default: false

    """

    def getElement(self, root, model_identifier):
        return etree.SubElement(root, 'ModelExchange',
                                modelIdentifier=model_identifier,
                                # needsExecutionTool="false",
                                completedIntegratorStepNotNeeded="false",
                                # canBeInstantiatedOnlyOncePerProcess="false",
                                # canNotUseMemoryManagementFunctions="false",
                                # canGetAndSetFMUstate="false",
                                # canSerializeFMUstate="false",
                                # providesDirectionalDerivative="false"
                                )


class Fmi2CoSimulation(AbstractFmi2ModelType):
    """
    Class for the ModelExchange xml-tag

    XML Schema Description
    ----------------------
    The FMU includes a model or the communication to a tool that provides a model.
    The environment provides the simulation engine for the model.

    modelIdentifier
    ---------------
    Short class name according to C-syntax, e.g. "A_B_C".
    Used as prefix for FMI2 functions if the functions are provided in C source code or in static libraries,
    but not if the functions are provided by a DLL/SharedObject.
    modelIdentifier is also used as name of the static library or DLL/SharedObject.

    needsExecutionTool (Optional)
    -----------------------------
    If true, a tool is needed to execute the model and the FMU just contains the communication to this tool.
    Type: Boolean, Default: false

    canHandleVariableCommunicationStepSize
    --------------------------------------
    The slave can handle variable communication step size. The communication step size (parameter
    communicationStepSize of fmi2DoStep(...) ) has not to be constant for each call.

    canInterpolateInputs
    --------------------
    The slave is able to interpolate continuous inputs. Calling of
    fmi2SetRealInputDerivatives(...) has an effect for the slave.

    maxOutputDerivativeOrder
    ------------------------
    The slave is able to provide derivatives of outputs with maximum order. Calling of
    fmi2GetRealOutputDerivatives(...) is allowed up to the order defined by
    maxOutputDerivativeOrder.

    canRunAsynchronuously
    ---------------------
    This flag describes the ability to carry out the fmi2DoStep(...) call asynchronously.

    canBeInstantiatedOnlyOncePerProcess (Optional)
    -------------------------------------------
    Type: Boolean, Default: false

    canNotUseMemoryManagementFunctions (Optional)
    ---------------------------------------------
    Type: Boolean, Default: false

    canGetAndSetFMUstate (Optional)
    -------------------------------
    Type: Boolean, Default: false

    canSerializeFMUstate (Optional)
    --------------------------------
    Type: Boolean, Default: false

    providesDirectionalDerivative (Optional)
    ----------------------------------------
    Type: Boolean, Default: false

    """
    def getElement(self, root, model_identifier):
        return etree.SubElement(root, 'CoSimulation',
                                modelIdentifier=model_identifier,
                                # needsExecutionTool="false",
                                # canHandleVariableCommunicationStepSize="false",
                                # canInterpolateInputs="false"
                                # canBeInstantiatedOnlyOncePerProcess="false",
                                # canNotUseMemoryManagementFunctions="false",
                                # canGetAndSetFMUstate="false",
                                # canSerializeFMUstate="false",
                                # providesDirectionalDerivative="false"
                                )