#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 959 $
$LastChangedDate:: 2019-06-18 09:45:53 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import base64
import numpy as np


class Serializable:

    @classmethod
    def toCBOR(cls, obj, f=1, nopack=False):
        """
        Create CBOR serializable instance from an given object

        ref: http://cbor.io/

        Parameters
        ----------
        obj : dict/np.ndarray/pm.PyTPeriodicMatrix/rbf.PeriodicMultiRBF/rbf.StackedMultiRBF
            Serializeable instance
        f : integer
            Angular frequency scaling factor, 1 (Default)
            Applies only if obj is of type pm.PyTPeriodicMatrix or rbf.PeriodicMultiRBF
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false

        Returns
        -------
        cbor_obj : dict/int/float/string/bytes/boolean

            Serialized instance
        """
        import lcm.utils.pytperiodic as pm
        import lcm.math.rbf as rbf

        if isinstance(obj, np.ndarray):
            ''' create cbor serializable dict '''
            cbor_obj = {'type': '__NDRealArray__',
                        'shape': list(obj.shape),
                        'data': obj.astype('>f8', order='C').tobytes()}

        elif isinstance(obj, pm.PyTPeriodicMatrix) or \
                isinstance(obj, rbf.PeriodicMultiRBF):
            '''call class method with frequency scaling'''
            cbor_obj = obj.toCBOR(f, nopack=nopack)
        elif isinstance(obj, rbf.StackedMultiRBF) or isinstance(obj, rbf.RBF):
            '''call class method'''
            cbor_obj = obj.toCBOR()
        elif isinstance(obj, dict):
            cbor_obj = {}
            for key, value in obj.items():
                '''call recursive'''
                cbor_obj[key] = cls.toCBOR(value, f)
        else:
            cbor_obj = obj
        return cbor_obj

    @classmethod
    def fromCBOR(cls, cbor_obj, nopack=False):
        """
        Create a deserialized instance from a CBOR serializabled object

        ref: http://cbor.io/

        Parameters
        ----------
        cbor_obj : dict/int/float/string/bytes/boolean
            Serialized instance
        nopack : boolean, optional
            Keyword argument suppressing packing of signal. default: false
        Returns
        -------
        obj : dict/np.ndarray/pm.PyTPeriodicMatrix/rbf.PeriodicMultiRBF/rbf.StackedMultiRBF
            Serializeable instance

        """
        import lcm.utils.pytperiodic as pm
        import lcm.math.rbf as rbf

        if isinstance(cbor_obj, dict):
            if 'type' in cbor_obj.keys():
                cbor_type = cbor_obj['type']

                if cbor_type == '__NDRealArray__':
                    buffer = cbor_obj['data']
                    shape = cbor_obj['shape']
                    obj = np.ndarray(shape=shape, buffer=buffer, dtype='>f8', order='C')

                elif cbor_type == '__periodicMatrix__':
                    obj = pm.PyTPeriodicMatrix.fromCBOR(cbor_obj, nopack=nopack)

                elif cbor_type == '__periodicRBF__':
                    obj = rbf.PeriodicMultiRBF.fromCBOR(cbor_obj, nopack=nopack)

                elif cbor_type == '__stackedRBF__':
                    obj = rbf.StackedMultiRBF.fromCBOR(cbor_obj)
                else:
                    raise TypeError(str(cbor_type) + 'is not supported!')
            else:
                obj = {}
                for key, value in cbor_obj.items():
                    '''call recursive'''
                    obj[key] = cls.fromCBOR(value, nopack=nopack)

        else:
            obj = cbor_obj
        return obj

    @classmethod
    def saveCBOR(cls, obj, file, f=1):
        """
        Create CBOR file representing an instance of this object

        ref: http://cbor.io/


        Parameters
        ----------
        obj : object
            instance to serialize
        file : String
            path to CBOR file
        f : int
            Angular frequency scaling factor, 1 (Default)
            Applies only if obj is of type pm.PyTPeriodicMatrix or rbf.PeriodicMultiRBF
        """
        import cbor2 as cbor

        cbor_obj = cls.toCBOR(obj, f)

        if not file.endswith('.cbor'):
            file += '.cbor'

        with open(file, "wb") as fh:
            cbor.dump(cbor_obj, fh)

    @staticmethod
    def loadCBOR(file):
        """
        Load a CBOR file and convert it to the corresponding instances

        ref: http://cbor.io/


        Parameters
        ----------
        file : String
            path to CBOR file

        Returns
        -------

        obj : dict/np.ndarray/pm.PyTPeriodicMatrix/rbf.PeriodicMultiRBF/rbf.StackedMultiRBF
            deserialized instance
        """
        import cbor2 as cbor

        path = os.path.abspath(file)

        if path.endswith('.cbor') or os.path.isfile(file + '.cbor'):
            with open(file, "rb") as fh:
                cbor_obj = cbor.load(fh)
        else:
            raise Exception('The filepath does not point to an *.cbor file!')

        return cbor_obj

    @classmethod
    def toBase64CBOR(cls, obj, f=1, decoding='ascii'):
        """
        Serialize CBOR file representing an instance to a string

        Parameters
        ----------

        obj : object
                Any object instance
        f : integer
            Angular frequency scaling factor, 1 (Default)
            Applies only if obj is of type pm.PyTPeriodicMatrix or rbf.PeriodicMultiRBF
        decoding : string
                Decoding style Default('ascii')

        Returns
        -------
        d : dict
            cls : string
                class of the instance
            data : string
                serialized data
            serialization : string
                type of serialization
            encoding : string
                type of encoding
        """

        ''' pickle instance '''
        import cbor2 as cbor

        cbor_obj = cls.toCBOR(obj, f)
        cbor_obj_string = cbor.dumps(cbor_obj)
        ''' encode to base64 '''
        encoded = base64.b64encode(cbor_obj_string)
        ''' decode to string'''
        encoded_cbor_obj = encoded.decode(decoding)

        d = dict(cls=obj.__class__.__name__,
                 encoding='base64',
                 serialization='cbor',
                 data=encoded_cbor_obj)

        return d

    @staticmethod
    def fromBase64(d):
        """
        Deserialize a base64 encoded
        string representation of a CBOR instance
        The serialized instance gets base64 decoded and converted to a serializeable object

        Parameters
        ----------
        d : dict
            cls : string
                class of the instance
            data : string
                serialized data
            serialization : string
                type of serialization
            encoding : string
                type of encoding

        Returns
        -------
        serialization_obj : object/dict
        """
        import cbor2 as cbor

        encoded_cbor_obj = d['data']
        encoding = d['encoding']
        serialization = d['serialization']

        if encoding.lower() == 'base64':
            ''' decode string to byte stream '''
            obj_string = base64.b64decode(encoded_cbor_obj)
        else:
            raise Exception('Encoding is not base64!')

        if serialization.lower() == 'cbor':
            serialization_obj = cbor.loads(obj_string)
        else:
            raise Exception('Serialization is not cbor!')

        return serialization_obj

    @classmethod
    def fromBase64CBOR(cls, d):
        """
        Deserialize a base64 encoded
        string representation of a CBOR instance
        The serialized instance gets base64 decoded and converted to corresponding python classes

        Parameters
        ----------
        d : dict
            cls : string
                class of the instance
            data : string
                serialized data
            serialization : string
                type of serialization
            encoding : string
                type of encoding

        Returns
        -------
        obj : object/dict
        """
        cbor_obj = cls.fromBase64(d)
        obj = cls.fromCBOR(cbor_obj)
        return obj

    @classmethod
    def savePickle(cls, obj, file):
        """
        Save Pickle file representing an instance of this object


        Parameters
        ----------
        obj : object
            instance to serialize
        file : String
            path to Pickle file
        """
        import pickle

        if not (file.endswith('.p') or file.endswith('.pickle')):
            file += '.p'

        print(file)

        with open(file, "wb") as fh:
            pickle.dump(obj, fh)

    @staticmethod
    def loadPickle(file):
        """
        Load a Pickle file and convert it to the corresponding instances


        Parameters
        ----------
        file : String
            path to Pickle file

        Returns
        -------

        obj : object
            deserialized instance
        """
        import pickle

        path = os.path.abspath(file)

        if path.endswith('.p') or os.path.isfile(file + '.p') or \
                path.endswith('.pickle') or os.path.isfile(file + '.pickle'):
            with open(file, "rb") as fh:
                obj = pickle.load(fh)
        else:
            raise Exception('The filepath does not point to an *.p or *.pickle file!')

        return obj

    @classmethod
    def loadFile(cls, file):
        """
        Load content of various file formats
        
        Pickle(.p, .pickle) CBOR (*.cbor) or MAT-files (*.mat)
        
        Parameters
        ----------
        file : str
            path to the file

        Returns
        -------
        obj
        """
        if file.endswith('.pickle') or file.endswith('.p'):
            obj = cls.loadPickle(file)
        elif file.endswith('.mat'):
            import scipy.io
            d = scipy.io.loadmat(file)
            if 'data' in d:
                obj = d['data']
            else:
                d.pop('__header__')
                d.pop('__version__')
                d.pop('__globals__')
                obj = d
            # obj = cls.fromCBOR(obj)
        elif file.endswith('.cbor'):
            obj = cls.loadCBOR(file)
        else:
            raise Exception('File type not supported!')

        return obj
    
    @classmethod
    def saveFile(cls, obj, file):       
        """
        Save data to various file formats
        
        Pickle(.p, .pickle) CBOR (*.cbor) or MAT-files (*.mat)
        
        Parameters
        ----------
        obj
            Data
        file : str
            path to the file
        """

        if file.endswith('.pickle') or file.endswith('.p'):
            cls.savePickle(obj, file)
        elif file.endswith('.mat'):
            import scipy.io
            # obj = cls.toCBOR(obj)
            d = dict(data=obj)
            scipy.io.savemat(file, d)
            print(file)
        elif file.endswith('.cbor'):
            cls.saveCBOR(obj, file)
        else:
            raise Exception('File type not supported!')
