#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1039 $
$LastChangedDate:: 2021-01-28 09:31:14 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import tempfile
import numpy as np
import matplotlib.pyplot as plt
import lcm.utils.pytperiodic as pm
# import bokeh.io as bkio
import plotly.offline as ply


def run_periodicMatrixRandom(nrows=2, ncols=3):
    """
    Example for PyTPeriodicMatrix
    using randomly generated signals with the defined size

    Parameters
    ----------
    nrows : int
        number of rows
    ncols : int
        number of cols

    """

    '''matrix 1'''
    t_per = pm.PyTPeriodicMatrix(nrows, ncols, mode='random', maxorder=100)
    # orders = range(0, 5)
    orders = [1, 3, 5]

    # orders = [0, 2, 4]
    modes = ['matrix', 'row', 'col', 'single']
    # modes = ['single', ]

    '''plot with matplotlib'''
    for mode in modes:
        '''matplot or java test also with CSyMSpace'''
        t_per.show_preview(orders=orders, mode=mode)
        t_per.show_spectrum(orders=orders, mode=mode)
        if mode == 'single':
            t_per.show_full_preview(orders=orders)

        '''matplot'''
        t_per.showPreviewMPL(orders=orders, mode=mode)
        t_per.showSpectrumMPL(orders=orders, mode=mode)
        if mode == 'single':
            t_per.showFullPreviewMPL(orders=orders)

        '''bokeh'''
        t_per.showPreviewBKH(orders=orders, mode=mode)
        # t_per.showSpectrumBKH(orders=orders, mode=mode) # not implemented
        # t_per.showFullPreviewBKH(orders=orders, mode=mode) # not implemented

        '''plotly'''
        t_per.showPreviewPLY(orders=orders, mode=mode)
        t_per.showSpectrumPLY(orders=orders, mode=mode)
        t_per.showFullPreviewPLY(orders=orders, mode=mode)


def run_periodicFit():
    """
    Example for PyTPeriodicMatrix
    using randomly generated signals with the defined size
    """

    '''matrix 1'''
    phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)
    order = 100
    samples = np.sin(order * phi)
    t_per = pm.PyTPeriodicMatrix.fromFit(phi, samples, maxorder=2)
    max_order = t_per.get_max_order()
    print('max_order', max_order)


def run_periodicMatrixSine(nrows=2, ncols=3):
    """
    Example for PyTPeriodicMatrix
    using cosine and sine signals with the defined size

    Parameters
    ----------
    nrows : int
        number of rows
    ncols : int
        number of cols
    """

    '''input matrix from 1x1 periodics'''
    xc = np.zeros((nrows, ncols), dtype=np.object)
    for r in range(nrows):
        for c in range(ncols):
            magnitude = r + c
            xc[r, c] = pm.PyTPeriodicMatrix(1., magnitude, 0.)  # cosine

    '''output matrix from 1x1 periodics'''
    zc = np.zeros((nrows, ncols), dtype=np.object)
    for r in range(nrows):
        for c in range(ncols):
            magnitude = r + c
            zc[r, c] = pm.PyTPeriodicMatrix(1, magnitude, -np.pi / 2)  # sine

    '''numpy to periodic'''
    xc_per = pm.PyTPeriodicMatrix(xc, mode='objarray')
    zc_per = pm.PyTPeriodicMatrix(zc, mode='objarray')

    '''print'''
    print('xc_per', xc_per)
    print('zc_per', zc_per)

    '''plot with plotly'''
    xc_per.showFullPreviewPLY(mode='matrix')
    zc_per.showFullPreviewPLY(mode='matrix')


def run_periodicMatrixPlot(nrows=2, ncols=3):
    """
    Example for PyTPeriodicMatrix
    using randomly generated signals with the defined size

    Parameters
    ----------
    nrows : int
        number of rows
    ncols : int
        number of cols

    """

    '''matrix 1'''
    xc_per = pm.PyTPeriodicMatrix(nrows, ncols, mode='random', maxorder=10)

    '''matrix 2'''
    zc_per = pm.PyTPeriodicMatrix(nrows, ncols, mode='random', maxorder=10)

    '''print'''
    print('xc_per', xc_per)
    print('zc_per', zc_per)

    '''plot with plotly'''
    # xc_per.showPreviewPLY(mode='matrix', titles=[str(x) for x in range(nrows * ncols)])
    # zc_per.showPreviewPLY(mode='single', titles=['abc', ])
    zc_per.showFullPreviewPLY(mode='single', titles=['abc', ], orders=np.arange(0, 30, 1))
    # zc_per.showPreviewPLY(mode='single', titles=['a', ])
    # zc_per.showPreviewPLY(mode='row', titles=[str(x) for x in range(nrows+1)])


if __name__ == '__main__':
    # run_periodicFit()
    run_periodicMatrixRandom(2, 3)
    # run_periodicMatrixRandom(1, 1)
    # run_periodicMatrixSine()
    # run_periodicMatrixPlot()
