# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 1039 $
$LastChangedDate:: 2021-01-28 09:31:14 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import unittest
import numpy as np
import tempfile
import numpy.testing as npt
import scipy.signal

from lcm.utils import pytperiodic as pm


class TestPyTPeriodicMatrix(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Sets up unittest data

        """
        ''' Create a temporary directory'''
        cls.test_dir = tempfile.mkdtemp()

        '''Create test data'''
        cls.npts = 100
        cls.phi = np.linspace(0, 2 * np.pi, cls.npts, endpoint=False)

        samples1 = np.empty((2, 3, cls.npts), dtype=np.float)
        samples1[0, 0] = np.zeros_like(cls.phi)
        samples1[0, 1] = 1.5 * np.ones_like(cls.phi)
        samples1[0, 2] = 1.5 + 2 * np.cos(cls.phi)
        samples1[1, 0] = 1.5 + 2 * np.cos(cls.phi) - 0.5 * np.sin(2 * cls.phi)
        samples1[1, 1] = 2 * np.cos(cls.phi) - 0.5 * np.sin(3 * cls.phi)
        samples1[1, 2] = 1.5 + 2 * np.cos(cls.phi) + 1e-6 * np.cos(7 * cls.phi)

        samples2 = np.empty((2, 3, cls.npts), dtype=np.float)
        samples2[0, 0] = np.zeros_like(cls.phi)
        samples2[0, 1] = 1.5 * np.ones_like(cls.phi)
        samples2[0, 2] = 1.5 + 2 * np.cos(cls.phi)
        samples2[1, 0] = 1.5 + 2 * np.cos(cls.phi) - 0.5 * np.sin(2 * cls.phi)
        samples2[1, 1] = 2 * np.cos(cls.phi) - 0.5 * np.sin(3 * cls.phi)
        samples2[1, 2] = 1.5 + 2 * np.cos(cls.phi) + 1e-6 * np.cos(7 * cls.phi)

        samples3 = np.empty((2, 3, cls.npts), dtype=np.float)
        samples3[0, 0] = 1 * np.cos(cls.phi)
        samples3[0, 1] = 2 * np.cos(cls.phi)
        samples3[0, 2] = 3 * np.cos(cls.phi)
        samples3[1, 0] = 4 * np.cos(cls.phi)
        samples3[1, 1] = 5 * np.cos(cls.phi)
        samples3[1, 2] = 6 * np.cos(cls.phi)

        samples4 = np.empty((2, 3, cls.npts), dtype=np.float)
        samples4[0, 0] = 6 * np.sin(cls.phi)
        samples4[0, 1] = 5 * np.sin(cls.phi)
        samples4[0, 2] = 4 * np.sin(cls.phi)
        samples4[1, 0] = 3 * np.sin(cls.phi)
        samples4[1, 1] = 2 * np.sin(cls.phi)
        samples4[1, 2] = 1 * np.sin(cls.phi)

        samples5 = np.empty((3, 3, cls.npts), dtype=np.float)
        samples5[0, 0] = np.cos(cls.phi)
        samples5[1, 0] = np.sin(cls.phi)
        samples5[2, 0] = 0
        samples5[0, 1] = - np.sin(cls.phi)
        samples5[1, 1] = np.cos(cls.phi)
        samples5[2, 1] = 0
        samples5[0, 2] = 0
        samples5[1, 2] = 0
        samples5[2, 2] = 1

        samples6 = np.empty((3, 3, cls.npts), dtype=np.float)
        samples6[0, 0] = np.cos(cls.phi)
        samples6[1, 0] = np.cos(cls.phi)
        samples6[2, 0] = 0
        samples6[0, 1] = - np.sin(cls.phi)
        samples6[1, 1] = np.sin(cls.phi)
        samples6[2, 1] = 0
        samples6[0, 2] = 0
        samples6[1, 2] = 0
        samples6[2, 2] = 1

        cls.rfft_vec = np.array([[[8.59242403e+02, 0.00000000e+00, - 0.00000000e+00, 0.00000000e+00,
                                   - 0.00000000e+00, 5.62690174e-02, - 8.67590066e-01, 0.00000000e+00,
                                   - 0.00000000e+00, 0.00000000e+00, - 0.00000000e+00, - 7.96651737e-01,
                                   8.39613013e-02, 0.00000000e+00, - 0.00000000e+00, 0.00000000e+00,
                                   - 0.00000000e+00, - 5.61147521e-01, 3.72622903e-02, 0.00000000e+00,
                                   - 0.00000000e+00, 0.00000000e+00, - 0.00000000e+00, - 1.18814570e+01,
                                   - 4.18951166e+01, 0.00000000e+00, - 0.00000000e+00, 0.00000000e+00,
                                   - 0.00000000e+00, - 6.09257085e-01, 2.15948499e-01, 0.00000000e+00,
                                   - 0.00000000e+00, 0.00000000e+00, - 0.00000000e+00, - 8.46854842e-01,
                                   4.14033568e-01, 0.00000000e+00, - 0.00000000e+00, 0.00000000e+00,
                                   - 0.00000000e+00, - 1.80240111e-02, 1.81582049e-01]]], dtype=np.float64)

        cls.samples1 = pm.PyTPeriodicMatrix(samples1, mode='samples')
        cls.samples2 = pm.PyTPeriodicMatrix(samples2, mode='samples')
        cls.samples3 = pm.PyTPeriodicMatrix(samples3, mode='samples')
        cls.samples4 = pm.PyTPeriodicMatrix(samples4, mode='samples')
        cls.samples5 = pm.PyTPeriodicMatrix(samples5, mode='samples')
        cls.samples6 = pm.PyTPeriodicMatrix(samples6, mode='samples')

        cls.eye = pm.PyTPeriodicMatrix(2, 3, mode='eye')
        cls.eye2x2 = pm.PyTPeriodicMatrix(2, 2, mode='eye')
        cls.zero = pm.PyTPeriodicMatrix(2, 3, mode='zero')
        cls.random = pm.PyTPeriodicMatrix(2, 3, mode='random', maxorder=10)
        cls.cos = pm.PyTPeriodicMatrix(1, 1., 0.)
        cls.sin = pm.PyTPeriodicMatrix(1, 1., -np.pi / 2)
        cls.dc = pm.PyTPeriodicMatrix(np.array([[1, 2], [3, 4]]), mode='dc')
        cls.rfft = pm.PyTPeriodicMatrix(cls.rfft_vec, mode='rfft', nopack=True)

        phi = 4. * np.pi * np.random.rand(4 * cls.npts)
        values = 2 * np.cos(phi) - 0.5 * np.sin(3 * phi)
        values2D = np.array([values, values])

        cls.fit = pm.PyTPeriodicMatrix(phi, values, mode="fit")
        cls.fit2D = pm.PyTPeriodicMatrix(phi, values2D, mode="fit")

        array = np.empty((2, 2), dtype=np.object)
        array[0, 0] = cls.sin
        array[1, 0] = cls.cos
        array[0, 1] = cls.sin
        array[1, 1] = cls.cos

        cls.objarray = pm.PyTPeriodicMatrix(array, mode='objarray')

        cls.ref_tps = dict(samples1=cls.samples1,
                           samples2=cls.samples2,
                           samples3=cls.samples3,
                           samples4=cls.samples4,
                           samples5=cls.samples5,
                           samples6=cls.samples6,
                           eye=cls.eye,
                           eye2x2=cls.eye2x2,
                           zero=cls.zero,
                           random=cls.random,
                           cos=cls.cos,
                           sin=cls.sin,
                           dc=cls.dc,
                           rfft=cls.rfft,
                           fit=cls.fit,
                           fit2D=cls.fit2D,
                           objarray=cls.objarray)

        cls.ref_scaled_tps = {}
        for key, value in cls.ref_tps.items():
            new_key = key + '_scaled'
            tmp_per: pm.PyTPeriodicMatrix = value[:, :]
            tmp_per.scaling_factor = 2
            cls.ref_scaled_tps[new_key] = tmp_per

    def test_fromCBOR(self):

        for key, tp_ref in self.ref_tps.items():
            for f in [1, 2, 6]:
                with self.subTest(msg='{}-f={}'.format(key, f)):
                    cbor_obj = tp_ref.toCBOR(f)

                    obj = pm.PyTPeriodicMatrix.fromCBOR(cbor_obj)

                    scaled_ref = tp_ref.transform_order(1, f)
                    scaled_obj = obj.transform_order(1, f)

                    npt.assert_almost_equal(scaled_ref.data, scaled_obj.data)

    def test_fromFit(self):
        phi = np.linspace(0, 2 * np.pi, self.npts, endpoint=False)
        # d = {'fit': self.fit, 'fit2D': self.fit2D}
        d = {'fit': self.fit}
        for key, tp_ref in d.items():
            with self.subTest(msg=key):
                values = tp_ref.get_samples(self.npts).flatten()
                fit = pm.PyTPeriodicMatrix.fromFit(phi, values)

                npt.assert_almost_equal(tp_ref.data, fit.data)

    def test_fromObjectArray(self):
        array = np.empty((2, 2), dtype=np.object)
        array[0, 0] = self.objarray[0, 0]
        array[1, 0] = self.objarray[1, 0]
        array[0, 1] = self.objarray[0, 1]
        array[1, 1] = self.objarray[1, 1]
        objarray = pm.PyTPeriodicMatrix.fromObjectArray(array)

        npt.assert_almost_equal(self.objarray.data, objarray.data)

    def test_fromSamples(self):
        n_samples = 360
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['rfft']:
                    self.skipTest('subtest {} skipped'.format(key))

                values = tp_ref.get_samples(n_samples)
                tmp_value = pm.PyTPeriodicMatrix.fromSamples(values, nopack=tp_ref.nopack)

                npt.assert_almost_equal(tp_ref.data, tmp_value.data)

    def test_fromRFFT(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                rfft = tp_ref.data
                tmp_per = pm.PyTPeriodicMatrix.fromRFFT(rfft, nopack=tp_ref.nopack)

                npt.assert_almost_equal(tp_ref.data, tmp_per.data)

    def test_fromDC(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                dc, cos_coeffs, sin_coeffs = tp_ref.get_coeffs()
                tmp_per = pm.PyTPeriodicMatrix.fromDC(dc)
                dc_tmp, cos_coeffs_tmp, sin_coeffs_tmp = tmp_per.get_coeffs()

                npt.assert_almost_equal(dc, dc_tmp)

    def test_fromCoeffs(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                dc, cos_coeffs, sin_coeffs = tp_ref.get_coeffs()
                tmp_per = pm.PyTPeriodicMatrix.fromCoeffs(dc, cos_coeffs, sin_coeffs, nopack=tp_ref.nopack)

                npt.assert_almost_equal(tp_ref.data, tmp_per.data)

    def test_zeros(self):
        np_zero = np.zeros(self.zero.shape)
        zero = pm.PyTPeriodicMatrix.zeros(self.zero.shape)

        npt.assert_almost_equal(self.zero.data, zero.data)
        npt.assert_almost_equal(np_zero, zero.data[:, :, 0])

    def test_eye(self):
        np_eye = np.eye(self.eye.nrows(), self.eye.ncols())
        eye = pm.PyTPeriodicMatrix.eye(self.eye.shape)

        npt.assert_almost_equal(self.eye.data, eye.data)
        npt.assert_almost_equal(np_eye, eye.data[:, :, 0])

    def test_random(self):
        shape = (2, 3)
        random = pm.PyTPeriodicMatrix.random(shape, maxorder=10)

        self.assertListEqual(list(random.shape), list(shape))

    def test_init(self):
        with self.subTest(msg='Exeption-ModeMissing'):
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(1)
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(1, 1)

        with self.subTest(msg='Exeption-UnkownMode'):
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(1, mode='foobar')
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(1, 1, mode='foobar')

        with self.subTest(msg='Exeption-WrongDimension'):
            tmp = np.zeros((1, 2, 3, 4))
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(tmp, mode='rfft')

        with self.subTest(msg='Exeption-Objarray-WrongType'):
            tmp_list = [[], []]
            tmp_list[0] = [self.objarray[0, 0], self.objarray[0, 1]]
            tmp_list[1] = [self.objarray[1, 0], self.objarray[1, 1]]

            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(tmp_list, mode='objarray')

        with self.subTest(msg='Exeption-Objarray-WrongDType'):
            tmp_array = self.objarray[:, :]
            tmp_array.dtype = np.float
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(tmp_array, mode='objarray')

        with self.subTest(msg='Exeption-Objarray-WrongDim'):
            tmp_array = np.empty((2, 2), dtype=np.object)
            tmp_array[0, 0] = self.objarray[0, 0]
            tmp_array[1, 0] = pm.PyTPeriodicMatrix(2, 2, mode='random')
            tmp_array[0, 1] = self.objarray[0, 1]
            tmp_array[1, 1] = self.objarray[1, 1]

            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix(tmp_array, mode='objarray')

    def test_data(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                self.assertIsInstance(tp_ref.data, np.ndarray)
                self.assertEqual(tp_ref.data.dtype, np.float)
                self.assertEqual(tp_ref.data.shape[0], tp_ref.nrows())
                self.assertEqual(tp_ref.data.shape[1], tp_ref.ncols())
                self.assertEqual(tp_ref.data.shape[2], tp_ref.nbins())

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key + '_setData'):
                tmp = pm.PyTPeriodicMatrix()
                tmp.data = tp_ref.data
                npt.assert_almost_equal(tmp.data, tp_ref.data)

    def test_shape(self):
        d = {'random 1': dict(mode='random', shape=(1, 1)),
             'random 2': dict(mode='random', shape=(3, 5)),
             'random 3': dict(mode='random', shape=(5, 3)),
             'random 4': dict(mode='random', shape=(30, 20)),
             'eye 1': dict(mode='eye', shape=(1, 1)),
             'eye 2': dict(mode='eye', shape=(3, 5)),
             'eye 3': dict(mode='eye', shape=(5, 3)),
             'eye 4': dict(mode='eye', shape=(30, 20)),
             'zero 1': dict(mode='zero', shape=(1, 1)),
             'zero 2': dict(mode='zero', shape=(3, 5)),
             'zero 3': dict(mode='zero', shape=(5, 3)),
             'zero 4': dict(mode='zero', shape=(30, 20)),

             }

        for key, data in d.items():
            with self.subTest(msg=key):
                tp = pm.PyTPeriodicMatrix(data['shape'][0], data['shape'][1], mode=data['mode'])

                npt.assert_almost_equal(tp.shape, data['shape'])

    def test_reshape(self):
        n = 1000
        tp = pm.PyTPeriodicMatrix(4, 3, mode='random')
        npt.assert_almost_equal(tp.shape, (4, 3))

        tp_reshaped = tp.reshape(3, 4)
        npt.assert_almost_equal(tp_reshaped.shape, (3, 4))

        tp = pm.PyTPeriodicMatrix(1, 10, mode='random')
        tp_samples = tp.get_samples(n)

        tp_reshaped = tp.reshape(2, 5)
        tp_reshaped_samples = tp_reshaped.get_samples(n)

        npt.assert_almost_equal(tp_samples[0, 5, :], tp_reshaped_samples[1, 0, :])
        npt.assert_almost_equal(tp_samples[0, 6, :], tp_reshaped_samples[1, 1, :])
        npt.assert_almost_equal(tp_samples[0, 7, :], tp_reshaped_samples[1, 2, :])
        npt.assert_almost_equal(tp_samples[0, 8, :], tp_reshaped_samples[1, 3, :])

        tp = pm.PyTPeriodicMatrix(3, 3, mode='random')
        npt.assert_almost_equal(tp.shape, (3, 3))

        tp_samples = tp.get_samples(n)

        tp_reshaped = tp.reshape(1, 9)
        npt.assert_almost_equal(tp_reshaped.shape, (1, 9))
        tp_reshaped_samples = tp_reshaped.get_samples(n)

        npt.assert_almost_equal(tp_samples[0, 0, :], tp_reshaped_samples[0, 0, :])
        npt.assert_almost_equal(tp_samples[0, 2, :], tp_reshaped_samples[0, 2, :])
        npt.assert_almost_equal(tp_samples[1, 0, :], tp_reshaped_samples[0, 3, :])
        npt.assert_almost_equal(tp_samples[1, 1, :], tp_reshaped_samples[0, 4, :])
        npt.assert_almost_equal(tp_samples[1, 2, :], tp_reshaped_samples[0, 5, :])
        npt.assert_almost_equal(tp_samples[2, 0, :], tp_reshaped_samples[0, 6, :])
        npt.assert_almost_equal(tp_samples[2, 1, :], tp_reshaped_samples[0, 7, :])
        npt.assert_almost_equal(tp_samples[2, 2, :], tp_reshaped_samples[0, 8, :])

    def test_nrows(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                self.assertEqual(tp_ref.data.shape[0], tp_ref.nrows())

    def test_ncols(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                self.assertEqual(tp_ref.data.shape[1], tp_ref.ncols())

    def test_nbins(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                self.assertEqual(tp_ref.data.shape[2], tp_ref.nbins())

    def test_min(self):
        d = {'sin-1': dict(tp=self.sin, min=-1.0),
             'cos-1': dict(tp=self.cos, min=-1.0),
             'zero-1': dict(tp=self.zero, min=0.0),
             'samples1-1': dict(tp=self.samples1, min=np.array([[0, 1.5, -0.5], [-0.701834737, -2.3127730621, -0.5]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                data['tp'].resolution = 100

                npt.assert_almost_equal(data['tp'].min(), data['min'], decimal=3)

                data['tp'].resolution = 10

    def test_max(self):
        d = {'sin-1': dict(tp=self.sin, max=1.0),
             'cos-1': dict(tp=self.cos, max=1.0),
             'zero-1': dict(tp=self.zero, max=0.0),
             'samples1-1': dict(tp=self.samples1, max=np.array([[0, 1.5, 3.5], [3.7018347375, 2.3127730621, 3.5]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                data['tp'].resolution = 100

                npt.assert_almost_equal(data['tp'].max(), data['max'], decimal=3)

                data['tp'].resolution = 10

    def test_pp(self):
        d = {'sin-1': dict(tp=self.sin, pp=2.0),
             'cos-1': dict(tp=self.cos, pp=2.0),
             'zero-1': dict(tp=self.zero, pp=0.0),
             'sin**2-1': dict(tp=self.sin * self.sin, pp=1.0),
             'cos**2-1': dict(tp=self.cos * self.cos, pp=1.0),
             'cos*sin-1': dict(tp=self.cos * self.sin, pp=1.0),
             'samples1-1': dict(tp=self.samples1, pp=np.array([[0., 0., 4.], [4.402878, 4.625535, 4.]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                data['tp'].resolution = 100

                npt.assert_almost_equal(data['tp'].pp(), data['pp'], decimal=3)

                data['tp'].resolution = 10

    def test_mean(self):
        d = {'sin-1': dict(tp=self.sin, mean=0.0),
             'cos-1': dict(tp=self.cos, mean=0.0),
             'zero-1': dict(tp=self.zero, mean=0.0),
             'sin**2-1': dict(tp=self.sin * self.sin, mean=1 / 2),
             'cos**2-1': dict(tp=self.cos * self.cos, mean=1 / 2),
             'cos*sin-1': dict(tp=self.cos * self.sin, mean=0.0),
             'samples1-1': dict(tp=self.samples1, mean=np.array([[0, 1.5, 1.5], [1.5, 0, 1.5]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                data['tp'].resolution = 1000000

                npt.assert_almost_equal(data['tp'].mean(), data['mean'], decimal=3)

                data['tp'].resolution = 10

    def test_rms(self):
        d = {'sin-1': dict(tp=self.sin, rms=1 / np.sqrt(2)),
             'cos-1': dict(tp=self.cos, rms=1 / np.sqrt(2)),
             'zero-1': dict(tp=self.zero, rms=0.0),
             'samples1-1': dict(tp=self.samples1, rms=np.array([[0, 1.5, 2.061552812],
                                                                [2.091650066, 1.45773797, 2.061552812]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(data['tp'].rms(), data['rms'])

    def test_thd(self):
        d = {'samples1-1': dict(tp=self.samples1, thd=np.array([[np.nan, np.nan, 0], [0.25, 0.25, 0]])),
             'samples4-1': dict(tp=self.samples4, thd=np.array([[0, 0, 0], [0, 0, 0]])),
             'sin-1': dict(tp=self.sin, thd=0.0),
             'cos-1': dict(tp=self.cos, thd=0.0),
             'zero-1': dict(tp=self.zero, thd=np.array([[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]))
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(data['tp'].thd(), data['thd'])

    def test_get_samples(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                samples = tp_ref.get_samples(self.npts)

                self.assertIsInstance(samples, np.ndarray)
                self.assertEqual(samples.dtype, np.float)
                self.assertEqual(samples.shape[0], tp_ref.nrows())
                self.assertEqual(samples.shape[1], tp_ref.ncols())
                self.assertEqual(samples.shape[2], self.npts)

    def test_value_at(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(tp_ref.get_samples(self.npts), tp_ref.value_at(self.phi))

    def test_value_at_pos(self):
        index = self.phi.size // 2
        phi = self.phi[index]

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                for row in range(tp_ref.nrows()):
                    for col in range(tp_ref.ncols()):
                        npt.assert_almost_equal(tp_ref.get_samples(self.npts)[row, col, index],
                                                tp_ref.value_at_pos(phi, row, col))

    def test_get_fft(self):
        d = {'sin-1': dict(tp=self.sin, fft=np.array([[[0. + 0.j, 0. - 1.5j, 0. + 1.5j]]]), dtype=np.complex),
             'cos-1': dict(tp=self.cos, fft=np.array([[[0. + 0.j, 1.5 + 0.j, 1.5 + 0.j]]]), dtype=np.complex),
             'zero-1': dict(tp=self.zero, fft=np.array([[[0. + 0.j], [0. + 0.j], [0. + 0.j]],
                                                        [[0. + 0.j], [0. + 0.j], [0. + 0.j]]], dtype=np.complex)),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(data['tp'].get_fft(), data['fft'])

    def test_get_magnitudes(self):

        d = {'sin-1': dict(tp=self.sin, mag=1),
             'cos-1': dict(tp=self.cos, mag=1),
             'zero-1': dict(tp=self.zero, mag=0.0),
             'samples1-1': dict(tp=self.samples1, mag=np.array([[[0., 0., 0.], [0., 0., 0.], [2., 0., 0.]],
                                                                [[2., 0.5, 0.], [2., 0., 0.5], [2., 0., 0.]]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(data['tp'].get_magnitudes(), data['mag'])

    def test_get_phases(self):
        d = {'sin-1': dict(tp=self.sin, phase=-np.pi / 2),
             'cos-1': dict(tp=self.cos, phase=0.0),
             'zero-1': dict(tp=self.zero, phase=0.0),
             'one-1': dict(tp=self.zero, phase=0.0),
             'samples1-1': dict(tp=self.samples1, phase=np.array([[[0., 0., 0., ],
                                                                   [0., 0., 0., ],
                                                                   [0., np.pi, np.pi]],

                                                                  [[0., np.pi / 2, np.pi],
                                                                   [0., 0., np.pi / 2],
                                                                   [0., np.pi, np.pi]]])),
             'samples2-1': dict(tp=self.samples2, phase=np.array([[[0., 0., 0., ],
                                                                   [0., 0., 0., ],
                                                                   [0., np.pi, np.pi]],

                                                                  [[0., np.pi / 2, np.pi],
                                                                   [0., 0., np.pi / 2],
                                                                   [0., np.pi, np.pi]]])),
             'samples3-1': dict(tp=self.samples3, phase=np.array([[[0.], [0.], [0.]], [[0.], [0.], [0.]]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                npt.assert_almost_equal(data['tp'].get_phases(), data['phase'])

    # @unittest.skip
    def test_get_max_order(self):
        phi = np.linspace(0, 2 * np.pi, 1000, endpoint=False)

        for order in range(2, 100, 5):
            with self.subTest(msg='fit-order={:d}'.format(order)):
                samples = np.sin(order * phi)
                t_per = pm.PyTPeriodicMatrix.fromFit(phi, samples, maxorder=None)
                max_order = t_per.get_max_order()
                npt.assert_equal(max_order, order)
            with self.subTest(msg='random-order={:d}'.format(order)):
                t_per = pm.PyTPeriodicMatrix.random((1, 1), maxorder=order)
                max_order = t_per.get_max_order()
                npt.assert_equal(max_order, order)

    @unittest.skip
    def test_get_coeffs(self):
        self.fail()

    def test_select(self):
        data_samples1_2 = np.array([[[0., 0., 0.],
                                     [4.5, 0., 0.],
                                     [4.5, 3., -0.]],
                                    [[4.5, 3., -0.],
                                     [-0., 3., -0.],
                                     [4.5, 3., -0.]]])
        data_samples1_3 = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                                     [10.5, 0., 0., 0., 0., 0., 0.],
                                     [10.5, 0, -0., 0., 0., 0., 0.]],
                                    [[10.5, 0, -0., 0., 1.75, 0., 0.],
                                     [-0., 0, -0., -0., -0., 0., 1.75],
                                     [10.5, 0, -0., 0., 0., 0., 0.]]])

        d = {'samples1-1': dict(tp=self.samples1, select=[1, 2, 3], data=self.samples1.data),
             'samples1-2': dict(tp=self.samples1, select=[1], data=data_samples1_2),
             'samples1-3': dict(tp=self.samples1, select=[2, 3], data=data_samples1_3),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                tp_sel = data['tp'].select(data['select'])

                npt.assert_almost_equal(tp_sel.data, data['data'])

    @unittest.skip
    def test_pack(self):
        self.fail()

    @unittest.skip
    def test_expand_to_odd(self):
        self.fail()

    def test_transform_order(self):

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['random', 'rfft']:
                    self.skipTest('subtest {} skipped'.format(key))

                # select third harmonic by dividing and multiplying
                tp_new = tp_ref.transform_order(1, 3, nopack=tp_ref.nopack)
                tp_new = tp_new.transform_order(3, 1, nopack=tp_ref.nopack)
                (dc1, cos_coeffs1, sin_coeffs1) = tp_new.get_coeffs()

                # selct third harmonic by select method
                (dc2, cos_coeffs2, sin_coeffs2) = tp_ref.select([3], nopack=tp_ref.nopack).get_coeffs()

                npt.assert_almost_equal(dc1, dc2)
                npt.assert_almost_equal(cos_coeffs1, cos_coeffs2)
                npt.assert_almost_equal(sin_coeffs1, sin_coeffs2)

    @unittest.skip
    def test_remove_dc(self):
        self.fail()

    @unittest.skip
    def test_remove_ac(self):
        self.fail()

    def test_det(self):
        d = {'eye2x2-1': dict(tp=self.eye2x2, det=1.0),
             'samples5-1': dict(tp=self.samples5, det=1.0),
             'samples6-1': dict(tp=self.samples6, det=2 * np.sin(self.phi) * np.cos(self.phi)),
             'dc-1': dict(tp=self.dc, det=-2)
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                tp_det = data['tp'].det()  # type: pm.PyTPeriodicMatrix
                det_vals = tp_det.get_samples(self.npts)
                npt.assert_almost_equal(det_vals.flatten(), data['det'] * np.ones(self.npts))

    def test_inv(self):
        I = np.ones(self.npts)
        O = np.zeros(self.npts)
        S = np.sin(self.phi)
        C = np.cos(self.phi)
        # CSC = 1/S
        # SEC = 1/C

        d = {
            'eye2x2-1': dict(tp=self.eye2x2, inv=np.array([[I, O],
                                                           [O, I]])),
            'samples5-1': dict(tp=self.samples5, inv=np.array([[C, S, O],
                                                               [-S, C, O],
                                                               [O, O, I]])),
            # 'samples6-1': dict(tp=self.samples6, inv=np.array([[SEC, SEC, O],
            #                                                    [-CSC, CSC, O],
            #                                                    [O, O, 2*I]])),
            'dc-1': dict(tp=self.dc, inv=np.array([[-2. * I, 1. * I],
                                                   [1.5 * I, -0.5 * I]]))
        }

        for key, data in d.items():
            with self.subTest(msg=key):
                tp_inv = data['tp'].inv()  # type: pm.PyTPeriodicMatrix
                inv_vals = tp_inv.get_samples(self.npts)
                npt.assert_almost_equal(inv_vals, data['inv'])

    def test_find_zerocrossing(self):
        cx_samples1 = np.zeros((2, 3, 2))
        cx_samples1[0, 2] = np.array([2.418858406, 3.864326901])
        cx_samples1[1, 0] = np.array([2.75283097, 4.1581980])
        cx_samples1[1, 1] = np.array([1.776239767, 4.917832421])
        cx_samples1[1, 2] = np.array([2.4188584057, 3.864326901])

        d = {'sin-all-1':
                 dict(tp=self.sin,
                      mode='all',
                      crossings=np.array([np.pi, 2 * np.pi]),
                      shift=0.0),
             'cos-all-1':
                 dict(tp=self.cos,
                      mode='all',
                      crossings=np.array([np.pi / 2, 3 * np.pi / 2]),
                      shift=0.0),
             'sin-all-2':
                 dict(tp=self.sin,
                      mode='all',
                      crossings=np.array([np.pi, 2 * np.pi]),
                      shift=np.pi / 100),
             'cos-all-2':
                 dict(tp=self.cos, mode='all',
                      crossings=np.array([np.pi / 2, 3 * np.pi / 2]),
                      shift=np.pi / 100),
             'x-all-1':
                 dict(tp=pm.PyTPeriodicMatrix(5, 5, 0),
                      mode='all',
                      crossings=np.array([np.pi / 2 / 5, 3 * np.pi / 2 / 5, 5 * np.pi / 2 / 5, 7 * np.pi / 2 / 5,
                                          9 * np.pi / 2 / 5, 11 * np.pi / 2 / 5, 13 * np.pi / 2 / 5,
                                          15 * np.pi / 2 / 5, 17 * np.pi / 2 / 5, 19 * np.pi / 2 / 5]),
                      shift=0.0),
             'samples1-all-1':
                 dict(tp=self.samples1, mode='all',
                      crossings=cx_samples1,
                      shift=0.0,
                      item=[[0, 2, None], [1, 0, None], [1, 1, None], [1, 2, None]]),
             'sin-guess-1':
                 dict(tp=self.sin,
                      mode='guess',
                      crossings=np.array([0.0]),
                      shift=0.0,
                      item=[[0, 0, 0.1], ]),
             'sin-guess-2':
                 dict(tp=self.sin,
                      mode='guess',
                      crossings=np.array([np.pi]),
                      shift=0.0,
                      item=[[0, 0, np.pi - 0.4], ]),
             'samples1-guess-1':
                 dict(tp=self.samples1,
                      mode='guess',
                      crossings=cx_samples1,
                      shift=0.0,
                      item=[[0, 2, 1.0], [1, 0, 2.0], [1, 1, 4.0], [1, 2, 2.35]]),

             }
        for key, values in d.items():
            with self.subTest(msg=key):
                if values['crossings'].ndim != 3:
                    values['crossings'] = values['crossings'][np.newaxis, np.newaxis, :]
                if 'item' not in values:
                    values['item'] = [[0, 0, None], ]

                for row, col, guess in values['item']:
                    crossings = values['tp'].shift(values['shift']).find_zerocrossing(row, col,
                                                                                      mode=values['mode'],
                                                                                      guess=guess)
                    if values['mode'] == 'all':
                        npt.assert_almost_equal(crossings, values['crossings'][row, col] - values['shift'])
                    else:
                        npt.assert_almost_equal(crossings, values['crossings'][row, col, 0] - values['shift'])

    def test_flip(self):
        d = {'samples1-1': dict(tp=self.samples1, angle=np.pi),
             'samples1-2': dict(tp=self.samples1, angle=np.pi / 7),
             'samples1-3': dict(tp=self.samples1, angle=np.pi / 34)}

        for key, data in d.items():
            with self.subTest(msg=key):
                flipped = data['tp'].flip()
                npt.assert_almost_equal(data['tp'].value_at(data['angle']), flipped.value_at(2 * np.pi - data['angle']))

    def test_shift(self):
        """Test shift function"""
        a = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                       [10.5, 0., 0., 0., 0., 0., 0.],
                       [10.5, 7., -0., 0., 0., 0., 0.]],
                      [[10.5, 7., -0., 0., 1.75, 0., 0.],
                       [-0., 7., -0., -0., -0., 0., 1.75],
                       [10.5, 7., -0., 0., 0., 0., 0.]]])
        angles = {'pi/6': np.pi / 6,
                  '2pi': 2 * np.pi}
        for key, angle in angles.items():
            with self.subTest(msg=str(angle)):

                shifted = self.samples1.shift(-angle)
                backshifted = shifted.shift(angle)

                npt.assert_almost_equal(backshifted.data, a)

                if key == '2pi':
                    npt.assert_almost_equal(shifted.data, a)

    def test_deriv(self):
        """ Test derivation method"""
        der = self.samples1.deriv(5.)

        a_der = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                           [0, 0., 0., 0., 0., 0., 0.],
                           [0, 0., 35., 0., 0., 0., 0.]],
                          [[0, 0., 35., -17.5, 0., 0., 0.],
                           [0., 0., 35., 0., 0., -26.25, 0.],
                           [0, 0., 35., 0., 0., 0., 0.]]])

        npt.assert_almost_equal(der.data, a_der)

    def test_integ(self):
        integ = self.samples1.integ(5.)

        a_integ = np.array([[[0., 0., 0., 0., 0., 0., 0.],
                             [0, 0., 0., 0., 0., 0., 0.],
                             [0, 0, -1.4, 0., 0., 0., 0.]],
                            [[0, 0, -1.4, 0.175, 0., 0., 0.],
                             [0., 0, -1.4, 0., 0., 0.116666666, 0.],
                             [0, 0., -1.4, 0, 0., 0., 0.]]])

        npt.assert_almost_equal(integ.data, a_integ)

    def test_integ_deriv(self):
        d = {'samples3': self.samples3,
             'samples4': self.samples4,
             'zero': self.zero,
             'cos': self.cos,
             'sin': self.sin,
             'fit': self.fit,
             'fit2D': self.fit2D}

        for key, tp_ref in d.items():
            tmp_value = pm.PyTPeriodicMatrix.fromRFFT(tp_ref.data, nopack=True)
            with self.subTest(msg=key):
                for i in range(1, 10):
                    tmp_value = tmp_value.integ(i, nopack=True)

                for j in reversed(range(1, 10)):
                    tmp_value = tmp_value.deriv(j, nopack=True)

                npt.assert_almost_equal(tmp_value.data, tp_ref.data)

    def test_tf(self):
        d = {'test1': dict(num=np.array([1]), den=np.array([1, 1]), omega=1, order=1, mag=3),
             'test2': dict(num=np.array([1]), den=np.array([1, 1]), omega=1, order=2, mag=2),
             'test3': dict(num=np.array([1]), den=np.array([1, 1]), omega=2, order=1, mag=5),
             'test4': dict(num=np.array([1]), den=np.array([1, 1]), omega=2, order=2, mag=10),
             'test5': dict(num=np.array([1, 2, 0.4]), den=np.array([1, 2, 4, 6, 4, 6, 8]), omega=1, order=1, mag=3),
             'test6': dict(num=np.array([1, 2, 0.4]), den=np.array([1, 2, 4, 6, 4, 6, 8]), omega=1, order=5, mag=2),
             'test7': dict(num=np.array([1, 2, 0.4]), den=np.array([1, 2, 4, 6, 4, 6, 8]), omega=2, order=7, mag=5),
             'test8': dict(num=np.array([1, 2, 0.4]), den=np.array([1, 2, 4, 6, 4, 6, 8]), omega=2, order=3, mag=10),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                w, h = scipy.signal.freqs(data['num'], data['den'], np.array([data['omega'] * data['order']]))
                tp = pm.PyTPeriodicMatrix(data['order'], data['mag'], 0).tf(data['num'], data['den'], data['omega'])

                npt.assert_equal(tp == pm.PyTPeriodicMatrix(data['order'], data['mag'] * np.abs(h), np.angle(h)), True)

    def test_enclosed_area(self):
        d = {'one circle radius 1':
                 dict(tpx=pm.PyTPeriodicMatrix(1, 1, 0), tpy=pm.PyTPeriodicMatrix(1, 1, -np.pi / 2), area=np.pi),
             'double circle radius 1':
                 dict(tpx=pm.PyTPeriodicMatrix(2, 1, 0), tpy=pm.PyTPeriodicMatrix(2, 1, -np.pi / 2), area=2 * np.pi),
             'one circle radius 2':
                 dict(tpx=pm.PyTPeriodicMatrix(1, 2, 0), tpy=pm.PyTPeriodicMatrix(1, 2, -np.pi / 2), area=4 * np.pi),
             'one circle radius 3':
                 dict(tpx=pm.PyTPeriodicMatrix(1, 3, 0), tpy=pm.PyTPeriodicMatrix(1, 3, -np.pi / 2), area=9 * np.pi),
             'three circle radius 3':
                 dict(tpx=pm.PyTPeriodicMatrix(3, 3, 0), tpy=pm.PyTPeriodicMatrix(3, 3, -np.pi / 2), area=27 * np.pi),
             'ellipse a = 2, b = 10 one revolution':
                 dict(tpx=pm.PyTPeriodicMatrix(1, 10, 0), tpy=pm.PyTPeriodicMatrix(1, 2, -np.pi / 2),
                      area=62.831853071),
             'ellipse a = 11, b = 0.57 two revolutions':
                 dict(tpx=pm.PyTPeriodicMatrix(2, 0.57, 0), tpy=pm.PyTPeriodicMatrix(2, 11, -np.pi / 2),
                      area=39.39557188
                      ),
             'test with matrix':
                 dict(tpx=self.samples3, tpy=self.samples4, area=np.array([[18.84955592, 31.41592653, 37.69911184],
                                                                           [37.69911184, 31.41592653, 18.84955592]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                area = data['tpx'].enclosed_area(data['tpy'])

                npt.assert_almost_equal(area, data['area'])

    def test_curve2dPerimeter(self):
        d = {'sin-cos':
                 dict(tpx=self.sin, tpy=self.cos, peri=2 * np.pi),
             'test with matrix':
                 dict(tpx=self.samples3, tpy=self.samples4, peri=np.array([[24.90007959, 23.01311259, 22.10349216],
                                                                           [22.10349216, 23.01311259, 24.9000795]])),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                peri = data['tpx'].curve2dPerimeter(1 * data['tpy'])

                npt.assert_almost_equal(peri, data['peri'], decimal=4)

    @unittest.skip
    def test_elementwise_mul(self):
        self.fail()

    @unittest.skip
    def test_elementwise_div(self):
        self.fail()

    @unittest.skip
    def test_elementwise_power(self):
        self.fail()

    @unittest.skip
    def test_skew(self):
        self.fail()

    @unittest.skip
    def test_transpose(self):
        self.fail()

    @unittest.skip('Skipped due to unreliable reproducibility')
    def test_sum(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                tp_per_sum = tp_ref.sum()
                tp_per_sum_ref = pm.PyTPeriodicMatrix(1, 1, mode='zero')

                for row in range(tp_ref.nrows()):
                    for col in range(tp_ref.ncols()):
                        tp_per_sum_ref = tp_per_sum_ref + tp_ref[row, col]

                # error = np.abs(tp_per_sum.data - tp_per_sum_ref.data).max()
                npt.assert_array_almost_equal(tp_per_sum.data, tp_per_sum_ref.data)
                npt.assert_allclose(tp_per_sum.data, tp_per_sum_ref.data)

    def test_toObjArray(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):
                objarray = tp_ref.toObjArray()

                self.assertIsInstance(objarray, np.ndarray)
                self.assertEqual(objarray.dtype, np.object)

                for row in range(objarray.shape[0]):
                    for col in range(objarray.shape[1]):
                        npt.assert_almost_equal(objarray[row, col].data, tp_ref[row, col].data)

    @unittest.skip
    def test_to_msgpack(self):
        self.fail()

    def test_toCBOR(self):
        for key, tp_ref in self.ref_tps.items():
            for f in [1, 2, 6]:
                with self.subTest(msg='{}-f={}'.format(key, f)):
                    d = tp_ref.toCBOR(f)
                    reduced_samples = tp_ref.transform_order(1, f)
                    buffer = d['bins']['data']
                    shape = d['bins']['shape']
                    rfft_array = np.ndarray(shape=shape, buffer=buffer, dtype='>f8', order='C')
                    tp = pm.PyTPeriodicMatrix(rfft_array, mode="rfft")
                    npt.assert_almost_equal(tp.data, reduced_samples.data)

    def test_saveCBOR(self):
        for key, tp_ref in self.ref_tps.items():
            for f in [1, 2, 6]:
                with self.subTest(msg='{}-f={}'.format(key, f)):
                    file = tempfile.mktemp(suffix='.cbor', dir=self.test_dir)
                    tp_ref.saveCBOR(file, f)
                    '''check if file was created'''
                    self.assertTrue(os.path.isfile(file))

    def test_hstack(self):

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['rfft', ]:
                    self.skipTest('subtest {} skipped'.format(key))

                tp_hstack = pm.PyTPeriodicMatrix.hstack((tp_ref, tp_ref))

                self.assertIsInstance(tp_hstack, pm.PyTPeriodicMatrix)

                nrows = tp_ref.nrows()
                ncols = tp_ref.ncols()

                for row in range(nrows):
                    for col in range(ncols):
                        npt.assert_almost_equal(tp_ref[row, col].data, tp_hstack[row, col].data)
                        npt.assert_almost_equal(tp_ref[row, col].data, tp_hstack[row, ncols + col].data)

    def test_vstack(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['rfft', ]:
                    self.skipTest('subtest {} skipped'.format(key))

                tp_vstack = pm.PyTPeriodicMatrix.vstack((tp_ref, tp_ref))

                self.assertIsInstance(tp_vstack, pm.PyTPeriodicMatrix)

                nrows = tp_ref.nrows()
                ncols = tp_ref.ncols()

                for row in range(nrows):
                    for col in range(ncols):
                        npt.assert_almost_equal(tp_ref[row, col].data, tp_vstack[row, col].data)
                        npt.assert_almost_equal(tp_ref[row, col].data, tp_vstack[nrows + row, col].data)

    def test_concatenate(self):
        for key, tp_ref in self.ref_tps.items():
            for axis in [0, 1]:
                with self.subTest(msg='{}-axis-{}'.format(key, axis)):

                    if key in ['rfft', ]:
                        self.skipTest('subtest {} skipped'.format(key))

                    tp_concatenate = pm.PyTPeriodicMatrix.concatenate((tp_ref, tp_ref), axis=axis)

                    self.assertIsInstance(tp_concatenate, pm.PyTPeriodicMatrix)

                    nrows = tp_ref.nrows()
                    ncols = tp_ref.ncols()

                    for row in range(nrows):
                        for col in range(ncols):
                            if axis == 0:
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_concatenate[row, col].data)
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_concatenate[nrows + row, col].data)
                            else:
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_concatenate[row, col].data)
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_concatenate[row, ncols + col].data)

        with self.subTest(msg='Exeption-WrongAxis'):
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix.concatenate((self.samples1, self.samples1), axis=3)

        with self.subTest(msg='Exeption-UnequalConcatAxis'):
            tp2x3 = pm.PyTPeriodicMatrix(2, 3, mode='random', maxorder=10)
            tp3x2 = pm.PyTPeriodicMatrix(3, 2, mode='random', maxorder=10)

            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix.concatenate((tp2x3, tp3x2), axis=0)

            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix.concatenate((tp2x3, tp3x2), axis=1)

    def test_insert(self):
        for key, tp_ref in self.ref_tps.items():
            for axis in [0, 1]:
                with self.subTest(msg='{}-axis-{}'.format(key, axis)):

                    if key in ['rfft', ]:
                        self.skipTest('subtest {} skipped'.format(key))

                    if tp_ref.nrows() > 1 and tp_ref.ncols() > 1:
                        index = 2
                    else:
                        index = 1

                    tp_insert = pm.PyTPeriodicMatrix.insert(tp_ref, tp_ref, index, axis=axis)

                    self.assertIsInstance(tp_insert, pm.PyTPeriodicMatrix)

                    nrows = tp_ref.nrows()
                    ncols = tp_ref.ncols()

                    for row in range(nrows):
                        for col in range(ncols):
                            if axis == 0:
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_insert[index + row, col].data)
                            else:
                                npt.assert_almost_equal(tp_ref[row, col].data, tp_insert[row, index + col].data)
        with self.subTest(msg='Exeption-WrongAxis'):
            with self.assertRaises(Exception):
                pm.PyTPeriodicMatrix.insert(self.samples1, self.samples1, index=1, axis=3)

    @unittest.skip
    def test_getOrders(self):
        self.fail()

    @unittest.skip
    def test_getSpectrum(self):
        self.fail()

    @unittest.skip
    def test_plot(self):
        self.fail()

    @unittest.skip
    def test_plot_spectrum(self):
        self.fail()

    @unittest.skip
    def test_show_preview(self):
        self.fail()

    @unittest.skip
    def test_show_full_preview(self):
        self.fail()

    @unittest.skip
    def test_show_spectrum(self):
        self.fail()

    @unittest.skip
    def test__repr_png_(self):
        self.fail()

    def test_getitem(self):
        """ tests indexing and packing of the new object """
        npt.assert_almost_equal(self.samples1[1, 0].data,
                                np.array([[[10.5, 7.0, 0., 0., 1.75, 0., 0.]]]))

    def test_setitem(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg='elem-{}'.format(key)):

                tmp = pm.PyTPeriodicMatrix.zeros(tp_ref.shape, nopack=True)
                for row in range(tp_ref.nrows()):
                    for col in range(tp_ref.ncols()):
                        tmp[row, col] = tp_ref[row, col]

                npt.assert_almost_equal(tp_ref.data, tmp.data)

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg='row-{}'.format(key)):

                tmp = pm.PyTPeriodicMatrix.zeros(tp_ref.shape, nopack=True)
                for row in range(tp_ref.nrows()):
                    tmp[row, :] = tp_ref[row, :]

                npt.assert_almost_equal(tp_ref.data, tmp.data)

        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg='col-{}'.format(key)):

                tmp = pm.PyTPeriodicMatrix.zeros(tp_ref.shape, nopack=True)
                for col in range(tp_ref.ncols()):
                    tmp[:, col] = tp_ref[:, col]

                npt.assert_almost_equal(tp_ref.data, tmp.data)

    def test_equality_func(self):
        for key, tp_ref in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['rfft', ]:
                    self.skipTest('subtest {} skipped'.format(key))

                npt.assert_equal(tp_ref == tp_ref, True)
                npt.assert_equal(tp_ref == tp_ref.shift(2 * np.pi), True)
                npt.assert_equal(tp_ref == tp_ref.shift(- 2 * np.pi), True)

        with self.subTest(msg='extra-sin-cos'):
            npt.assert_equal(self.sin == self.sin, True)
            npt.assert_equal(self.cos == self.sin, False)
            npt.assert_equal(self.cos == self.sin.shift(np.pi / 2), True)
            npt.assert_equal(self.sin == self.cos.shift(-np.pi / 2), True)

    def test_rightside_division(self):
        d = {'sin-1': dict(tp=self.sin, rdiv=2, mul=0.5),
             'cos-1': dict(tp=self.cos, rdiv=2, mul=0.5),
             'zero-1': dict(tp=self.zero, rdiv=2, mul=0.5),
             'samples1-1': dict(tp=self.samples1, rdiv=2, mul=0.5),
             'samples1-2': dict(tp=self.samples1, rdiv=5, mul=1 / 5),
             }

        for key, data in d.items():
            with self.subTest(msg=key):
                tmp_mul = data['tp'] / data['rdiv']
                tmp_rdiv = data['mul'] * data['tp']

                npt.assert_almost_equal(tmp_mul.data, tmp_rdiv.data)

    def test_greater_than(self):
        margin = 1e-3

        for key, tp in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['samples3', 'samples4']:
                    self.skipTest('subtest {} skipped'.format(key))

                npt.assert_equal(tp > tp, False)
                npt.assert_equal(tp - margin > tp, False)
                npt.assert_equal(tp + margin > tp, True)

    def test_less_than(self):
        margin = 1e-3

        for key, tp in self.ref_tps.items():
            with self.subTest(msg=key):

                if key in ['samples3', 'samples4']:
                    self.skipTest('subtest {} skipped'.format(key))

                npt.assert_equal(tp < tp, False)
                npt.assert_equal(tp - margin < tp, True)
                npt.assert_equal(tp + margin < tp, False)

    def test_indexing(self):
        n = 1000
        tp = pm.PyTPeriodicMatrix(3, 2, mode='random')
        samples = tp.get_samples(n)
        tp1 = tp[0, 0]
        tp2 = tp[0, 1]
        tp3 = tp[1, 0]
        tp4 = tp[1, 1]
        tp5 = tp[2, 0]
        tp6 = tp[2, 1]

        npt.assert_almost_equal(samples[0, 0, :], tp1.get_samples(n).flatten())
        npt.assert_almost_equal(samples[0, 1, :], tp2.get_samples(n).flatten())
        npt.assert_almost_equal(samples[1, 0, :], tp3.get_samples(n).flatten())
        npt.assert_almost_equal(samples[1, 1, :], tp4.get_samples(n).flatten())
        npt.assert_almost_equal(samples[2, 0, :], tp5.get_samples(n).flatten())
        npt.assert_almost_equal(samples[2, 1, :], tp6.get_samples(n).flatten())

    def test_slicing(self):
        tp = pm.PyTPeriodicMatrix(3, 2, mode='random')
        tp_slice = tp[0:3, 0:2]
        npt.assert_almost_equal(tp.get_samples(), tp_slice.get_samples())

    def test_mixed_indexing(self):
        tp1 = pm.PyTPeriodicMatrix(3, 3, mode='random')
        tp2 = tp1[0:2, 0]
        npt.assert_equal(tp2.shape, (2, 1))
        samples1 = tp1.get_samples(100)
        samples2 = tp2.get_samples(100)
        npt.assert_almost_equal(samples1[0:2, 0, :].flatten(), samples2.flatten())

        tp3 = tp1[2, :]
        npt.assert_equal(tp3.shape, (1, 3))
        samples1 = tp1.get_samples(100)
        samples2 = tp3.get_samples(100)
        npt.assert_almost_equal(samples1[2, :, :].flatten(), samples2.flatten())

    def test_allclose(self):
        # margin = 1e-5
        # rtol = 1e-5
        # atol = 1e-8
        for key, tp in self.ref_tps.items():
            with self.subTest(msg=key):
                if key in ['rfft', ]:
                    self.skipTest('subtest {} skipped'.format(key))

                # if key in ['samples3', 'samples4', 'sin', 'cos', 'rfft', 'fit', 'fit2D', 'objarray']:
                #     margin = 1e-3
                # elif key in ['dc', ]:
                #     margin = 1e-4
                # else:
                #     margin = 1e-5

                self.assertTrue(tp.allclose(tp))
                # self.assertFalse(tp.allclose(tp + margin, rtol=rtol, atol=atol))
                # self.assertFalse(tp.allclose(tp - margin, rtol=rtol, atol=atol))
                self.assertTrue(tp.allclose(tp.shift(2 * np.pi)))
                self.assertTrue(tp.allclose(tp.shift(- 2 * np.pi)))

        with self.subTest(msg='extra-sin-cos'):
            npt.assert_equal(self.sin == self.sin, True)
            npt.assert_equal(self.cos == self.sin, False)
            npt.assert_equal(self.cos == self.sin.shift(np.pi / 2), True)
            npt.assert_equal(self.sin == self.cos.shift(-np.pi / 2), True)

    def test_scaling_factor(self):
        for key, tp in self.ref_tps.items():
            with self.subTest(msg=key):
                self.assertEqual(tp.scaling_factor, 1)

        for key, tp in self.ref_scaled_tps.items():
            with self.subTest(msg=key):
                self.assertEqual(tp.scaling_factor, 2)

        with self.subTest(msg='Exeption-UnequalScaling Factor'):
            tmp1 = pm.PyTPeriodicMatrix(1, 1, mode='eye')
            tmp2 = pm.PyTPeriodicMatrix(1, 1, mode='eye')
            tmp2.scaling_factor = 2
            with self.assertRaises(Exception):
                _ = tmp1 + tmp2


if __name__ == '__main__':
    from lcm.utils.testing import runUnittest

    runUnittest(pm.PyTPeriodicMatrix.getInterpreter())
