/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.utils.simulate;

import java.awt.AWTError;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.util.Arrays;

/**
 * LayoutManger to order components in a line with dynamic width. Width of
 * container is divide corresponding the weight.
 * 
 * @author Werner Koppelsttter
 */
public class ProgressLayout implements LayoutManager {

	/** Target container of the LayoutManager */
	private Container parent;

	/** Weight of the components */
	private int[] weight = new int[0];

	/**
	 * Set the weights of the components
	 * 
	 * @param weight
	 *            Array with the weights of the components
	 */
	public void setWeight(int[] weight) {
		if (weight != null) {
			this.weight = weight;
		}
		this.layoutContainer(this.parent);
	}

	/**
	 * Create a layout-manager for container
	 * 
	 * @param parent
	 *            Container for the layout-manager
	 */
	public ProgressLayout(Container parent) {
		this.parent = parent;
	}

	/**
	 * Check if given container is the same as the one stored
	 * 
	 * @param parent
	 *            Container to check
	 */
	private void checkContainer(Container parent) {
		if (this.parent != parent) {
			throw new AWTError("StatusLayout can't be shared");
		}
	}

	@Override
	public void addLayoutComponent(String name, Component comp) {
	}

	@Override
	public void removeLayoutComponent(Component comp) {
	}

	@Override
	public Dimension preferredLayoutSize(Container parent) {
		Dimension d;
		Component c;
		int nComps, width, height;
		Insets insets = parent.getInsets();

		this.checkContainer(parent);
		width = 0;
		height = 0;

		nComps = parent.getComponentCount();
		for (int i = 0; i < nComps; i++) {
			c = parent.getComponent(i);
			d = c.getPreferredSize();
			width = Math.max(width, d.width);
			height = Math.max(height, d.height);
		}

		return new Dimension(width + insets.left + insets.right,
				height + insets.top + insets.bottom);
	}

	@Override
	public Dimension minimumLayoutSize(Container parent) {
		Insets insets = parent.getInsets();

		this.checkContainer(parent);

		return new Dimension(insets.left + insets.right,
				insets.top + insets.bottom);
	}

	@Override
	public void layoutContainer(Container parent) {
		Component c;
		Insets insets = parent.getInsets();
		int nComps = parent.getComponentCount();
		int width = parent.getWidth() - (insets.left + insets.right);
		int height = parent.getHeight() - (insets.top + insets.bottom);
		int x = insets.left, sum, w, ground;
		int[] wght;

		this.checkContainer(parent);

		if ((width > 0) && (height > 0)) {
			if (this.weight.length < nComps) {
				this.weight = Arrays.copyOf(this.weight, nComps);
			}

			// get sum of weight and ground value of weight
			sum = 0;
			for (int i = 0; i < nComps; i++) {
				sum += this.weight[i];
			}
			ground = (10 * sum) / width;

			// calculate expanded weight
			wght = new int[nComps];
			sum = 0;
			for (int i = 0; i < nComps; i++) {
				if (this.weight[i] == 0) {
					wght[i] = 0;
				} else if (this.weight[i] < ground) {
					wght[i] = ground;
				} else {
					wght[i] = this.weight[i];
				}
				sum += wght[i];
			}

			// set size of components
			for (int i = 0; i < nComps; i++) {
				c = parent.getComponent(i);
				try {
					w = (width * wght[i]) / sum;
				} catch (Exception e) {
					w = 0;
				}
				c.setBounds(insets.left + x, insets.top, w, height);

				x += w;
			}
		}
	}
}
