/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.utils.simulate;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.MouseListener;
import java.util.Arrays;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

/**
 * StatusPanel to display number of different status
 * 
 * @author Werner Koppelsttter
 */
public class ProgressPanel extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Panel to display each status */
	private JPanel[] panels;

	/** Labels to display number in status */
	private JLabel[] labels;

	/** LayoutManager of panel */
	private ProgressLayout layout;

	/**
	 * Create panel with sub-panel for each status
	 * 
	 * @param color
	 *            Array with color of wanted status
	 */
	public ProgressPanel(Color[] color) {
		this.panels = new JPanel[color.length];
		this.labels = new JLabel[color.length];
		this.layout = new ProgressLayout(this);
		this.setLayout(this.layout);

		// create panels
		for (int k = 0; k < color.length; k++) {
			this.panels[k] = new JPanel();
			this.panels[k].setLayout(new BorderLayout());
			this.panels[k].setBackground(color[k]);
			this.labels[k] = new JLabel("0");
			this.labels[k].setHorizontalAlignment(SwingConstants.CENTER);
			this.labels[k].setVerticalAlignment(SwingConstants.CENTER);
			this.panels[k].add(this.labels[k]);
			this.add(this.panels[k]);
		}
	}

	/**
	 * Set the weights of the components
	 * 
	 * @param weight
	 *            Array with the weights of the components
	 */
	public void setWeight(int[] weight) {

		if (weight != null) {
			if (weight.length < this.labels.length) {
				weight = Arrays.copyOf(weight, this.labels.length);
			}

			for (int k = 0; k < this.labels.length; k++) {
				this.labels[k].setText(Integer.toString(weight[k]));
				this.labels[k].setToolTipText(Integer.toString(weight[k]));
			}

			this.layout.setWeight(weight);
		}
	}

	/**
	 * Adds the mouse listener to panel with the index
	 * 
	 * @param index
	 *            Index of panel
	 * @param listener
	 *            Listener to add
	 */
	public void addMouseListener(int index, MouseListener listener) {
		if ((0 <= index) && (index <= this.panels.length)) {
			this.labels[index].addMouseListener(listener);
		}
	}

	/**
	 * Get mouse listeners of panel with the index
	 * 
	 * @param index
	 *            Index of panel
	 * @return List of listeners
	 */
	public MouseListener[] getMouseListeners(int index) {
		MouseListener[] list = null;

		if ((0 <= index) && (index <= this.panels.length)) {
			list = this.panels[index].getMouseListeners();
		}

		return list;
	}

	/**
	 * Remove the mouse listener of panel with the index
	 * 
	 * @param index
	 *            Index of panel
	 * @param listener
	 *            Listener to remove
	 */
	public void removeMouseListener(int index, MouseListener listener) {
		if ((0 <= index) && (index <= this.panels.length)) {
			this.panels[index].removeMouseListener(listener);
		}
	}

}
