/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9715 $
 * $Date: 2020-03-19 10:01:15 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.symspace.center.views.tree;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Parameter;

/**
 * 
 * @author Markus Priglinger
 * @author Werner Koppelsttter
 */
public class ProParameterModel extends AbstractTreeTableModel {

	private ArrayList<IndexRendererPair> _irpList = null;

	private ProTreeTableIndexViewManager _indexViewManager = null;

	/**
	 * Constructor for creating a ProParameterModel.
	 */

	public ProParameterModel(Parameter root) {
		super(root);

		this._irpList = new ArrayList<IndexRendererPair>();
	}

	public void setIndexViewManager(ProTreeTableIndexViewManager indexViewManager) {

		this._indexViewManager = indexViewManager;
	}

	public void setColumnStructure(ArrayList<IndexRendererPair> irpList) {

		this._irpList = irpList;
	}

	//
	// TreeModel interface
	//

	/**
	 * TreeModel method to return the number of children of a particular node. Since
	 * <code>node</code> is a TreeNode, this can be answered via the TreeNode method
	 * <code>getChildCount</code>.
	 */
	@Override
	public int getChildCount(Object parent) {

		int count = 0;
		if (parent instanceof ParamContainer) {
			count = ((ParamContainer) parent).getParam().length;
		}

		return count;
	}

	/**
	 * TreeModel method to locate a particular child of the specified node. Since
	 * <code>node</code> is a TreeNode, this can be answered via the TreeNode method
	 * <code>getChild</code>.
	 */
	@Override
	public Object getChild(Object parent, int index) {

		Parameter child = null;
		Parameter[] param;

		if (parent instanceof ParamContainer) {
			param = ((ParamContainer) parent).getParam();
			if ((index < param.length) && (index >= 0)) {
				child = param[index];
			}
		}

		return child;
	}

	/**
	 * TreeModel method to determine if a node is a leaf. Since <code>node</code> is
	 * a TreeNode, this can be answered via the TreeNode method <code>isLeaf</code>.
	 */
	@Override
	public boolean isLeaf(Object node) {
		// TODO number of childs (insert in empty container fails)
		return !(node instanceof ParamContainer);
		// return this.getChildCount(node) == 0;
	}

	//
	// The TreeTable interface.
	//

	/**
	 * Returns the number of column names passed into the constructor.
	 */
	@Override
	public int getColumnCount() {

		return this._irpList.size();
	}

	/**
	 * Returns the column name passed into the constructor.
	 */
	@Override
	public String getColumnName(int column) {

		return this._irpList.get(column).getIndex() + " (no header)";
	}

	/**
	 * Returns the column class for column <code>column</code>. This is set in the
	 * constructor.
	 */
	@Override
	@SuppressWarnings("rawtypes")
	public Class getColumnClass(int column) {

		if (column == 0) {
			return TreeTableModel.class;
		} else {
			return TData.class;
		}
	}

	/**
	 * Returns the value for the column <code>column</code> and object
	 * <code>node</code>. The return value is determined by invoking the method
	 * specified in constructor for the passed in column.
	 */
	@Override
	public Object getValueAt(Object node, int column) {

		if (node instanceof ParamField) {
			ParamField parField = (ParamField) node;
			if (column >= 0) {
				int idx = this._indexViewManager.getIndexForColumnPosition(column);
				return parField.getData(idx);
			}
		}
		return null;
	}

	/**
	 * <code>isCellEditable</code> is invoked by the JTreeTable to determine if a
	 * particular entry can be added. This is overridden to return true for the
	 * first column, assuming the node isn't the root, as well as returning two for
	 * the second column if the node is a BookmarkEntry. For all other columns this
	 * returns false.
	 */
	@Override
	public boolean isCellEditable(Object node, int column) {
		boolean edit = false;
		TData data;

		if (column == 0) {
			edit = true;
		} else if (column == 1) {
			if (node instanceof ParamField) {
				ParamField parField = (ParamField) node;
				if (!parField.getLock() && !parField.getReadOnly()) {
					data = parField.getData();
					if (data.getClass().equals(TNumeric.class)) {
						edit = data.toString().equals(((TNumeric) data).toFullString());
					} else if (data.getClass().equals(TString.class)) {
						edit = data.toString().equals("'" + ((TString) data).getString() + "'");
					} else if (data.getClass().equals(TBoolean.class)) {
						edit = true;
					} else if (data.getClass().equals(TNull.class)) {
						edit = true;
					}
				}
			}
		}

		return edit;
	}

	@Override
	public void setValueAt(Object aValue, Object node, int column) {

		if (node instanceof ParamField) {

			ParamField parField = (ParamField) node;
			TData currentData = parField.getData();
			TData newData = null;

			try {
				newData = TData.createData((String) aValue);
			} catch (TDataException e1) {
				e1.printStackTrace();
			}

			boolean dataChanged = !(currentData.equals(newData));
			if (!currentData.getClass().equals(newData.getClass())) {

				dataChanged = true;
				Object[] options = { "Yes", "No" };
				int ret = JOptionPane.showOptionDialog(null,
						"Do you want to change the current data type?\n" + "Please confirm!",
						"Data type change confirmation", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null,
						options, options[0]);
				if (ret == JOptionPane.NO_OPTION) {
					return;
				}
			}

			try {
				parField.setData(newData);
			} catch (Exception e) {
				e.printStackTrace();
			}

			if (dataChanged) {
				// FIXME KW
				// GUIManager.getInstance().setProjectDirty();

				// The value changed, fire an event to notify listeners.
				ParamContainer parent = ((Parameter) node).getParent();

				this.fireTreeNodesChanged(this, this.getPathToRoot(parent),
						new int[] { this.getIndexOfChild(parent, node) }, new Object[] { node });
			}

		}
	}

	/**
	 * Builds the parents of the node up to and including the root node, where the
	 * original node is the last element in the returned array. The length of the
	 * returned array gives the node's depth in the tree.
	 * 
	 * @param aNode
	 *            the TreeNode to get the path for
	 */
	public Parameter[] getPathToRoot(Parameter aNode) {
		return this.getPathToRoot(aNode, 0);
	}

	/**
	 * Builds the parents of the node up to and including the root node, where the
	 * original node is the last element in the returned array. The length of the
	 * returned array gives the node's depth in the tree.
	 * 
	 * @param aNode
	 *            the TreeNode to get the path for
	 * @param depth
	 *            an int giving the number of steps already taken towards the root
	 *            (on recursive calls), used to size the returned array
	 * @return an array of TreeNodes giving the path from the root to the specified
	 *         node
	 */
	private Parameter[] getPathToRoot(Parameter aNode, int depth) {

		Parameter[] retNodes;
		// This method recurses, traversing towards the root in order
		// size the array. On the way back, it fills in the nodes,
		// starting from the root and working back to the original node.

		/*
		 * Check for null, in case someone passed in a null node, or they passed in an
		 * element that isn't rooted at root.
		 */
		if (aNode == null) {
			if (depth == 0) {
				return null;
			} else {
				retNodes = new Parameter[depth];
			}
		} else {
			depth++;
			if (aNode == this.root) {
				retNodes = new Parameter[depth];
			} else {
				retNodes = this.getPathToRoot(aNode.getParent(), depth);
			}
			retNodes[retNodes.length - depth] = aNode;
		}
		return retNodes;
	}
}
