/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10034 $
 * $Date: 2020-10-21 18:25:19 +0200 (Wed, 21 Oct 2020) $
 */
package at.lcm.symspace.center.views.tree;

import java.awt.Color;
import java.awt.Component;
import java.rmi.RemoteException;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.tree.TreeCellRenderer;

import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.magopt.structure.Access;
import at.lcm.bu21.magopt.structure.IParamCollector;
import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.IParamDirectory;
import at.lcm.bu21.magopt.structure.IParamField;
import at.lcm.bu21.magopt.structure.IParamFile;
import at.lcm.bu21.magopt.structure.IParamFormula;
import at.lcm.bu21.magopt.structure.IParamFunction;
import at.lcm.bu21.magopt.structure.IParamJavaFunction;
import at.lcm.bu21.magopt.structure.IParamLinkedField;
import at.lcm.bu21.magopt.structure.IParamMaterial;
import at.lcm.bu21.magopt.structure.IParamMatlabFunction;
import at.lcm.bu21.magopt.structure.IParamPythonFunction;
import at.lcm.bu21.magopt.structure.IParamPythonFunction_old;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.symspace.center.icons.IconParameter;
import at.lcm.symspace.center.icons.IconProperty;

/**
 * TreeCellRenderer for MagOpt-Projects. Define a costume style with several
 * icons
 * 
 * @author Werner Koppelsttter
 */
public class ProjectTreeCellRenderer extends JPanel implements TreeCellRenderer {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Set the style of the current node
	 * 
	 * @param tree     Tree to insert
	 * @param value    Object to display
	 * @param sel      If node is selected
	 * @param expanded If node if expanded
	 * @param leaf     If node is a leaf
	 * @param row      Number of row
	 * @param hasFocus If node has focus
	 * @return Component to draw node
	 */
	@Override
	public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf,
			int row, boolean hasFocus) {

		JLabel label;
		IParameter param = (IParameter) value;

		this.setOpaque(false);
		this.removeAll();
		this.setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		this.add(new JLabel(new ImageIcon(IconParameter.getIcon(param))));
		try {
			if (param instanceof IProject) {
				this.addIcons((IProject) param);
			} else if (param instanceof IParamFormula) {
				this.addIcons((IParamFormula) param);
			} else if (param instanceof IParamLinkedField) {
				this.addIcons((IParamLinkedField) param);
			} else if (param instanceof IParamField) {
				this.addIcons((IParamField) param);
			} else if (param instanceof IParamMatlabFunction) {
				this.addIcons((IParamMatlabFunction) param);
			} else if (param instanceof IParamPythonFunction) {
				this.addIcons((IParamPythonFunction) param);
			} else if (param instanceof IParamPythonFunction_old) {
				this.addIcons((IParamPythonFunction_old) param);
			} else if (param instanceof IParamJavaFunction) {
				this.addIcons((IParamJavaFunction) param);
			} else if (param instanceof IParamMaterial) {
				this.addIcons((IParamMaterial) param);
			} else if (param instanceof IParamCollector) {
				this.addIcons((IParamCollector) param);
			} else if (param instanceof IParamContainer) {
				this.addIcons((IParamContainer) param);
			} else if (param instanceof IParamFile) {
				this.addIcons((IParamFile) param);
			} else if (param instanceof IParamDirectory) {
				this.addIcons((IParamDirectory) param);
			}

			// this.add(Box.createRigidArea(new Dimension(5, 0)));

			label = new JLabel(param.getName());

			if (sel) {
				this.setBackground(Color.LIGHT_GRAY);
				if (param.getUsed()) {
					label.setForeground(Color.BLACK);
				} else {
					label.setForeground(Color.WHITE);
				}
			} else {
				this.setBackground(tree.getBackground());
				if (param.getUsed()) {
					label.setForeground(Color.BLACK);
				} else {
					label.setForeground(Color.LIGHT_GRAY);
				}
			}

			this.add(label);

			this.setToolTipText(param.getDescription());
			this.revalidate();

			// this.setPreferredSize(null);
		} catch (RemoteException e) {
			// TODO RMI
		}
		return this;
	}

	/**
	 * Add icons for class Parameter
	 * 
	 * @param param Parameter object
	 * @throws RemoteException
	 */
	protected void addIcons(IParameter param) throws RemoteException {
		if (param.getConstant()) {
			this.add(new JLabel(new ImageIcon(IconProperty.CONSTANT)));
		}
		if (param.getReadOnly()) {
			this.add(new JLabel(new ImageIcon(IconProperty.READONLY)));
		}
		if (param.getModified()) {
			this.add(new JLabel(new ImageIcon(IconProperty.MODIFIED)));
		}
		if (param.getUncertain()) {
			this.add(new JLabel(new ImageIcon(IconProperty.UNCERTAIN)));
		}
		if (param.getError()) {
			this.add(new JLabel(new ImageIcon(IconProperty.ERROR)));
		}
		if (param.getWarning()) {
			String str = MagOptConfig.getConfig("ShowTreeWarning");
			if (str != null) {
				if (Boolean.parseBoolean(str)) {
					this.add(new JLabel(new ImageIcon(IconProperty.WARNING)));
				}
			}
		}
	}

	/**
	 * Add icons for class Project
	 * 
	 * @param param Project object
	 * @throws RemoteException
	 */
	protected void addIcons(IProject param) throws RemoteException {
		if (param.isWriteProtected()) {
			this.add(new JLabel(new ImageIcon(IconProperty.WRITE_PROTECTED)));
		}
		this.addIcons((IParamContainer) param);
	}

	/**
	 * Add icons for class ParamFormula
	 * 
	 * @param param ParamFormula object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamFormula param) throws RemoteException {
		this.addIcons((IParamField) param);
		if (param.getUpdateRequired()) {
			this.add(new JLabel(new ImageIcon(IconProperty.UPDATEREQUIRED)));
		}
	}

	/**
	 * Add icons for class LinkedField
	 * 
	 * @param param ParamLinkedField object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamLinkedField param) throws RemoteException {
		this.addIcons((IParamField) param);
		if (param.getUpdateRequired()) {
			this.add(new JLabel(new ImageIcon(IconProperty.UPDATEREQUIRED)));
		}
	}

	/**
	 * Add icons for class ParamField
	 * 
	 * @param param ParamField object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamField param) throws RemoteException {
		this.addIcons((IParameter) param);
		if (param.getLock()) {
			this.add(new JLabel(new ImageIcon(IconProperty.LOCK)));
		}
		if (param.getData() instanceof TNull) {
			this.add(new JLabel(new ImageIcon(IconProperty.EMPTY)));
		}
	}

	/**
	 * Add icons for class ParamFunction
	 * 
	 * @param param ParamFunction object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamFunction param) throws RemoteException {
		this.addIcons((IParamContainer) param);
		if (param.getUpdateRequired()) {
			this.add(new JLabel(new ImageIcon(IconProperty.UPDATEREQUIRED)));
		}
	}

	/**
	 * Add icons for class ParamMatlabFunction
	 * 
	 * @param param ParamMatlabFunction object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamMatlabFunction param) throws RemoteException {
		this.addIcons((IParamFunction) param);
	}

	/**
	 * Add icons for class ParamPythonFunction
	 * 
	 * @param param ParamPythonFunction object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamPythonFunction param) throws RemoteException {
		this.addIcons((IParamFunction) param);
	}

	/**
	 * Add icons for class ParamPythonFunction_old
	 * 
	 * @param param ParamPythonFunction_old object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamPythonFunction_old param) throws RemoteException {
		this.addIcons((IParamFunction) param);
	}

	/**
	 * Add icons for class ParamJavaFunction
	 * 
	 * @param param ParamJavaFunction object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamJavaFunction param) throws RemoteException {
		this.addIcons((IParamFunction) param);
	}

	/**
	 * Add icons for class ParamMaterial
	 * 
	 * @param param ParamMaterial object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamMaterial param) throws RemoteException {
		this.addIcons((IParamContainer) param);
	}

	/**
	 * Add icons for class ParamCollector
	 * 
	 * @param param ParamCollector object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamCollector param) throws RemoteException {
		this.addIcons((IParamContainer) param);
	}

	/**
	 * Add icons for class ParamContainer
	 * 
	 * @param param ParamContainer object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamContainer param) throws RemoteException {
		if (param.getComponent() != null) {
			this.add(new JLabel(new ImageIcon(IconProperty.COMPONENT)));
		}
		this.addIcons((IParameter) param);
		if (param.getIncomplete()) {
			this.add(new JLabel(new ImageIcon(IconProperty.INCOMPLETE)));
		}
		if (param.getAccess() == Access.PROTECTED) {
			this.add(new JLabel(new ImageIcon(IconProperty.PROTECTED)));
		}
		if (param.getAccess() == Access.PRIVATE) {
			this.add(new JLabel(new ImageIcon(IconProperty.PRIVATE)));
		}
		if (!param.getEnable()) {
			this.add(new JLabel(new ImageIcon(IconProperty.DISABLED)));
		}
	}

	/**
	 * Add icons for class ParamFile
	 * 
	 * @param param ParamFile object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamFile param) throws RemoteException {
		this.addIcons((IParamDirectory) param);
	}

	/**
	 * Add icons for class ParamDirectory
	 * 
	 * @param param ParamDirectory object
	 * @throws RemoteException
	 */
	protected void addIcons(IParamDirectory param) throws RemoteException {
		this.addIcons((IParameter) param);
	}

}
