/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9733 $
 * $Date: 2020-04-06 09:48:37 +0200 (Mon, 06 Apr 2020) $
 */
package at.lcm.symspace.gui.utils;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

/**
 * Storage setting to json-file
 *
 * @author Werner Koppelsttter
 */
public class Storage {

	/** Setting-file */
	private File file;

	/** Maximal number of recent files */
	private int numFileList = 10;

	/** Maximal number of recent optimizations */
	private int numOptList = 10;

	/**
	 * Description of an optimization
	 * 
	 * @author Werner Koppelstaetter
	 */
	public class OptimizationDescription {

		/** Base project of optimization */
		String mop;

		/** Name of archive */
		String archive;

		/** Name of problem */
		String problem;

		/**
		 * Create object from list
		 * 
		 * @param list
		 *            List with description-info
		 */
		protected OptimizationDescription(List<String> list) {
			this(list.get(0), list.get(1), list.get(2));
		}

		/**
		 * Create object
		 * 
		 * @param mop
		 *            Path ot base project
		 * @param archive
		 *            Name of archive
		 * @param problem
		 *            Name of problem
		 */
		public OptimizationDescription(String mop, String archive, String problem) {
			this.mop = mop;
			this.archive = archive;
			this.problem = problem;
		}

		/**
		 * Get path of base.project
		 * 
		 * @return Path to mop
		 */
		public String getMopFile() {
			return this.mop;
		}

		/**
		 * Get name of archive
		 * 
		 * @return Archive name
		 */
		public String getArchive() {
			return this.archive;
		}

		/**
		 * Get name of problem
		 * 
		 * @return Problem name
		 */
		public String getProblem() {
			return this.problem;
		}

		/**
		 * Get info as list
		 * 
		 * @return Description-info as list
		 */
		protected List<String> asList() {
			List<String> list;

			list = new ArrayList<String>();
			list.add(this.mop);
			list.add(this.archive);
			list.add(this.problem);

			return list;
		}

		@Override
		public String toString() {
			return "[" + this.mop + " : " + this.archive + " : " + this.problem + "]";
		}
	}

	/**
	 * Create storage-object for recent-files
	 */
	public Storage() {
		this.file = new File(new File(System.getProperty("user.home"), ".SyMSpace"), "storage.json");
	}

	/**
	 * Read JSON-object from file
	 * 
	 * @return JSON-object
	 */
	private JSONObject readStorage() {
		FileReader reader = null;
		JSONObject obj;

		if (this.file.exists()) {
			try {
				reader = new FileReader(this.file);
				obj = (JSONObject) new JSONParser().parse(reader);
			} catch (Exception e) {
				obj = new JSONObject();
				System.err.println(e.getMessage());
				e.printStackTrace();
			} finally {
				if (reader != null) {
					try {
						reader.close();
					} catch (IOException e) {
						System.err.println(e.getMessage());
						e.printStackTrace();
					}
				}
			}
		} else {
			obj = new JSONObject();
		}

		return obj;
	}

	/**
	 * Write JSON-object to file
	 * 
	 * @param obj
	 */
	private void writeStorage(JSONObject obj) {
		FileWriter writer = null;

		if (!this.file.getParentFile().exists()) {
			this.file.getParentFile().mkdirs();
		}
		try {
			writer = new FileWriter(this.file);
			obj.writeJSONString(writer);
		} catch (Exception e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}
			}
		}

	}

	/**
	 * Get list of recent opened projects
	 * 
	 * @return Recent-projects
	 */
	@SuppressWarnings("unchecked")
	public List<String> getRecentProjects() {
		JSONObject obj;
		List<String> list;

		obj = this.readStorage();
		list = (List<String>) obj.get("RecentFiles");

		return list;
	}

	/**
	 * Add path to recent-project
	 * 
	 * @param mop
	 *            Project-file
	 */
	@SuppressWarnings("unchecked")
	public void addRecentProject(String mop) {
		JSONObject obj;
		List<String> list;

		obj = this.readStorage();

		list = (List<String>) obj.get("RecentFiles");
		if (list == null) {
			list = new ArrayList<String>();
			obj.put("RecentFiles", list);
		}

		list.remove(mop);
		list.add(0, mop);
		if (list.size() > this.numFileList) {
			list.remove(list.size() - 1);
		}

		this.writeStorage(obj);
	}

	/**
	 * Get list of recent opened optimization
	 * 
	 * @return Recent-optimizations
	 */
	@SuppressWarnings("unchecked")
	public List<OptimizationDescription> getRecentOptimizations() {
		JSONObject obj;
		List<List<String>> list;
		List<OptimizationDescription> opts = null;

		obj = this.readStorage();
		list = (List<List<String>>) obj.get("RecentOptimizations");
		if (list != null) {
			opts = new ArrayList<Storage.OptimizationDescription>();
			for (List<String> entry : list) {
				opts.add(new OptimizationDescription(entry));
			}
		}

		return opts;
	}

	/**
	 * Add info to recent opened optimization
	 * 
	 * @param opt
	 *            Optimization description
	 */
	@SuppressWarnings("unchecked")
	public void addRecentOptimization(OptimizationDescription opt) {
		JSONObject obj;
		List<List<String>> list;

		obj = this.readStorage();

		list = (List<List<String>>) obj.get("RecentOptimizations");
		if (list == null) {
			list = new ArrayList<List<String>>();
			obj.put("RecentOptimizations", list);
		}

		list.remove(opt.asList());
		list.add(0, opt.asList());
		if (list.size() > this.numOptList) {
			list.remove(list.size() - 1);
		}

		this.writeStorage(obj);
	}

	/**
	 * Get user-macro path
	 * 
	 * @return User-path
	 */
	public String getUserMacroPath() {
		JSONObject obj;
		String path;

		obj = this.readStorage();
		path = (String) obj.get("UserMacroPath");

		return path;
	}

	/**
	 * Set user-macro-path
	 * 
	 * @param path
	 *            Project-file
	 */
	@SuppressWarnings("unchecked")
	public void setUserMacroPath(String path) {
		JSONObject obj;

		if ((path != null) && !path.isEmpty()) {
			obj = this.readStorage();
			obj.put("UserMacroPath", path);
			this.writeStorage(obj);
		}
	}

}
