/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9814 $
 * $Date: 2020-04-30 16:26:03 +0200 (Thu, 30 Apr 2020) $
 */
package at.lcm.symspace.gui.views;

import java.util.LinkedList;

import javax.swing.JComponent;
import javax.swing.JTabbedPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Tabbed view for different views
 *
 * @author Werner Koppelsttter
 */
public class TabView extends View {

	/** List of views */
	LinkedList<ViewItem> views;

	/** Tabbed-pane */
	JTabbedPane tab = null;

	/**
	 * Create tabbed view for view-manager
	 * 
	 * @param manager
	 *            View-manger
	 */
	public TabView(ViewManager<?> manager) {
		super(manager);
		this.views = new LinkedList<ViewItem>();
		this.tab = new JTabbedPane();
		this.tab.addChangeListener(new ChangeListener() {

			@Override
			public void stateChanged(ChangeEvent e) {
				TabView.this.refresh();
			}
		});
	}

	@Override
	public JComponent getComponent() {
		return this.tab;
	}

	/**
	 * Add view to list
	 * 
	 * @param item
	 *            View
	 */
	public void addView(ViewItem item) {
		this.addView(item, this.views.size());
	}

	/**
	 * Add view to list
	 * 
	 * @param item
	 *            View
	 * @param index
	 *            Index to insert
	 */
	public void addView(ViewItem item, int index) {
		this.views.add(index, item);
		this.tab.insertTab(item.getName(), null, this.manager.getComponent(item), null, index);
		this.refresh();
	}

	/**
	 * Remove view from list
	 * 
	 * @param item
	 *            View
	 */
	public void removeView(ViewItem item) {
		int index;

		index = this.views.indexOf(item);
		this.views.remove(index);
		this.tab.remove(index);
		this.refresh();
	}

	@Override
	public void refresh() {
		int index;
		index = this.tab.getSelectedIndex();
		if (index != -1) {
			while ((index > 0) && this.views.get(index).getName().trim().isEmpty()) {
				index--;
			}
			while ((index < this.tab.getTabCount()) && this.views.get(index).getName().trim().isEmpty()) {
				index++;
			}
			if ((index != this.tab.getSelectedIndex()) && (index < this.tab.getTabCount())) {
				this.tab.setSelectedIndex(index);
			}

			this.manager.refresh(this.views.get(this.tab.getSelectedIndex()));
		}
	}

	@Override
	public View getView(ViewItem item) {
		View view = null;

		for (ViewItem i : this.views) {
			if (i == item) {
				view = this;
			}
		}

		return view;
	}

	/**
	 * Select displayed index
	 * 
	 * @param index
	 */
	public void selectIndex(int index) {
		if ((index > 0) && (index < this.tab.getTabCount())) {
			this.tab.setSelectedIndex(index);
		}
	}

	@Override
	public void setEnabled(boolean aFlag) {
		for (int k = 0; k < this.tab.getTabCount(); k++) {
			this.tab.getComponentAt(k).setEnabled(aFlag);
		}
	}
}
