/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9868 $
 * $Date: 2020-06-04 12:11:54 +0200 (Thu, 04 Jun 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.structure.client.CDependItem;
import at.lcm.bu21.magopt.structure.rmi.RDependItem;

/**
 * Structure for dependency-items
 * 
 * @author Werner Koppelsttter
 */
public class DependItem implements IDependItem {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Variable description */
	private PortDesc var = new PortDesc();

	/** Name of assigned Parameter */
	private String param = "";

	/** Reference to the assigned Parameter */
	private Parameter handle = null;

	/**
	 * Create empty item
	 */
	protected DependItem() {

	}

	/**
	 * Create item with description
	 * 
	 * @param var
	 *            Description of item
	 */
	public DependItem(PortDesc var) {
		this(var, "");
	}

	/**
	 * Create item
	 * 
	 * @param var
	 *            Description of item
	 * @param param
	 *            Parameter or item
	 */
	public DependItem(PortDesc var, String param) {
		this.var = var;
		this.setParam(param);
	}

	/**
	 * Create a deep clone of given parameter
	 * 
	 * @return Deep clone of parameter
	 * @throws CloneException
	 */
	public DependItem deepClone() throws CloneException {
		return (DependItem) Clone.deepClone(this);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		DependItem other = (DependItem) obj;
		if (this.var == null) {
			if (other.var != null) {
				return false;
			}
		} else if (!this.var.equals(other.var)) {
			return false;
		}
		if (this.param == null) {
			if (other.param != null) {
				return false;
			}
		} else if (!this.param.equals(other.param)) {
			return false;
		}
		return true;
	}

	/**
	 * Clone item
	 * 
	 * @return cloned object
	 * @throws CloneException
	 */
	DependItem cloneDepend() throws CloneException {
		DependItem clone = new DependItem();

		clone.var = (PortDesc) Clone.deepClone(this.var);
		clone.param = this.param;

		return clone;
	}

	@Override
	public String getVarName() {
		return this.var.getName();
	}

	@Override
	public PortDesc getVarDesc() {
		return this.var;
	}

	/**
	 * Set Description of DependItem
	 * 
	 * @param var
	 *            Description
	 */
	protected void setVarDesc(PortDesc var) {
		this.var = var;
	}

	@Override
	public String getParam() {
		return this.param;
	}

	/**
	 * Set parameter of DependItem
	 * 
	 * @param param
	 *            Parameter
	 */
	protected void setParam(String param) {
		this.param = param;
	}

	@Override
	public Parameter getHandle() {
		return this.handle;
	}

	/**
	 * Set handle of DependItem
	 * 
	 * @param handle
	 */
	protected void setHandle(Parameter handle) {
		this.handle = handle;
	}

	/**
	 * Read structure from xml-file
	 * 
	 * @param xmlReader
	 *            xml-reader
	 * @throws Exception
	 */
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, var;

		xmlReader.next();

		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("var")) {
				xmlReader.next();
				var = "";
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					var += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
				// FIXME DEPEND
				this.var = new PortDesc(var, "", "?", true);

			} else if (element.equals("param")) {
				xmlReader.next();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.param += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            xml-writer
	 * @throws Exception
	 */
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("var");
		if (this.var != null) {
			// FIXME DEPEND
			xmlWriter.writeCharacters(this.getVarName());
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeStartElement("param");
		if (this.param != null) {
			xmlWriter.writeCharacters(this.param);
		}
		xmlWriter.writeEndElement();
	}

	/**
	 * Return string representation of item
	 * 
	 * @return String representation
	 */
	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("DependItem");
		str.append("\n\tVar        : ");
		str.append(this.var);
		str.append("\n\tParam      : ");
		str.append(this.param);
		str.append("\n\tIdentifier : ");
		if (this.handle != null) {
			str.append(this.handle.getIdentifier());
		}

		return str.toString();
	}

	/**
	 * Get RMI-adapter of depend-item
	 * 
	 * @return RMI-adapter
	 * @throws RemoteException
	 */
	public IDependItem getRMIImpl() throws RemoteException {
		return new RDependItem(this);
	}

	@Override
	public IDependItem getClientImpl() {
		return new CDependItem(this);
	}
}
