/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.Serializable;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Map;

/**
 * Interface of depend
 * 
 * @author Werner Koppelsttter
 */
public interface IDepend extends Remote, Serializable, Cloneable {

	/**
	 * Get imports of parameter
	 * 
	 * @return imports of parameter
	 * @throws RemoteException
	 */
	public IDependItem[] getImports() throws RemoteException;

	/**
	 * Get exports of parameter
	 * 
	 * @return exports of parameter
	 * @throws RemoteException
	 */
	public IDependItem[] getExports() throws RemoteException;

	/**
	 * Set variable-list of the imports
	 *
	 * @param vars
	 *            list of variables
	 * @throws RemoteException
	 */
	public void setImportVars(PortDesc[] vars) throws RemoteException;

	/**
	 * Set variable-list of the exports
	 *
	 * @param vars
	 *            list of variables
	 * @throws RemoteException
	 */
	public void setExportVars(PortDesc[] vars) throws RemoteException;

	/**
	 * Assign import Parameter to the given variable list
	 * 
	 * @param vars
	 *            list of assignments
	 * @throws RemoteException
	 */
	public void setImportAssignments(String[][] vars) throws RemoteException;

	/**
	 * Assign import Parameter to the given variable list
	 * 
	 * @param vars
	 *            map of assignments
	 * @throws RemoteException
	 */
	public void setImportAssignmentsMap(Map<String, String> vars)
			throws RemoteException;

	/**
	 * Assign export Parameter to the given variable list
	 * 
	 * @param vars
	 *            list of assignments
	 * @throws RemoteException
	 */
	public void setExportAssignments(String[][] vars) throws RemoteException;

	/**
	 * Assign export Parameter to the given variable list
	 * 
	 * @param vars
	 *            map of assignments
	 * @throws RemoteException
	 */
	public void setExportAssignmentsMap(Map<String, String> vars)
			throws RemoteException;

	/**
	 * Get number of depends of parameter
	 * 
	 * @param direction
	 *            Direction of depends
	 * @return imports of parameter
	 * @throws RemoteException
	 */
	public int getDependsLength(DependDirection direction)
			throws RemoteException;

	/**
	 * Get depend of parameter
	 * 
	 * @param direction
	 *            Direction of depends
	 * @param index
	 *            Index of depend
	 * @return imports of parameter
	 * @throws RemoteException
	 */
	public IDependItem getDepend(DependDirection direction, int index)
			throws RemoteException;

	/**
	 * Assign Parameter to the given variable list
	 * 
	 * @param var
	 *            Name of variable
	 * @param param
	 *            Name of parameter to assign
	 * @param direction
	 *            flag for import/export
	 * @throws RemoteException
	 */
	public void setAssignment(String var, String param,
			DependDirection direction) throws RemoteException;

	/**
	 * Get client-implementation of interface
	 * 
	 * @return Client-implementation
	 * @throws RemoteException
	 */
	public IDepend getClientImpl() throws RemoteException;
}
