/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.Pair;
import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.structure.event.EventLog;

/**
 * Parameter-data structure for all datasets
 * 
 * @author Werner Koppelsttter
 */
public class ParamDataStruct implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Index or current dataset */
	private int index = 0;

	/** Next number for Parameter-ID */
	private long nextPId = 0;

	/** Next number for dataset-ID */
	private long nextDsId = 0;

	/** ID for each dataset */
	private ArrayList<String> dsId = new ArrayList<String>();

	/** Dataset with parameter data */
	private ArrayList<Dataset> dataset = new ArrayList<Dataset>();

	/**
	 * Create data structure
	 */
	public ParamDataStruct() {
		String dsId;

		dsId = this.getNextDsId();
		this.dsId.add(dsId);
		this.dataset.add(new Dataset(dsId));
	}

	/**
	 * Check if selected datasets are equal
	 * 
	 * @param obj
	 *            Object to check if equal
	 * @param link
	 *            List of linked parameter-identifier
	 * @param thisIndex
	 *            Index of dataset of this object
	 * @param otherIndex
	 *            Index of dataset of other object
	 * @return If datasets are equal
	 */
	public boolean equals(Object obj, ArrayList<Pair<String, String>> link,
			int thisIndex, int otherIndex) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamDataStruct other = (ParamDataStruct) obj;
		if (this.dataset == null) {
			if (other.dataset != null) {
				return false;
			}
		} else {
			if ((thisIndex < 0) || (this.dataset.size() <= thisIndex)) {
				return false;
			}
			if ((otherIndex < 0) || (other.dataset.size() <= otherIndex)) {
				return false;
			}
			if (!this.dataset.get(thisIndex)
					.equals(other.dataset.get(otherIndex), link)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Clone parameter. Prepare to add cloned dataset
	 * 
	 * @param clone
	 */
	void cloneParam(ParamDataStruct clone) {
		clone.nextPId = this.nextPId;
		clone.dataset.clear();
		clone.dsId.clear();
	}

	/**
	 * Cleanup datasets
	 * 
	 * @param proj
	 *            Project to cleanup
	 */
	void cleanup(Project proj) {
		// cleanup directories
		if (proj.getAllFilesDirectory(false).exists()) {
			for (File f : proj.getAllFilesDirectory().listFiles()) {
				if (!f.getName().startsWith("_")) {
					if (!this.dsId.contains(f.getName())) {
						FileUtils.delete(f);
					}
				}
			}
		}

		// cleanup dataset
		for (int k = 0; k < this.dataset.size(); k++) {
			this.dataset.get(k).cleanup(proj, proj.getFilesDirectory(k, false));
		}
	}

	/**
	 * Move dataset
	 * 
	 * @param from
	 *            Move from index
	 * @param to
	 *            Move to index
	 */
	void move(int from, int to) {

		if ((0 <= from) && (from < this.dsId.size()) && (0 <= to)
				&& (to < this.dsId.size())) {
			if (to < from) {
				to--;
			}
			this.dsId.add(to, this.dsId.remove(from));
			this.dataset.add(to, this.dataset.remove(from));
		}
	}

	/**
	 * Add dataset to structure
	 * 
	 * @param ds
	 *            Dataset to add
	 * @throws CloneException
	 */
	void addParamDataset(Dataset ds) throws CloneException {
		ds.setDsId(this.getNextDsId());
		this.dsId.add(ds.getDsId());
		this.dataset.add(ds);

	}

	/**
	 * Get length (number) of datasets
	 * 
	 * @return Length
	 */
	int getLength() {
		return this.dataset.size();
	}

	/**
	 * Get current index
	 * 
	 * @return Index
	 */
	int getIndex() {
		if (this.index >= this.getLength()) {
			this.index = this.getLength() - 1;
		}
		return this.index;
	}

	/**
	 * Set current index
	 * 
	 * @param index
	 *            Index
	 */
	void setIndex(int index) {
		this.index = index;
	}

	/**
	 * Get current dataset
	 * 
	 * @return Current dataset
	 */
	Dataset getDataset() {
		return this.getDataset(this.index);
	}

	/**
	 * Get dataset of index
	 * 
	 * @param index
	 *            Index of dataset
	 * @return Dataset
	 */
	Dataset getDataset(int index) {
		Dataset ds = null;

		if ((index >= 0) && (index < this.getLength())) {
			ds = this.dataset.get(index);
		}

		return ds;
	}

	/**
	 * Get next dataset-ID
	 * 
	 * @return Next dataset-ID
	 */
	private String getNextDsId() {
		return Long.toString(this.nextDsId++, 36);
	}

	/**
	 * Get next parameter-ID
	 * 
	 * @return Next parameter-ID
	 */
	String getNextPId() {
		return Long.toString(this.nextPId++, 36);
	}

	/**
	 * Duplicate dataset
	 * 
	 * @param directory
	 *            Files-directory of project
	 * @param index
	 *            Index of dataset to clone
	 * @return duplicated dataset
	 * @throws CloneException
	 * @throws IOException
	 */
	Dataset duplicate(File directory, int index)
			throws CloneException, IOException {
		Dataset ds;
		File source, target;

		ds = this.getDataset(index);
		if (ds != null) {
			ds = ds.deepClone();
			source = new File(directory, ds.getDsId());
			ds.setDsId(this.getNextDsId());
			target = new File(directory, ds.getDsId());
			this.dsId.add(ds.getDsId());
			this.dataset.add(ds);

			if (source.exists()) {
				if (target.exists()) {
					FileUtils.delete(target);
				}
				FileUtils.copy(source, target, true);
			}
		}

		return ds;
	}

	/**
	 * Reduce dataset
	 * 
	 * @param directory
	 *            Files-directory of project
	 * @param index
	 *            Index of dataset
	 * @return Removed Dataset
	 */
	Dataset reduce(File directory, int index) {
		Dataset ds;
		File dir;

		ds = this.getDataset(index);
		if (ds != null) {
			this.dsId.remove(index);
			this.dataset.remove(index);
			dir = new File(directory, ds.getDsId());
			if (dir.exists()) {
				FileUtils.delete(dir);
			}
		}

		return ds;
	}

	/**
	 * Read object from xml-file
	 * 
	 * @param xmlReader
	 *            XMLReader to read xml-file
	 * @throws Exception
	 */
	public void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;
		String str;
		Dataset ds;
		EventLog log, oldLog;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("CurIndex")) {
				xmlReader.next();
				this.setIndex(Integer.parseInt(xmlReader.getText()) - 1);
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("NextPID")) {
				xmlReader.next();
				this.nextPId = Long.parseLong(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("NextDSID")) {
				xmlReader.next();
				this.nextDsId = Long.parseLong(xmlReader.getText());
				xmlReader.next();
				xmlReader.next();
			} else if (element.equals("DSID")) {
				xmlReader.next();
				this.dsId.clear();
				while (xmlReader
						.getEventType() == XMLStreamConstants.START_ELEMENT) {
					xmlReader.next();
					str = new String();
					while (xmlReader
							.getEventType() != XMLStreamConstants.END_ELEMENT) {
						str += xmlReader.getText();
						xmlReader.next();
					}
					this.dsId.add(str);
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("DS")) {
				xmlReader.next();
				oldLog = this.dataset.get(0).getEventLog();
				this.dataset.clear();
				while (xmlReader
						.getEventType() == XMLStreamConstants.START_ELEMENT) {
					xmlReader.next();
					ds = new Dataset("");
					this.dataset.add(ds);
					ds.xml2obj(xmlReader);
					xmlReader.next();
				}
				xmlReader.next();

				// transfer events
				ds = this.getDataset();
				if (ds != null) {
					log = ds.getEventLog();
					for (Event e : oldLog.getBuffer()) {
						log.add(e);
					}
				}

			} else {
				throw new Exception(
						"Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	/**
	 * Write object to xml-file
	 * 
	 * @param xmlWriter
	 *            XMLWriter to write xml-file
	 * @throws Exception
	 */
	public void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamDataStruct");

		// CurIndex
		xmlWriter.writeStartElement("CurIndex");
		xmlWriter.writeCharacters(Integer.toString(this.getIndex() + 1));
		xmlWriter.writeEndElement();

		// NextID
		xmlWriter.writeStartElement("NextPID");
		xmlWriter.writeCharacters(Long.toString(this.nextPId));
		xmlWriter.writeEndElement();

		// NextID
		xmlWriter.writeStartElement("NextDSID");
		xmlWriter.writeCharacters(Long.toString(this.nextDsId));
		xmlWriter.writeEndElement();

		// DatasetID
		xmlWriter.writeStartElement("DSID");
		for (String dsId : this.dsId) {
			xmlWriter.writeStartElement("item");
			xmlWriter.writeCharacters(dsId);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

		// Dataset
		xmlWriter.writeStartElement("DS");
		for (Dataset ds : this.dataset) {
			xmlWriter.writeStartElement("item");
			ds.obj2xml(xmlWriter);
			xmlWriter.writeEndElement();
		}
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

}
