/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10035 $
 * $Date: 2020-10-22 13:14:36 +0200 (Thu, 22 Oct 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.Desktop;
import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamDirectory;
import at.lcm.bu21.magopt.structure.rmi.RParamDirectory;

/**
 * Parameter representation of directory
 * 
 * @author Werner Koppelsttter
 */
public class ParamDirectory extends Parameter implements IParamDirectory {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Reference location for relative directory from constant directory */
	private Location location = Location.ABSOLUTE;

	/** Directory from constant parameter */
	private String directory = "";

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.add(0, "Directory Info");
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamDirectory() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamDirectory(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamDirectory(String name, String description) throws Exception {
		this(name, description, "");
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param directory   Directory
	 * @throws Exception
	 */
	public ParamDirectory(String name, String description, String directory) throws Exception {
		super(name, description);
		this.setAbsoluteDirectoryPath(directory);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamDirectory other = (ParamDirectory) obj;
		if (this.location == null) {
			if (other.location != null) {
				return false;
			}
		} else if (!this.location.equals(other.location)) {
			return false;
		}
		if (this.directory == null) {
			if (other.directory != null) {
				return false;
			}
		} else if (!this.directory.equals(other.directory)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamDirectory deepClone() throws CloneException {
		return (ParamDirectory) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamDirectory) {
			ParamDirectory other = (ParamDirectory) clone;

			other.location = this.location;
			other.directory = this.directory;
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamDirectory) {
			ParamDirectory other = (ParamDirectory) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.LOCATION);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.LOCATION, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.LOCATION);
			}

			data = dsThis.getValue(this.getPId(), ParamSetKey.DIRECTORY);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.DIRECTORY, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.DIRECTORY);
			}
		}
	}

	@Override
	public void setConstant(boolean constant) {
		Dataset ds;
		TString value;
		Location loc;
		String dir;

		if (this.getConstant() != constant) {
			if (constant) {
				this.location = Location.ABSOLUTE;
				this.directory = "";

				// define constant values
				ds = this.getDataset();
				if (ds != null) {
					value = (TString) ds.getValue(this.getPId(), ParamSetKey.LOCATION);
					if (value != null) {
						this.location = Location.valueOf(value.getString());
					}

					value = (TString) ds.getValue(this.getPId(), ParamSetKey.DIRECTORY);
					if (value != null) {
						this.directory = value.getString();
					}
				}

				// check changes and update
				for (int i = 0; i < this.getLength(); i++) {
					dir = "";
					loc = Location.ABSOLUTE;
					ds = this.getDataset(i);
					if (ds != null) {
						value = (TString) ds.getValue(this.getPId(), ParamSetKey.LOCATION);
						if (value != null) {
							loc = Location.valueOf(value.getString());
						}

						value = (TString) ds.getValue(this.getPId(), ParamSetKey.DIRECTORY);
						if (value != null) {
							dir = value.getString();
						}
					}

					if (!this.location.equals(loc) || !this.directory.equals(dir)) {
						this.setModified(i, true);
					}

					if (ds != null) {
						ds.rmValue(this.getPId(), ParamSetKey.LOCATION);
						ds.rmValue(this.getPId(), ParamSetKey.DIRECTORY);
					}
				}

				// Check location (not Param or Dataset Location for constant
				// directory)
				if (!this.location.isConstantPossible()) {
					this.location = Location.PROJECT;
					this.directory = FileUtils.getFile(
							FileUtils.getRelativePath(Location.PROJECT.getPath(this), Location.PARAM.getPath(this)),
							this.directory).getAbsolutePath();
				}
			} else {

				// update values
				for (int i = 0; i < this.getLength(); i++) {
					ds = this.getDataset(i);
					if (ds != null) {
						ds.setValue(this.getPId(), ParamSetKey.LOCATION, new TString(this.location.name()));
						ds.setValue(this.getPId(), ParamSetKey.DIRECTORY, new TString(this.directory));
					}
				}
				this.location = Location.ABSOLUTE;
				this.directory = "";
			}
		}
		super.setConstant(constant);
	}

	@Override
	public TData getData(int index) {
		return new TString(this.getAbsoluteDirectoryPath(index));
	}

	@Override
	public void setData(TData data) throws Exception {
		if (data instanceof TString) {
			this.setAbsoluteDirectoryPath(((TString) data).getString());
		} else {
			throw new Exception("Only Strings are allowed to set to " + this.getClass().getSimpleName());
		}
	}

	@Override
	public Location getLocation() {
		return this.getLocation(this.getIndex());
	}

	/**
	 * Get reference location of dataset
	 * 
	 * @param index Index of dataset
	 * @return Location
	 */
	public Location getLocation(int index) {
		Dataset ds;
		TData value;
		Location loc = Location.ABSOLUTE;

		if (this.getConstant()) {
			loc = this.location;
		} else {
			ds = this.getDataset(index);
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.LOCATION);
				if (value != null) {
					loc = Location.valueOf(((TString) value).getString());
				}
			}
		}

		return loc;
	}

	/**
	 * Set relative location of directory
	 * 
	 * @param location  Location
	 * @param setModify Set modify if location changed
	 */
	private void setLocation(Location location, boolean setModify) {
		Dataset ds;
		TData value;
		Location loc = Location.ABSOLUTE;

		if (this.getConstant()) {
			if (!location.equals(this.location)) {
				this.location = location;
				if (setModify) {
					for (int i = 0; i < this.getLength(); i++) {
						this.setModified(i, true);
					}
				}
			}
		} else {
			ds = this.getDataset();
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.LOCATION);
				if (value != null) {
					loc = Location.valueOf(((TString) value).getString());
				}
				if (!location.equals(loc)) {
					ds.setValue(this.getPId(), ParamSetKey.LOCATION, new TString(location.name()));
					if (setModify) {
						this.setModified(true);
					}
				}
			}
		}
	}

	@Override
	public void changeLocation(Location location) {
		String absPath, relPath, locPath;

		if (!this.getLocation().equals(location)) {
			absPath = this.getAbsoluteDirectoryPath();
			if (location == Location.ABSOLUTE) {
				this.setLocation(location, false);
				this.setRelDirPath(absPath, false);
			} else {
				locPath = location.getPath(this);
				if (locPath != null) {
					relPath = FileUtils.getRelativePath(locPath, absPath);
					if (!relPath.contains(":") && !relPath.startsWith("\\\\")) {
						this.setLocation(location, false);
						this.setRelDirPath(relPath, false);
					}
				}
			}
		}
	}

	/**
	 * Get location-path
	 * 
	 * @return Location-path
	 */
	public String getLocationPath() {
		return this.getLocationPath(this.getIndex());
	}

	/**
	 * Get location-path
	 * 
	 * @param index Index of dataset
	 * @return Location-path
	 */
	public String getLocationPath(int index) {
		return this.getLocation(index).getPath(this);
	}

	@Override
	public String getRelDirPath() {
		return this.getRelDirPath(this.getIndex());
	}

	/**
	 * Get relative directory of dataset
	 * 
	 * @param index Index of dataset
	 * @return Relative directory
	 */
	public String getRelDirPath(int index) {
		Dataset ds;
		TData value;
		String relDir = "";

		if (this.getConstant()) {
			relDir = this.directory;
		} else {
			ds = this.getDataset(index);
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.DIRECTORY);
				if (value != null) {
					relDir = ((TString) value).getString();
				}
			}
		}

		return relDir;
	}

	@Override
	public void setRelDirPath(String relPath) {
		this.setRelDirPath(relPath, true);
	}

	/**
	 * Set relative path of current dataset
	 * 
	 * @param relPath   Relative path
	 * @param setModify Set modify if location changed
	 */
	private void setRelDirPath(String relPath, boolean setModify) {
		Dataset ds;
		TString value;

		// check string
		if (relPath == null) {
			relPath = "";
		} else {
			relPath = FileUtils.trimPath(relPath);
		}

		// set directory
		if (this.getConstant()) {
			if (!this.directory.equals(relPath)) {
				this.directory = relPath;
				if (setModify) {
					for (int i = 0; i < this.getLength(); i++) {
						this.setModified(i, true);
					}
				}
			}
		} else {
			ds = this.getDataset();
			if (ds != null) {
				value = (TString) ds.getValue(this.getPId(), ParamSetKey.DIRECTORY);
				if ((value == null) || (!value.getString().equals(relPath))) {
					ds.setValue(this.getPId(), ParamSetKey.DIRECTORY, new TString(relPath));
					if (setModify) {
						this.setModified(true);
					}
				}
			}
		}
	}

	/**
	 * Get absolute path to directory
	 * 
	 * @return Absolute path
	 */
	public String getAbsoluteDirectoryPath() {
		return this.getAbsoluteDirectoryPath(this.getIndex());
	}

	/**
	 * Get absolute path to directory of dataset
	 * 
	 * @param index Index of dataset
	 * @return Absolute path
	 */
	public String getAbsoluteDirectoryPath(int index) {
		return this.getAbsoluteDirectory(index).getPath();
	}

	/**
	 * Set absolute path to directory
	 * 
	 * @param dir Absolute path
	 */
	public void setAbsoluteDirectoryPath(String dir) {
		// check string
		if (dir == null) {
			dir = "";
		} else {
			dir = FileUtils.trimPath(dir);
		}

		if (!this.getAbsoluteDirectoryPath().equals(dir)) {
			this.updateLocation(dir, true);
		}
	}

	/**
	 * Update location of path
	 */
	public void updateLocation() {
		this.updateLocation(this.getAbsoluteDirectoryPath(), false);
	}

	/**
	 * Update location from absPath
	 * 
	 * @param absPath   Absolute path to set
	 * @param setModify Set modify flag is changed
	 */
	private void updateLocation(String absPath, boolean setModify) {
		String locPath;

		for (Location loc : Location.fileValues()) {
			locPath = loc.getPath(this);
			if (loc == Location.ABSOLUTE) {
				this.setLocation(loc, setModify);
				this.setRelDirPath(absPath, setModify);
				break;
			} else if (this.getConstant() && (!loc.isConstantPossible())) {
			} else if (locPath != null) {
				if (absPath.contains(locPath)) {
					this.setLocation(loc, setModify);
					this.setRelDirPath(absPath.substring(loc.getPath(this).length()), setModify);
					break;
				}
			}
		}
	}

	@Override
	public File getAbsoluteDirectory() {
		return this.getAbsoluteDirectory(this.getIndex());
	}

	/**
	 * Get absolute-path of directory
	 * 
	 * @param index Index of dataset
	 * @return absolute-path
	 */
	public File getAbsoluteDirectory(int index) {
		String dir, path;
		File file;

		path = this.getLocationPath(index);
		dir = this.getRelDirPath(index);

		if ((path == null) || path.isEmpty()) {
			file = new File(dir);
		} else {
			file = new File(path, dir);
		}

		return FileUtils.trimPath(file);
	}

	@Override
	public void setAbsoluteDirectory(File dir) {
		String path;

		if (dir == null) {
			path = "";
		} else {
			path = dir.getPath();
		}

		this.setAbsoluteDirectoryPath(path);
	}

	@Override
	public void show() {
		File file;

		file = new File(((TString) this.getData()).getString());

		if (file.exists()) {
			if (Desktop.isDesktopSupported()) {
				try {
					Desktop.getDesktop().open(file);
				} catch (IOException e) {
					// FIXME winopen
					this.warning("Unable to open (Exception while open)");
				}
			} else {
				this.warning("Unable to open (Unsupported)");
			}
		} else {
			this.warning("Unable to open (file doesn't exist)");
		}

	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamDirectory\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tDirectory   : " + this.getAbsoluteDirectory());

		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, str;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("Parameter")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Location")) {
				xmlReader.next();
				str = "";
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					str += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
				this.location = Location.valueOf(str);
			} else if (element.equals("Directory")) {
				xmlReader.next();
				this.directory = "";
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.directory += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamDirectory");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Location
		xmlWriter.writeStartElement("Location");
		xmlWriter.writeCharacters(this.location.name());
		xmlWriter.writeEndElement();

		// Directory
		xmlWriter.writeStartElement("Directory");
		xmlWriter.writeCharacters(this.directory);
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	void reportTable(DocTexWriter content, int indent) {
		String path = "";
		Location loc;

		loc = this.getLocation();
		if (loc != Location.ABSOLUTE) {
			path = "[" + loc.name() + "]" + File.separator;
		}
		path += this.getRelDirPath();

		super.reportTable(content, indent);
		content.add("\\multicolumn{3}{>{\\centering}m{7.5cm}|}{\\mopath{");
		if (!this.getUsed()) {
			content.add("\\textcolor{gray}{");
		}
		content.addString(path.replace(File.separatorChar, '/'));
		if (!this.getUsed()) {
			content.add("}");
		}
		content.add("}}");
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tDirectory & \\moparam{");
		content.addString((this.getRelDirPath()).replace(File.separatorChar, '/'));
		content.add("} \\\\\n");

		content.add("\tLocation of directory & \\mopath{");
		content.addString(this.getLocation().toString());
		content.add("} \\\\\n");
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.LOCATION);
		list.add(ParamSetKey.DIRECTORY);

		return list;
	}

	@Override
	public IParamDirectory getRMIImpl() throws RemoteException {
		return new RParamDirectory(this);
	}

	@Override
	public IParamDirectory getClientImpl() {
		return new CParamDirectory(this);
	}
}
