/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9868 $
 * $Date: 2020-06-04 12:11:54 +0200 (Thu, 04 Jun 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;

import at.lcm.bu21.general.utils.Clone;

/**
 * Descriptor of dependency (import/export)
 * 
 * @author Werner Koppelsttter
 */
public class PortDesc implements Serializable, Cloneable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Name of dependency */
	protected String name = "";

	/** Unit of dependency */
	protected String unit = "";

	/** Dimension of dependency */
	protected String dim = "";

	/** Description of dependency */
	protected String desc = "";

	/** Is dependency optional */
	protected boolean optional = false;

	/** Is dependency optional */
	protected PortDesc[] struct = null;

	/**
	 * Create empty port-description
	 */
	public PortDesc() {
	}

	/**
	 * Create port-description
	 * 
	 * @param name
	 *            Name of depend
	 * @param description
	 *            Description of depend
	 * @param unit
	 *            Unit of depend
	 * @param optional
	 *            Is depend optional
	 */
	public PortDesc(String name, String description, String unit, boolean optional) {
		this(name, description, unit, optional, "");
	}

	/**
	 * Create port-description
	 * 
	 * @param name
	 *            Name of depend
	 * @param description
	 *            Description of depend
	 * @param unit
	 *            Unit of depend
	 * @param dimension
	 *            Dimension of depend
	 * @param optional
	 *            Is depend optional
	 */
	public PortDesc(String name, String description, String unit, boolean optional, String dimension) {
		this.setName(name);
		this.setDescription(description);
		this.setUnit(unit);
		this.setOptional(optional);
		this.setDimension(dimension);
	}

	/**
	 * Create port-description
	 * 
	 * @param name
	 *            Name of depend
	 * @param description
	 *            Description of depend
	 * @param unit
	 *            Unit of depend
	 * @param dimension
	 *            Dimension of depend
	 * @param optional
	 *            Is depend optional
	 * @param struct
	 *            Structure of depend
	 */
	public PortDesc(String name, String description, String unit, boolean optional, String dimension, PortDesc struct) {
		this(name, description, unit, optional, dimension);
		if (struct != null) {
			this.setStructure(new PortDesc[] { struct });
		}
	}

	/**
	 * Create port-description
	 * 
	 * @param name
	 *            Name of depend
	 * @param description
	 *            Description of depend
	 * @param unit
	 *            Unit of depend
	 * @param dimension
	 *            Dimension of depend
	 * @param optional
	 *            Is depend optional
	 * @param struct
	 *            Structure of depend
	 */
	public PortDesc(String name, String description, String unit, boolean optional, String dimension,
			PortDesc[] struct) {
		this(name, description, unit, optional, dimension);
		this.setStructure(struct);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		PortDesc other = (PortDesc) obj;
		if (this.desc == null) {
			if (other.desc != null) {
				return false;
			}
		} else if (!this.desc.equals(other.desc)) {
			return false;
		}
		if (this.dim == null) {
			if (other.dim != null) {
				return false;
			}
		} else if (!this.dim.equals(other.dim)) {
			return false;
		}
		if (this.name == null) {
			if (other.name != null) {
				return false;
			}
		} else if (!this.name.equals(other.name)) {
			return false;
		}
		if (this.optional != other.optional) {
			return false;
		}
		if (!Arrays.equals(this.struct, other.struct)) {
			return false;
		}
		if (this.unit == null) {
			if (other.unit != null) {
				return false;
			}
		} else if (!this.unit.equals(other.unit)) {
			return false;
		}
		return true;
	}

	/**
	 * Create a deep clone of given descriptor
	 * 
	 * @return Deep clone of descriptor
	 * @throws Exception
	 */
	public DependItem deepClone() throws Exception {
		return (DependItem) Clone.deepClone(this);
	}

	/**
	 * Get name of depend
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Set name of depend
	 * 
	 * @param name
	 *            Name
	 */
	public void setName(String name) {
		if (name != null) {
			this.name = name;
		} else {
			this.name = "";
		}
	}

	/**
	 * Get unit of depend
	 * 
	 * @return Unit
	 */
	public String getUnit() {
		return this.unit;
	}

	/**
	 * Set unit of depend
	 * 
	 * @param unit
	 *            Unit
	 */
	public void setUnit(String unit) {
		if (unit != null) {
			this.unit = unit;
		} else {
			this.unit = "";
		}
	}

	/**
	 * Get dimension of depend
	 * 
	 * @return Dimension
	 */
	public String getDimension() {
		return this.dim;
	}

	/**
	 * Set dimension of depend
	 * 
	 * @param dimension
	 *            Dimension
	 */
	public void setDimension(String dimension) {
		if (dimension != null) {
			this.dim = dimension;
		} else {
			this.dim = "";
		}
	}

	/**
	 * Get description of depend
	 * 
	 * @return Description
	 */
	public String getDescription() {
		return this.desc;
	}

	/**
	 * Set description of depend
	 * 
	 * @param description
	 *            Description
	 */
	public void setDescription(String description) {
		if (description != null) {
			this.desc = description;
		} else {
			this.desc = "";
		}
	}

	/**
	 * Get optional-flag of depend
	 * 
	 * @return Optional-flag
	 */
	public boolean getOptional() {
		return this.optional;
	}

	/**
	 * Set optional-flag of depend
	 * 
	 * @param optional
	 *            Optional-flag
	 */
	public void setOptional(boolean optional) {
		this.optional = optional;
	}

	/**
	 * Set optional-flag of depend
	 * 
	 * @param optional
	 *            Optional-flag
	 */
	public void setOptional(String optional) {
		if (optional != null) {
			this.optional = Boolean.parseBoolean(optional);
		} else {
			this.optional = false;
		}
	}

	/**
	 * Get structure of depend as list of depend-descriptors
	 * 
	 * @return Structure
	 */
	public PortDesc[] getStructure() {
		return this.struct;
	}

	/**
	 * Set structure of depend as list of depend-descriptors
	 * 
	 * @param structure
	 *            Structure
	 */
	public void setStructure(PortDesc[] structure) {
		this.struct = structure;
	}

	/**
	 * Set structure of depend as list of depend-descriptors
	 * 
	 * @param structure
	 *            Structure
	 */
	public void setStructureList(List<PortDesc> structure) {
		this.struct = structure.toArray(new PortDesc[0]);
	}

}
