/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10034 $
 * $Date: 2020-10-21 18:25:19 +0200 (Wed, 21 Oct 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.Desktop;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.Pair;
import at.lcm.bu21.general.utils.TexWriter;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.docu.FillTexReport;
import at.lcm.bu21.magopt.docu.ReportStructure;
import at.lcm.bu21.magopt.structure.DatasetEvent.DatasetType;
import at.lcm.bu21.magopt.structure.ReportParameter.ReportProperty;
import at.lcm.bu21.magopt.structure.client.CProject;
import at.lcm.bu21.magopt.structure.converter.XMLConverter;
import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.structure.event.EventLog;
import at.lcm.bu21.magopt.structure.jep.JepProxy;
import at.lcm.bu21.magopt.structure.rmi.RProject;
import at.lcm.bu21.magopt.structure.xsd.XMLValidator;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import jobmanager.JobDesc;
import jobmanager.JobStatus;
import magopt.utils.ComponentInfo;
import magopt.utils.Execute;
import magopt.utils.HtmlWriter;
import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import magopt.utils.SaveNotAlloowedException;

/**
 * Project parameter which manage updates, saving ...
 * 
 * @author Werner Koppelsttter
 */
public class Project extends ParamContainer implements IProject, MopAccess {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Version-history of MagOpt */
	private static final String[][] versionHistory = { { "V.0.0", "Start Version numbering" },
			{ "V.0.1", "Event update" }, { "V.0.2", "ParamFnc modified (add Container functionality)" },
			{ "V.0.3", "New Geometry structure, change FncWinding" }, { "V.0.4", "update structure (Depend, GeoData)" },
			{ "V.0.5", "update flags of constants" },
			{ "V.0.6", "Replace structure of periodic with periodic-structure" },
			{ "V.0.7", "update error flags and event-log" },
			{ "V.0.8", "Update flags, ParamFnc and rename ParamFieldFnc" }, { "V.0.9", "Update function structure" },
			{ "V.0.10", "Update combobox" }, { "V.0.11", "Update acess and enable" },
			{ "V.0.12", "Update Null and Lock" }, { "V.0.13", "Update Lock flag" },
			{ "V.0.14", "Update Material Parameter" }, { "V.0.15", "Update files-directories and DocFiles" },
			{ "V.0.16", "Updatae material" }, { "V.0.17", "Update material-diragram" }, { "V.0.18", "Update docFiles" },
			{ "V.0.19", "Remove SimRunning-flag" }, { "V.0.20", "Add ID for parameter and dataset" },
			{ "V.0.21", "Refill ID for parameter and dataset" }, { "V.0.22", "Change event structure" },
			{ "V.0.23", "Add report-parameters" }, { "V.0.24", "Rename dataset-  and paramter-direcotries" },
			{ "V.0.25", "ReRename dataset-  and paramter-direcotries" },
			{ "V.0.26", "Remove combobox-index and temorary-flag" }, { "V.0.27", "Partition of structure and data" },
			{ "V.0.28", "Bug ParamFile and ParamFormula" }, { "V.0.29", "Bug in cloneParam of ParamDatatStruct" },
			{ "V.0.30", "Optimizer structure changed" }, { "V.0.31", "Optimizer savepoints changed" },
			{ "V.0.32", "Optimizer round variable" },
			{ "V.0.33", "Add Linked-Field and public-interface to Container" },
			{ "V.0.34", "Optmizer split solver-settings and solver-state" },
			{ "V.0.35", "ParamDirectory/ParamFile changed" }, { "V.0.36", "Femm class-name changed" },
			{ "V.0.37", "Formula export changed" }, { "V.0.38", "Add component to container" },
			{ "V.0.39", "Add python function" }, { "V.0.40", "Add formula-type" },
			{ "V.0.41", "Change python function" }, { "V.0.42", "LinkedField export changed" }
//			,{ "V.0.43", "Add collector-container" } 
	};

	/** Project file */
	private MopFile mopFile = null;

	/** Handle to gui (matlab) */
	private Double guiHandle = -1.0;

	/** Flag whether events should print on console */
	private boolean printEvent = false;

	/** Report-parameters */
	private ReportParameter reportParam = new ReportParameter();

	/** DataStruct with parameter data */
	protected ParamDataStruct paramDataStruct = new ParamDataStruct();

	private ArrayList<String> templates = new ArrayList<String>();

	/** List of index-listeners */
	private transient LinkedList<DatasetListener> datasetListeners;

	/** Jep connection to python */
	private transient JepProxy jep = null;

	public void addTemplate(String template) {
		this.addTemplate(new File(template));
	}

	public void addTemplate(File template) {
		String relPath;

		relPath = FileUtils.getRelativePath(this.getMopFile().getParentFile(), template).getPath();
		if (!this.templates.contains(relPath)) {
			this.templates.add(relPath);
		}
	}

	public void rmTemplate(String template) {
		this.rmTemplate(new File(template));
	}

	public void rmTemplate(File template) {
		String relPath;
		relPath = FileUtils.getRelativePath(this.getMopFile().getParentFile(), template).getPath();
		this.templates.remove(relPath);
	}

	public int getNumTemplates() {
		return this.templates.size();
	}

	public File getTemplate(int index) {
		return FileUtils.trimPath(new File(this.getMopFile().getParentFile(), this.templates.get(index)));
	}

	/**
	 * List name of tabs
	 * 
	 * @return List of tab-names
	 */
	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.set(0, "Project Info");
		list.add(1, "Report");
		return list;
	}

	/**
	 * Constructor of Project
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        Project file
	 * @throws Exception
	 */
	public Project(String name, String description, String file) throws Exception {
		this(name, description, new MopFile(file));
	}

	/**
	 * Constructor for clone a project
	 * 
	 * @throws Exception
	 */
	Project() throws Exception {
		this("Clone", "Clone", new MopFile("clone.mop"));
	}

	/**
	 * Constructor of Project
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        Project file
	 * @throws Exception
	 */
	public Project(String name, String description, File file) throws Exception {
		this(name, description, new MopFile(file.getAbsolutePath()));
	}

	/**
	 * Constructor of Project
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        Project file
	 * @throws Exception
	 */
	public Project(String name, String description, MopFile file) throws Exception {
		super(name, description);

		this.mopFile = file;

		this.setPId(this.getParameterStruct().getNextPId());
	}

	/**
	 * Check if structure is equal
	 * 
	 * @param obj Object to compare
	 * @return Return if structure is equal
	 */
	public boolean equalsStructure(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Project other = (Project) obj;
		if (this.reportParam == null) {
			if (other.reportParam != null) {
				return false;
			}
		} else if (!this.reportParam.equals(other.reportParam)) {
			return false;
		}
		return true;
	}

	@Override
	public boolean equals(Object obj) {
		ArrayList<Pair<String, String>> link;

		if (!this.equalsStructure(obj)) {
			return false;
		}
		Project other = (Project) obj;
		if (this.paramDataStruct == null) {
			if (other.paramDataStruct != null) {
				return false;
			}
		} else {
			if (this.getLength() != other.getLength()) {
				return false;
			}

			link = this.getlinkPid(other);
			for (int k = 0; k < this.getLength(); k++) {
				if (!this.paramDataStruct.equals(other.paramDataStruct, link, k, k)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Check if structure and selected datasets are equal
	 * 
	 * @param obj        Object to check if equal
	 * @param thisIndex  Index of dataset of this project
	 * @param otherIndex Index of dataset of other project
	 * @return If selected datasets are equal
	 */
	public boolean equals(Object obj, int thisIndex, int otherIndex) {
		ArrayList<Pair<String, String>> link;

		if (!this.equalsStructure(obj)) {
			return false;
		}
		Project other = (Project) obj;
		if (this.paramDataStruct == null) {
			if (other.paramDataStruct != null) {
				return false;
			}
		} else {
			link = this.getlinkPid(other);
			if (!this.paramDataStruct.equals(other.paramDataStruct, link, thisIndex, otherIndex)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Get parameter-identifier-linked-list of this and other project
	 * 
	 * @param other Project to link
	 * @return List of pid-link
	 */
	private ArrayList<Pair<String, String>> getlinkPid(Project other) {
		Parameter[] thisParams, otherParams;
		ArrayList<Pair<String, String>> link;

		thisParams = this.getParameters(Parameter.class);
		otherParams = other.getParameters(Parameter.class);
		link = new ArrayList<Pair<String, String>>(thisParams.length);
		for (int k = 0; k < thisParams.length; k++) {
			link.add(new Pair<String, String>(thisParams[k].getPId(), otherParams[k].getPId()));
		}

		return link;
	}

	/**
	 * Get install-path
	 * 
	 * @return Install-path
	 */
	public static String getInstallPath() {
		return MagOptConfig.getHome();
	}

	/**
	 * Get Path of library
	 * 
	 * @return Library-path
	 */
	public static String getLibraryPath() {
		// TODO MagOptConfig.getLibraryPath()
		return new File(Project.getInstallPath(), "Library").getPath();
	}

	/**
	 * Get MagOpt property
	 * 
	 * @param name Name of property
	 * @return Property value
	 */
	public static String getProperty(String name) {
		return MagOptConfig.getConfig(name);
	}

	@Override
	public Project deepClone() throws CloneException {
		return (Project) Clone.deepClone(this);
	}

	// @Override
	// public Parameter cloneParam(Parameter clone, int index)
	// throws CloneException, InstantiationException,
	// IllegalAccessException {
	//
	// clone = super.cloneParam(clone, index);
	// clone.refresh(true);
	//
	// return clone;
	// }

	@Override
	public void close() {
		if (this.jep != null) {
			try {
				this.jep.close();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
			this.jep = null;
		}
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {

		super.cloneParam(clone);

		if (clone instanceof Project) {
			Project other = (Project) clone;

			other.printEvent = this.printEvent;
			other.reportParam = (ReportParameter) Clone.deepClone(this.reportParam);
		}
	}

	/**
	 * Get current program version
	 * 
	 * @return program version
	 */
	public static String getVersion() {
		return Project.versionHistory[Project.versionHistory.length - 1][0];
	}

	/**
	 * Get version history of program
	 * 
	 * @return version history
	 */
	public static String[][] getVersionHistory() {
		return Project.versionHistory;
	}

	/**
	 * Get project directory with dataset-files
	 * 
	 * @return Project-directory
	 */
	@Override
	public File getAllFilesDirectory() {
		return this.getAllFilesDirectory(true);
	}

	/**
	 * Get project directory with dataset-files
	 * 
	 * @param create Create directory if not exists
	 * @return Project-directory
	 */
	File getAllFilesDirectory(boolean create) {
		File dir;

		dir = this.mopFile.getFilesDirectory();
		if (create && !dir.exists()) {
			dir.mkdirs();
		}
		return dir;
	}

	@Override
	public File getFilesDirectory() {
		return this.getFilesDirectory(this.getIndex());
	}

	/**
	 * Get directory of dataset-files
	 * 
	 * @param create Create directory if not exists
	 * @return Dataset-directory
	 */
	File getFilesDirectory(boolean create) {
		return this.getFilesDirectory(this.getIndex(), create);
	}

	/**
	 * Get directory of dataset-files
	 * 
	 * @param index Index of directory
	 * @return Dataset-directory
	 */
	public File getFilesDirectory(int index) {
		return this.getFilesDirectory(index, true);
	}

	/**
	 * Get directory of dataset-files
	 * 
	 * @param index  Index of directory
	 * @param create Create directory if not exists
	 * @return Dataset-directory
	 */
	File getFilesDirectory(int index, boolean create) {
		File dir = null;
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			dir = new File(this.getAllFilesDirectory(create), ds.getDsId());
			if (create && !dir.exists()) {
				dir.mkdirs();
			}
		}

		return dir;
	}

	@Override
	public String getDSComment() {
		return this.getDSComment(this.getIndex());
	}

	/**
	 * Get comment of current dataset
	 * 
	 * @param index Index of directory
	 * @return Comment
	 */
	public String getDSComment(int index) {
		Dataset ds;
		String comment = "";

		ds = this.getDataset(index);
		if (ds != null) {
			comment = ds.getComment();
		}

		return comment;
	}

	@Override
	public String[] getAllDSComment() {
		Dataset ds;

		String[] list = new String[this.getLength()];
		for (int i = 0; i < this.getLength(); i++) {
			ds = this.getDataset(i);
			if (ds != null) {
				list[i] = ds.getComment();
			} else {
				list[i] = "";
			}
		}
		return list;
	}

	@Override
	public void setDSComment(String comment) {
		this.setDSComment(this.getIndex(), comment);
	}

	/**
	 * Set comment of current dataset
	 * 
	 * @param index   Index of directory
	 * @param comment Comment
	 */
	void setDSComment(int index, String comment) {
		Dataset ds;

		ds = this.getDataset(index);
		if (ds != null) {
			if (((comment == null) && (ds.getComment() != null)) || !comment.equals(ds.getComment())) {
				ds.setComment(comment);
				this.fireDatasetIndexChanged(this, index, index);
			}
		}
	}

	@Override
	public void setConstant(boolean constant) {
		if (constant) {
			this.warning("Set Project to constant is not allowed!");
		}
	}

	/**
	 * Get status of event-print
	 * 
	 * @return Event-print status
	 */
	public boolean getPrintEvent() {
		return this.printEvent;
	}

	/**
	 * Set status of event-print
	 * 
	 * @param printEvent Event-print status
	 */
	public void setPrintEvent(boolean printEvent) {
		this.printEvent = printEvent;
	}

	/**
	 * Get path of project-file
	 * 
	 * @return Project-path
	 */
	String getPath() {
		return this.mopFile.getParent();
	}

	@Override
	public String getFile() {
		return this.mopFile.getPath();
	}

	@Override
	public MopFile getMopFile() {
		return this.mopFile;
	}

	@Override
	public TData getData(int index) {
		return new TString(this.getFile());
	}

	@Override
	public double getGuiHandle() {
		return this.guiHandle;

	}

	@Override
	public void setGuiHandle(double guiHandle) {
		this.guiHandle = guiHandle;
	}

	/**
	 * Load project file
	 * 
	 * @param file Project-file
	 * @return Project
	 * @throws Exception
	 */
	public static Project load(String file) throws Exception {
		return Project.load(new MopFile(file));
	}

	/**
	 * Load project file
	 * 
	 * @param file Project-file
	 * @return Project
	 * @throws Exception
	 */
	public static Project load(File file) throws Exception {
		return Project.load(new MopFile(file));
	}

	/**
	 * Load project file
	 * 
	 * @param file Project-file
	 * @return Project
	 * @throws Exception
	 */
	public static Project load(MopFile file) throws Exception {
		Project project;

		project = new Project("Loader-ERROR", "", new File(file.getParent(), "error.mop"));

		if (!file.exists()) {
			project.error("File does not exists!");
		} else {
			try {
				// unzip component files
				file.expandComponent();

				// load project
				project.loadFile(file);
				project.mopFile = file;

				// refresh handle
				project.info("refresh project");
				project.refresh(true);

				project.info("Reading File successfully finished");
			} catch (Exception ex) {
				project.error(ex.getMessage(), ex.getStackTrace());
			}
		}
		return project;
	}

	/**
	 * Load project file
	 * 
	 * @param file Project-file
	 * @return Project
	 * @throws Exception
	 */
	public static Project onlyLoad(File file) throws Exception {
		Project project;
		MopFile mop;

		mop = new MopFile(file.getAbsolutePath());
		project = new Project("Loader-ERROR", "", new File(file.getParent(), "error.mop"));

		if (!mop.exists()) {
			project.error("File does not exists!");
		} else {
			try {
				project.loadFile(mop);
				project.mopFile = mop;

				project.warning("Project not refreshed");
				project.info("Reading File successfully finished");
			} catch (Exception ex) {
				project.error(ex.getMessage(), ex.getStackTrace());
			}
		}
		return project;
	}

	/**
	 * Delete project
	 * 
	 * @param mop Mop-File to delete
	 */
	public static void delete(String mop) {
		Project.delete(new MopFile(mop));
	}

	/**
	 * Delete project
	 * 
	 * @param mop Mop-File to delete
	 */
	public static void delete(File mop) {
		Project.delete(new MopFile(mop));
	}

	/**
	 * Delete project
	 * 
	 * @param mop Mop-File to delete
	 */
	public static void delete(MopFile mop) {
		mop.delete();
	}

	/**
	 * Load project into project
	 * 
	 * @param file project-file
	 * @throws Exception
	 */
	protected void loadFile(MopFile file) throws Exception {

		// update File
		XMLConverter xmlFile = new XMLConverter(file, Project.getVersion());
		int status = xmlFile.compareVersion();
		if (status < 0) {
			throw new IOException("File version is greater than program version");
		} else if (status > 0) {
			xmlFile.convert();
			this.info("Convert File " + xmlFile.getOldVersion() + " => " + xmlFile.getNewVersion());
			xmlFile = null;
		}

		// validation of File
		this.info("Validating file");
		new XMLValidator().validateMop(file, "Project.xml", "Project.xsd");
		// TODO Mop enthlt kein Project.xml -> ERROR
		// throw new IOException("Invalid mop/mopt file! " + file);

		// open File for reading
		this.info("Read file content");
		file.readStream(this, "Project.xml");
		// TODO Mop enthlt kein Project.xml -> ERROR
		// throw new IOException("Invalid mop/mopt file! " + file);

	}

	@Override
	public void readStream(InputStream is) throws Exception {

		// Open xml-reader
		XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
		XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(is, "utf-8");

		// read file
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		if ("Project".compareTo(xmlReader.getLocalName()) != 0) {
			throw new Exception("File is not a project-file");
		}
		this.xml2obj(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new IOException("Undefined error by reading file!");
		}

		// end file
		xmlReader.close();

	}

	@Override
	public void writeStream(OutputStream os) throws Exception {

		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance().createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");
		xmlWriter.writeComment(Project.getVersion());

		// write Project to File
		this.obj2xml(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();

	}

	/**
	 * Import template into project at given parameter
	 * 
	 * @param file       Template-file
	 * @param identifier Identifier of Parameter to insert
	 * @throws Exception
	 */
	public void importTemplate(String file, String identifier) throws Exception {
		this.importTemplate(new MopFile(file), identifier);
	}

	/**
	 * Import template into project at given parameter
	 * 
	 * @param file       Template-file
	 * @param identifier Identifier of Parameter to insert
	 * @throws Exception
	 */
	public void importTemplate(File file, String identifier) throws Exception {
		this.importTemplate(new MopFile(file), identifier);
	}

	/**
	 * Import template into project at given parameter
	 * 
	 * @param file       Template-file
	 * @param identifier Identifier of Parameter to insert
	 */
	public void importTemplate(MopFile file, String identifier) {
		Parameter obj = this.getObject(identifier);
		ParamContainer insert;

		if (!(obj instanceof ParamContainer)) {
			this.error("Unable to insert template to " + identifier);
		} else {
			insert = (ParamContainer) obj;
			try {
				if (!file.exists()) {
					this.error("File does not exists!");
				} else {
					this.importTemplate(Project.load(file), insert);
				}
			} catch (Exception ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
		}
	}

	/**
	 * Import template into project at given parameter
	 * 
	 * @param insert   Location to insert template
	 * @param template Template to insert
	 */
	public void importTemplate(Project template, ParamContainer insert) {
		Parameter[] params;
		Event[] buffer;

		try {
			// transfer event-log of template to current project
			buffer = template.getEventLog().getBuffer();
			for (int i = 0; i < buffer.length; i++) {
				buffer[i].setDescription("TEMPLATE: " + buffer[i].getDescription());
				this.getEventLog().add(buffer[i]);
			}

			// move all sub-parameter to current project
			if (template.getComponent() != null) {
				insert.addComponent();
				template.getComponent().cloneComponent(insert.getComponent());
			}
			params = template.getParam();
			for (int i = 0; i < params.length; i++) {
				insert.addParam(params[i].cloneParam());
			}
			this.refresh(true);
		} catch (Exception ex) {
			this.error(ex.getMessage(), ex.getStackTrace());
		}
	}

	/**
	 * Export component from given parameter
	 * 
	 * @param file           Component-file
	 * @param compIdentifier Identifier of component
	 * @param identifiers    Identifier-list of Parameter to export
	 * @param infos          Component infos
	 * @throws Exception
	 */
	public void exportComponentList(MopFile file, String compIdentifier, List<String> identifiers,
			final ComponentInfo infos) throws Exception {
		this.exportComponent(file, compIdentifier, identifiers.toArray(new String[0]), infos);
	}

	/**
	 * Export component from given parameter
	 * 
	 * @param file           Component-file
	 * @param compIdentifier Identifier of component
	 * @param identifiers    Identifier-list of Parameter to export
	 * @param infos          Component infos
	 * @throws Exception
	 */
	public void exportComponent(MopFile file, String compIdentifier, String[] identifiers, final ComponentInfo infos)
			throws Exception {
		Component comp;
		Project tmp;
		String relPath;
		File parent;
		ParamContainer container;
		boolean inside = false;
		String uIdent, cuIdent, relIdent = "";

		try {
			tmp = this.exportTemplateProject(file, identifiers);

			// external component
			container = (ParamContainer) this.getObject(compIdentifier);
			cuIdent = container.getIdentifier();
			for (String ident : identifiers) {
				if (!inside) {
					uIdent = this.getObject(ident).getIdentifier();
					if (cuIdent.equals(uIdent) || (cuIdent.startsWith(uIdent + "."))) {
						relIdent = uIdent.substring(0, uIdent.lastIndexOf("."));
						relIdent = cuIdent.substring(relIdent.length() + 1);
						inside = true;
					}
				}
			}
			if (!inside) {
				relIdent = ".";
				tmp.addComponent();
				container.getComponent().cloneComponent(tmp.getComponent());
				tmp.refresh();
			}
			tmp.save();

			// component info
			comp = ((ParamContainer) tmp.getObject(relIdent)).getComponent();
			infos.setHelp(comp.getHelpFile());
			infos.setImage(comp.getImage());

			// zip files
			for (Parameter p : tmp.getParameters(ParamContainer.class)) {
				comp = ((ParamContainer) p).getComponent();
				if (comp != null) {
					parent = file.getFilesDirectory();

					// component files
					for (File f : FileUtils.listFiles(comp.getFolder())) {
						if (f.isFile()) {
							relPath = FileUtils.getRelativePath(parent, f).getPath().replaceAll("\\\\", "/");
							tmp.getMopFile().writeFile(f, MopFile.componentTag + relPath);
						}
					}

					// linked files
					for (File f : comp.listLinkedFiles()) {
						if (f.isFile()) {
							relPath = FileUtils.getRelativePath(parent, f).getPath().replaceAll("\\\\", "/");
							tmp.getMopFile().writeFile(f, MopFile.componentTag + relPath);
						}

					}
				}

			}

			// write component infos
			tmp.getMopFile().writeStream(new MopAccess() {

				@Override
				public void writeStream(OutputStream os) throws Exception {
					infos.getProperties().store(os, "");
				}

				@Override
				public void readStream(InputStream is) throws Exception {

				}
			}, "Component.info");

			// close project
			tmp.close();
		} catch (Exception ex) {
			this.error(ex.getMessage(), ex.getStackTrace());
		}

	}

	/**
	 * Export template from given parameter
	 * 
	 * @param file        Template-file
	 * @param identifiers Identifier-list of Parameter to export
	 * @throws Exception
	 */
	public void exportTemplate(File file, String[] identifiers) throws Exception {
		this.exportTemplate(new MopFile(file), identifiers);
	}

	/**
	 * Export template from given parameter
	 * 
	 * @param file        Template-file
	 * @param identifiers Identifier-list of Parameter to export
	 * @throws Exception
	 */
	public void exportTemplate(String file, String[] identifiers) throws Exception {
		this.exportTemplate(new MopFile(file), identifiers);
	}

	/**
	 * Export template from given parameter
	 * 
	 * @param file        Template-file
	 * @param identifiers Identifier-list of Parameter to export
	 * @throws Exception
	 */
	public void exportTemplate(MopFile file, String[] identifiers) throws Exception {
		try {
			this.exportTemplateProject(file, identifiers).save();
		} catch (Exception ex) {
			this.error(ex.getMessage(), ex.getStackTrace());
		}

	}

	/**
	 * Export template from given parameter and retrun template
	 * 
	 * @param file        Template-file
	 * @param identifiers Identifier-list of Parameter to export
	 * @return Template
	 * @throws Exception
	 */
	private Project exportTemplateProject(MopFile file, String[] identifiers) throws Exception {
		Project template = null;
		Event[] buffer;
		int index = this.getIndex();

		template = new Project("Template", "Template", file);
		// copy parameter to template
		for (int i = 0; i < identifiers.length; i++) {
			template.addParam(this.getObject(identifiers[i]).cloneParam(null, index));
		}
		template.refresh(true);

		// transfer event-log of template to current project
		buffer = template.getEventLog().getBuffer();
		for (int i = 0; i < buffer.length; i++) {
			buffer[i].setDescription("TEMPLATE: " + buffer[i].getDescription());
			this.getEventLog().add(buffer[i]);
		}

		return template;
	}

	@Override
	public EventLog getEventLog() {
		Dataset ds;
		EventLog log = null;

		ds = this.getDataset();
		if (ds != null) {
			log = ds.getEventLog();
		}

		return log;
	}

	/**
	 * Refresh class-file location
	 */
	protected void refreshClasses() {
		ParamFunction func;
		Parameter[] pf;

		pf = this.getParameters(ParamFunction.class);
		for (Parameter param : pf) {
			func = (ParamFunction) param;
			try {
				func.refreshClass();
			} catch (Exception ex) {
				this.error(ex.getMessage(), ex.getStackTrace());
			}
		}
	}

	@Override
	public boolean isWriteProtected() {
		boolean protect = false;

		try {
			if (this.mopFile.exists()) {
				for (String name : this.mopFile.getFileNames()) {
					protect |= name.startsWith("GeneticOptimization/");
				}
			}
		} catch (Exception ex) {
			protect = true;
		}

		return protect;
	}

	@Override
	public void save() throws Exception {
		// refresh
		this.info("Refresh project");
		this.refreshClasses();
		this.refresh();
		this.cleanup();

		// save Project
		this.info("Save project");
		try {
			if (this.isWriteProtected()) {
				throw new SaveNotAlloowedException("Save Project with optimization is not allowed!");
			}
			this.mopFile.writeStream(this, "Project.xml");
			this.info("Save project successfully finished");
		} catch (Exception ex) {
			this.error(ex.getMessage(), ex.getStackTrace());
			throw (ex);
		}
	}

	@Override
	public void saveAs(String file) throws Exception {
		this.saveAs(new File(file));
	}

	@Override
	public void saveAs(String file, boolean override) throws Exception {
		this.saveAs(new File(file), override);
	}

	@Override
	public void saveAs(File file) throws Exception {
		this.saveAs(file, false);
	}

	@Override
	public void saveAs(File file, boolean override) throws Exception {
		MopFile sourceMop, destMop;
		File sourceFiles, destFiles, sourceData, destData;

		if (file.equals(this.mopFile)) {
			this.save();
		} else {
			sourceMop = this.mopFile;
			destMop = new MopFile(file);
			sourceData = sourceMop.getDataDirectory();
			destData = destMop.getDataDirectory();
			sourceFiles = sourceMop.getFilesDirectory();
			destFiles = destMop.getFilesDirectory();

			if (override) {
				FileUtils.delete(destMop);
				FileUtils.delete(destData);
				FileUtils.delete(destFiles);
			}

			if (destMop.exists() || destData.exists() || destFiles.exists()) {
				this.warning("Unable to save as. Destination-files already exists!");
			} else {
				destMop.getParentFile().mkdirs();
				if (sourceFiles.exists()) {
					FileUtils.copy(sourceFiles, destFiles);
				}
				if (sourceData.exists()) {
					FileUtils.copy(sourceData, destData);
				}
				this.mopFile = destMop;
				this.save();

				this.info("Project saveAs ('" + sourceMop.getPath() + "' -> '" + destMop.getPath() + "')");
			}
		}
	}

	/**
	 * Rename Project to given filename
	 * 
	 * @param filename Project-filename
	 */
	public void rename(String filename) {
		MopFile sourceMop, destMop;
		File sourceFiles, destFiles, sourceData, destData;

		sourceMop = this.mopFile;
		destMop = new MopFile(new File(sourceMop.getParent(), filename).getPath());
		sourceData = sourceMop.getDataDirectory();
		destData = destMop.getDataDirectory();
		sourceFiles = sourceMop.getFilesDirectory();
		destFiles = destMop.getFilesDirectory();

		if ((sourceMop.exists() && !sourceMop.canWrite()) || (sourceData.exists() && !sourceData.canWrite())
				|| (sourceFiles.exists() && !sourceFiles.canWrite())) {
			this.warning("Unable to rename. Source-files not able to write!");
		} else if (destMop.exists() || destData.exists() || destFiles.exists()) {
			this.warning("Unable to rename. Destination-files already exists!");
		} else {
			if (sourceMop.exists()) {
				sourceMop.renameTo(destMop);
			}
			this.mopFile = destMop;
			if (sourceData.exists()) {
				sourceData.renameTo(destData);
			}
			if (sourceFiles.exists()) {
				sourceFiles.renameTo(destFiles);
			}
			this.info("Project renamed ('" + sourceMop.getName() + "' -> '" + destMop.getName() + "')");
		}
	}

	/**
	 * Cleanup datasets and directories
	 */
	public void cleanup() {
		// cleanup datasets
		this.getParameterStruct().cleanup(this);

		// cleanup components
		Component.cleanup(this);
	}

	@Override
	public void check() {
		this.check(this.getParameters(Function.class), this.getParameters(Dependency.class));
	}

	/**
	 * Check structure which updates are required
	 * 
	 * @param functions    List of functions
	 * @param dependencies List of dependencies
	 */
	void check(Parameter[] functions, Parameter[] dependencies) {
		DependItem[] imports, exports;
		Function func;
		ParamMaterial mat;
		Parameter[] matList;
		int n;

		this.refresh(false, false, dependencies);

		// check material dependency
		matList = this.getParameters(ParamMaterial.class);
		for (Parameter p : matList) {
			mat = (ParamMaterial) p;

			imports = mat.getDepend().getImports();
			for (int j = 0; j < imports.length; j++) {
				// check if import is valid
				if (!mat.getModified() && (imports[j].getHandle() != null) && imports[j].getHandle().getModified()) {
					mat.setModified(true);
				}
			}
		}

		// check all functions
		for (Parameter p : functions) {
			// check imports of function are modified
			func = (Function) p;

			imports = func.getDepend().getImports();
			for (int j = 0; j < imports.length; j++) {
				// check if import is valid
				if ((imports[j].getHandle() != null) && imports[j].getHandle().getModified()) {
					// check for cycle (import is in export)
					exports = func.getDepend().getExports();
					n = 0;
					while ((n < exports.length) && (imports[j].getHandle() != exports[n].getHandle())) {
						n++;
					}
					if (n == exports.length) {
						func.setUpdateRequired(true);
					}
				}

				// check for incorrect imports
				if ((imports[j].getHandle() == null) && (imports[j].getParam() != null)
						&& (imports[j].getParam().length() != 0)) {
					String ident = ((Parameter) func).getIdentifier();
					((Parameter) func).warning("Import " + imports[j].getParam() + " in " + ident + " not found");
				}
			}

			// check for incorrect exports
			exports = func.getDepend().getExports();
			for (int j = 0; j < exports.length; j++) {
				if ((exports[j].getHandle() == null) && (exports[j].getParam() != null)
						&& (exports[j].getParam().length() != 0)) {
					String ident = ((Parameter) func).getIdentifier();
					((Parameter) func).warning("Export " + exports[j].getParam() + " in " + ident + " not found");
				}
			}
		}

		// clear modified flags
		this.execStructure(new ExecStructure() {
			@Override
			public void execute(Parameter param) {
				param.setModified(false);
			}
		});

		// FIXME setUncertain (refresh)
		this.refreshFlags();
	}

	@Override
	public int regenerate(Map<String, Object> args) {
		DependMatrix depend;
		String[] aims;
		int status, count, sizeA_Old, idx;
		Parameter handle;
		Function handleFnc;
		ArrayList<Boolean> updateStopped;
		String msg;

		// clear error flags
		this.execStructure(new ExecStructure() {
			@Override
			public void execute(Parameter param) {
				param.setError(false);
			}
		});

		// FIXME check ??
		this.refresh();
		depend = new DependMatrix(this);
		aims = (String[]) args.get("Aims");
		if (aims != null) {
			depend.reset(aims);
		}
		this.check();

		// FIXME regenerate
		count = 1;
		if (!depend.isEmpty()) {
			count = depend.size();

			updateStopped = new ArrayList<Boolean>();
			for (int k = 0; k < depend.size(); k++) {
				updateStopped.add(false);
			}
			sizeA_Old = 0;
			while (!depend.isEmpty() && (depend.size() != sizeA_Old)) { // adjency
																		// matrix
																		// is
																		// not
																		// empty
																		// and
																		// its
																		// size
																		// has
																		// changed
																		// since
																		// the
																		// last
																		// cycle
				sizeA_Old = depend.size();
				idx = 0;
				while (idx < depend.size()) {
					if (!depend.depends(idx)) { // function may be updated
												// if according column in
												// the adjancency matrix is
												// empty
						handle = depend.getFunction(idx);
						handleFnc = (Function) handle;
						if (handleFnc.getUpdateRequired()) {
							if (!updateStopped.get(idx)) {
								this.info("Updating " + handle.getName());
								// FIXME waitbar
								this.updateWaitbar(this.getGuiHandle(), count, depend.size());
								try {
									if (handleFnc.isEnable()) {
										if ((handleFnc instanceof ParamFunction)
												&& ((ParamFunction) handleFnc).getResetFlag()) {
											((ParamFunction) handleFnc).reset();
										}
										status = handleFnc.update(args);
									} else {
										this.warning("Function '" + handle.getName() + "' is not enabled!");
										status = 1;
									}
								} catch (Exception ex) {
									msg = ex.getMessage();
									if ((msg == null) || msg.isEmpty()) {
										msg = ex.getClass().getSimpleName();
									}
									handle.error(msg, ex.getStackTrace());
									status = 1;
								}
							} else { // function-update has already been
										// executed but was not successfull
								status = 1;
							}
						} else { // no update required
							status = 0;
						}
						if (status == 0) { // update succesful
							depend.remove(idx);
							updateStopped.remove(idx);
						} else { // update has not been successful
							updateStopped.set(idx, true);
							idx = idx + 1;
						}
					} else { // function may not be updated - inputs haven't
								// been updated yet
						idx = idx + 1;
					}
				}
			}
		}

		if (depend.isEmpty()) {
			this.info("Global Update completed successfully");
			status = 0;
		} else {
			this.info("Stopped Updates");
			status = 1;
		}

		// refresh handles and flags
		this.refresh();

		// FIXME Waitbar
		this.updateWaitbar(this.getGuiHandle(), count, depend.size());

		return status;
	}

	/**
	 * Update waitbar of Matlab gui
	 * 
	 * @param handle Handle of gui
	 * @param count  Value of waitbar
	 * @param max    Max value of waitbar
	 */
	// FIXME Waitbar
	private void updateWaitbar(double handle, int count, int max) {
		try {
			Class<?> c = Class.forName("at.lcm.bu21.magopt.utils.MatlabControl");
			Method m = c.getMethod("updateWaitbar", double.class, int.class, int.class);
			m.invoke(null, handle, count, max);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("Project\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tCurIndex    : " + (this.getIndex() + 1));
		str.append("\n\tLength      : " + this.getLength());
		str.append("\n\tDS-Comment  : " + this.getDSComment());
		str.append("\n\tParams      : ");

		for (Parameter p : this.getParam()) {
			str.append("\n\t\tName: " + p.getName());
		}

		return str.toString();
	}

	@Override
	public void duplicate() throws CloneException, IOException {
		this.duplicate(this.getIndex());
	}

	/**
	 * Duplicate given dataset to the end of list
	 * 
	 * @param index Index of dataset
	 * @throws CloneException
	 * @throws IOException
	 */
	public void duplicate(int index) throws CloneException, IOException {
		Dataset ds;

		ds = this.paramDataStruct.duplicate(this.getAllFilesDirectory(), index);
		if (ds != null) {
			this.fireDatasetAdded(this, this.getLength() - 1, this.getLength() - 1);
			this.setIndex(this.getLength() - 1);
		} else {
			this.warning("Unable to duplicate!");
		}
	}

	@Override
	public void reduce() {
		this.reduce(this.getIndex());
	}

	/**
	 * Reduce given dataset of list
	 * 
	 * @param index Index of dataset
	 */
	void reduce(int index) {
		Dataset ds;
		Hashtable<String, JobStatus> list;

		ds = this.getDataset(index);
		if ((ds != null) && (this.getLength() > 1)) {
			this.setIndex(index);

			// delete jobs from JobManager
			list = JobDesc.getStatusList(this.getFile(), ds.getDsId(), null);
			if (list != null) {
				for (String jobId : list.keySet()) {
					JobDesc.remove(jobId);
				}
			}

			this.paramDataStruct.reduce(this.getAllFilesDirectory(), index);

			// correct index
			if (this.getIndex() >= this.getLength()) {
				this.setIndex(this.getLength() - 1);
			}

			this.fireDatasetRemoved(this, index, index);
		} else {
			this.warning("Unable to reduce!");
		}
	}

	@Override
	public void info(String identifier, String message, StackTraceElement[] stack) {
		if (this.printEvent) {
			System.out.println("INFO: " + message);
		}
		this.getEventLog().info(identifier, message, stack);
	}

	@Override
	public void warningStructure(String identifier, String message, StackTraceElement[] stack) {
		if (this.printEvent) {
			System.out.println("WARNING: " + message);
		}
		this.getEventLog().warning(identifier, message, stack);
	}

	@Override
	public void errorStructure(String identifier, String message, StackTraceElement[] stack) {
		if (this.printEvent) {
			System.out.println("ERROR: " + message);
		}
		this.getEventLog().error(identifier, message, stack);
	}

	@Override
	public void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamContainer")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("ReportParameter")) {
				this.getReportParameter().xml2obj(xmlReader);
			} else if (element.equals("ParamDataStruct")) {
				this.paramDataStruct.xml2obj(xmlReader);
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	public void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("Project");
		xmlWriter.writeAttribute("object", this.getClassName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// ReportParameter
		this.getReportParameter().obj2xml(xmlWriter);

		// ParamDataStruct
		this.paramDataStruct.obj2xml(xmlWriter);

		xmlWriter.writeEndElement();
	}

	/**
	 * Create doc-files content-overview
	 * 
	 * @param directory Report-directory
	 * @throws IOException
	 */
	public void createDocFilesContent(String directory) throws IOException {
		this.createDocFilesContent(new File(directory));
	}

	/**
	 * Create doc-files content-overview
	 * 
	 * @param directory Report-directory
	 * @throws IOException
	 */
	public void createDocFilesContent(File directory) throws IOException {
		Parameter[] params = null;
		String fileName;

		directory.mkdirs();
		HtmlWriter content = new HtmlWriter(directory);

		// header
		content.add("<html>\n");
		content.add("<head>");
		content.add("<title>" + "Doc file content" + "</title>");
		content.add("<link rel=\"stylesheet\" type=\"text/css\" " + "href=\".\\LibContents_ACCM.css\">");
		content.add("</head>\n");
		content.add("<body>\n");

		// html header line
		content.add("<div id=\"Header\">");
		content.add("<table><tr>");
		content.add("<td align=\"left\">MagOpt " + Project.getVersion() + "</td>");
		content.add("<td align=\"right\">Doc file content</td>");
		content.add("</tr></table>");
		content.add("</div>\n");

		content.add("<div id=\"Documentation\">\n");
		content.add("<h1>" + this.getDescription() + " - " + this.getName() + "</h1>");

		// Projectinfo
		content.add("<table class=\"ProjectInfo\"><tr>");
		content.add("<td class=\"ProjectInfo\">Index #" + this.getIndex() + "</td>");
		content.add("<td style=\"border:none;\">" + this.getDSComment() + "</td>");
		content.add("<meta http-equiv=\"content-type\" content=\"text/html; charset=ISO-8859-1\">");
		content.add("</tr></table>\n");

		// Content table
		content.add("<table>\n\n");
		content.add("<tr>");
		content.add("<td class=\"Master\">Parameter Name</td>");
		content.add("<td class=\"Master\">Name</td>");
		content.add("<td class=\"Master\">File</td>");
		content.add("<td class=\"Master\">Description</td>");
		content.add("<td class=\"Master\">Checksum</td>");
		content.add("</tr>\n");

		params = this.getParameters(Parameter.class);
		for (Parameter param : params) {
			DocFile[] files = param.getDocFiles();
			if (files != null) {
				for (DocFile file : files) {
					fileName = (new File(file.getFile())).getName();
					content.add("<tr>");
					content.add("<td class=\"Slave\">" + file.getParamName() + "</td>");
					content.add("<td class=\"Slave\">" + file.getName() + "</td>");
					content.add("<td class=\"Slave\"><a href=\".\\" + fileName + "\">" + fileName + "</a></td>");
					content.add("<td class=\"Slave\">" + file.getDescription() + "</td>");
					content.add("<td class=\"Slave\" style=\"text-align:right;\">" + file.getChecksum() + "</td>");
					content.add("</tr>\n");
				}
			}
		}

		content.add("\n\n</table>\n");
		content.add("</div>\n");

		// html footer line
		content.add("<div id=\"Footer\">");
		content.add("<table><tr>");
		content.add(
				"<td align=\"left\">&copy; 2008 Austrian Center of Mechatronics GmbH &diams; Altenberger Str. 69 &diams; A-4040 Linz</td>");
		content.add("</tr></table>");
		content.add("</div>\n");

		// html end
		content.add("</body>\n");
		content.add("</html>\n");

		// write content to File
		content.write();
	}

	/**
	 * Create Tex-report of project
	 * 
	 * @param struct
	 * @throws IOException
	 */
	public void createTexReport(ReportStructure struct) throws IOException {
		File texDir;

		texDir = new File(this.getFilesDirectory(), "tex");
		if (FileUtils.deleteContentFiles(texDir)) {
			this.generateDocumentationFiles();
			this.reportTex(new File(texDir, "report.tex"), struct);
		} else {
			this.error("Unable to delete directory");
		}
	}

	/**
	 * Report current-dataset
	 * 
	 * @param directory Report-directory
	 * @param struct    Report-structure
	 * @throws IOException
	 */
	public void reportTex(String directory, ReportStructure struct) throws IOException {
		this.reportTex(new File(directory), struct);
	}

	/**
	 * Report current-dataset
	 * 
	 * @param file   Report-file
	 * @param struct Report-structure
	 * @throws IOException
	 */
	public void reportTex(File file, ReportStructure struct) throws IOException {
		String home;
		DocTexWriter content, localContent, partContent;

		if (struct == null) {
			struct = new ReportStructure();
		}
		file.getParentFile().mkdirs();
		content = new DocTexWriter(file);

		// header
		content.add("\\documentclass[11pt,ngerman]{article}\n\n");

		home = Project.getInstallPath().replace(File.separatorChar, '/');
		content.add("\\newcommand{\\MoHomePath}{\"" + home + "\"}\n");
		content.add("\\usepackage{\\MoHomePath/" + this.getReportParameter().getStylePath() + "}\n\n");

		// add style-properties
		ReportProperty[] list = this.getReportParameter().getProperties();
		for (ReportProperty e : list) {
			content.add("\\renewcommand{\\");
			content.addString(e.getKey());
			content.add("}{");
			content.addString(e.getValue());
			content.add("}\n");
		}
		content.add("\n");

		content.add("\\begin{document}\n");
		content.add("\\selectlanguage{" + this.getReportParameter().getLanguage().getTex() + "}\n");
		content.add("\\maketitle\n");
		content.add("\\newpage\n\n");

		// Summary
		if (struct.getSummary()) {
			partContent = new DocTexWriter(new File(file.getParentFile(), "Summary.tex"));

			partContent.add("\\begin{landscape}\n");
			partContent.add("\\part*{Summary}\n");
			partContent.add("\\ifdefined\\usehtml\\else\n");
			partContent.add("\t\\addcontentsline{toc}{part}{Summary}\n");
			partContent.add("\\fi\n\n");

			partContent.add("\\section*{Index \\#" + (this.getIndex() + 1) + " - ");
			partContent.addString(this.getDSComment());
			partContent.add("}\n");
			partContent.add("\\ifdefined\\usehtml\\else\n");
			partContent.add("\t\\addcontentsline{toc}{section}{Index \\#" + this.getIndex() + " - ");
			partContent.addString(this.getDSComment());
			partContent.add("}\n");
			partContent.add("\\fi\n\n");

			// Content table
			partContent.add("\\ifdefined\\usehtml\n");
			partContent
					.add("\t\\begin{tabular}{|m{6cm} m{6cm} >{\\raggedleft}m{3cm} >{\\centering}m{3cm} m{1.5cm}|}\n");
			partContent.add("\\else\n");
			partContent
					.add("\t\\begin{longtable}{|m{6cm} m{6cm} >{\\raggedleft}m{3cm} >{\\centering}m{3cm} m{1.5cm}|}\n");
			partContent.add("\\fi\n\n");
			partContent.add("\\hline ");
			this.reportTable(partContent, 0);
			partContent.add("\\\\\n");
			partContent.add("\\hline\n\n");
			partContent.add("\\ifdefined\\usehtml");
			partContent.add("\t\\end{tabular}");
			partContent.add("\\else");
			partContent.add("\t\\end{longtable}");
			partContent.add("\\fi\n");
			partContent.add("\\end{landscape}\n\n");

			partContent.write();
			content.add("\\input{Summary}\n\n");
		}

		// Details description
		if (struct.getDetail()) {
			localContent = this.reportTexDetails(file, home, struct);
			if (localContent.length() > 0) {
				partContent = new DocTexWriter(new File(file.getParentFile(), "Detail.tex"));

				partContent.add("\\part*{Detail description}\n");
				partContent.add("\\ifdefined\\usehtml\\else\n");
				partContent.add("\t\\addcontentsline{toc}{part}{Detail description}\n");
				partContent.add("\\fi\n\n");
				partContent.addWriter(localContent);
				partContent.add("\n");

				partContent.write();
				content.add("\\input{Detail}\n\n");
			}
		}

		// function doku
		if (struct.getFunction()) {
			localContent = this.reportTexFunctions(file, home, struct);
			if (localContent.length() > 0) {
				partContent = new DocTexWriter(new File(file.getParentFile(), "Functions.tex"));

				partContent.add("\\newpage{}\n");
				partContent.add("\\part*{Function description}\n");
				partContent.add("\\ifdefined\\usehtml\\else\n");
				partContent.add("\t\\addcontentsline{toc}{part}{Function description}\n");
				partContent.add("\\fi\n\n");
				partContent.addWriter(localContent);
				partContent.add("\n\\graphicspath{{.}}\n");
				partContent.add("\\renewcommand{\\InputPath}{.} \n\n");

				partContent.write();
				content.add("\\input{Functions}\n\n");
			}
		}

		// package doku
		if (struct.getPackage()) {
			localContent = this.reportTexPackages(file, home, struct);
			if (localContent.length() > 0) {
				partContent = new DocTexWriter(new File(file.getParentFile(), "Packages.tex"));

				partContent.add("\\newpage{}\n");
				partContent.add("\\part*{Package description}\n");
				partContent.add("\\ifdefined\\usehtml\\else\n");
				partContent.add("\t\\addcontentsline{toc}{part}{Package description}\n");
				partContent.add("\\fi\n\n");
				partContent.addWriter(localContent);
				partContent.add("\n\\graphicspath{{.}}\n");
				partContent.add("\\renewcommand{\\InputPath}{.} \n\n");

				partContent.write();
				content.add("\\input{Packages}\n\n");
			}
		}

		// data-sheet doku
		if (struct.getDatasheet()) {
			localContent = this.reportTexDatasheets(file, home, struct);
			if (localContent.length() > 0) {
				partContent = new DocTexWriter(new File(file.getParentFile(), "Datasheets.tex"));

				partContent.add("\\newpage{}\n");
				partContent.add("\\part*{Datasheet}\n");
				partContent.add("\\ifdefined\\usehtml\\else\n");
				partContent.add("\t\\addcontentsline{toc}{part}{Data-sheet}\n");
				partContent.add("\\fi\n\n");
				partContent.addWriter(localContent);
				partContent.add("\n\\graphicspath{{.}}\n");
				partContent.add("\\renewcommand{\\InputPath}{.} \n\n");

				partContent.write();
				content.add("\\input{Datasheets}\n\n");
			}
		}

		content.add("\\ifdefined\\usehtml\\else\n");
		content.add("\t\\newpage{}\n");
		content.add("\t\\part*{Appendix}\n");

		content.add("\t\\addcontentsline{toc}{part}{Appendix}\n\n");

		// list of figures
		content.add("\t\\addcontentsline{toc}{section}{List of figures}\n");
		content.add("\t\\listoffigures{}\n");

		// index list
		content.add("\t\\addcontentsline{toc}{section}{Index}\n");
		content.add("\t\\printindex{}\n");
		content.add("\\fi\n\n");

		content.add("\\end{document}\n");

		// write content to File and compile
		content.write();
		if (content.compile(true) != 0) {
			this.error("Unable to compile tex-file");
		}
	}

	/**
	 * Report detail-description of all parameters
	 * 
	 * @param file   Name of detail-tex-file
	 * @param home   Path to magopt
	 * @param struct Report-structure
	 * @return Tex-Content with detail-informations
	 * @throws IOException
	 */
	protected DocTexWriter reportTexDetails(File file, String home, ReportStructure struct) throws IOException {
		DocTexWriter content, localContent;
		Parameter[] params;
		String name;

		content = new DocTexWriter(file);
		params = this.getParameters(Parameter.class);
		for (Parameter param : params) {
			name = "D_" + param.getIdentifier();
			name = name.replace(' ', '_');
			name = name.replace('/', '_');
			name = name.replace('\\', '_');
			localContent = new DocTexWriter(new File(content.getFile().getParentFile(), name + ".tex"));
			param.reportDetail(localContent, struct);

			if (localContent.length() > 0) {
				// localContent.write();
				// content.add("\\input{" + name + "}\n\n");
				content.addWriter(localContent);
			}
		}

		return content;
	}

	/**
	 * Report description of used functions
	 * 
	 * @param file   Name of functions-tex-file
	 * @param home   Path to magopt
	 * @param struct Report-structure
	 * @return Tex-Content with function-informations
	 */
	protected DocTexWriter reportTexFunctions(File file, String home, ReportStructure struct) {
		String name, path;
		DocTexWriter content;
		Parameter[] params;
		ParamFunction func;
		File texFile;
		LinkedList<String> list;

		list = new LinkedList<String>();
		content = new DocTexWriter(file);
		params = this.getParameters(ParamFunction.class);
		for (Parameter param : params) {
			func = (ParamFunction) param;
			if (!list.contains(func.getClassName())) {
				list.add(func.getClassName());
				texFile = new File(func.getClassPath());
				if (func instanceof ParamJavaFunction) {
					texFile = texFile.getParentFile();
				}
				name = func.getClassName().replace(".", File.separator);
				texFile = FileUtils.getFile(texFile.getPath(), "tex", name + "_Doc", new File(name).getName() + ".tex");
				if (!texFile.exists()) {
					texFile = new File(texFile.getParentFile(), new File(name).getName() + ".en.tex");
					if (!texFile.exists()) {
						texFile = new File(texFile.getParentFile(), new File(name).getName() + ".ger.tex");
					}
				}

				if (texFile.exists()) {
					content.add("\n\\hypertarget{");
					content.addName("F_" + func.getClassName());
					content.add("}{}");

					path = texFile.getParent().replace(File.separatorChar, '/');
					if ((path + "/").startsWith(home + "/")) {
						path = "\\MoHomePath" + path.substring(home.length());
					}
					content.add("\\graphicspath{{" + path + "/}}\n");
					content.add("\\renewcommand{\\InputPath}{\"" + path + "/\"}\n");
					content.add("\\input{" + path + "/" + texFile.getName() + "}\n");

					if (struct.getPackage()) {
						content.add("\\vspace{2ex}\\hyperlink{");
						content.addName("P_" + func.getLibraryPackage());
						content.add("}{Package: ");
						content.addString(func.getLibraryPackage().replace("\\", "/"));
						content.add("}\n");
					}
					content.add("\n");
				}
			}
		}

		return content;
	}

	/**
	 * Report description of used packages
	 * 
	 * @param file   Name of package-tex-file
	 * @param home   Path to magopt
	 * @param struct Report-structure
	 * @return Tex-Content with package-informations
	 */
	protected DocTexWriter reportTexPackages(File file, String home, ReportStructure struct) {
		String name, path;
		DocTexWriter content;
		Parameter[] params;
		ParamFunction func;
		File texFile;
		LinkedList<String> list;

		list = new LinkedList<String>();
		content = new DocTexWriter(file);
		params = this.getParameters(ParamFunction.class);
		for (Parameter param : params) {
			func = (ParamFunction) param;
			if (!list.contains(func.getLibraryPackage())) {
				list.add(func.getLibraryPackage());
				name = new File(func.getLibraryPackage()).getName();
				texFile = FileUtils.getFile(func.getLocation(), func.getLibraryPackage(), "Doc", name + ".en.tex");
				if (!texFile.exists()) {
					texFile = new File(texFile.getParentFile(), name + ".ger.tex");
					if (!texFile.exists()) {
						texFile = new File(texFile.getParentFile(), name + ".tex");
					}
				}
				if (texFile.exists()) {
					content.add("\\hypertarget{");
					content.addName("P_" + func.getLibraryPackage());
					content.add("}{}");

					path = texFile.getParent().replace(File.separatorChar, '/');
					if ((path + "/").startsWith(home + "/")) {
						path = "\\MoHomePath" + path.substring(home.length());
					}
					content.add("\\graphicspath{{" + path + "/}}\n");
					content.add("\\renewcommand{\\InputPath}{\"" + path + "/\"}\n");
					content.add("\\input{" + path + "/" + texFile.getName() + "}\n");
				}
			}
		}

		return content;
	}

	/**
	 * Report description of used packages
	 * 
	 * @param file   Name of package-tex-file
	 * @param home   Path to magopt
	 * @param struct Report-structure
	 * @return Tex-Content with package-informations
	 */
	protected DocTexWriter reportTexDatasheets(File file, String home, ReportStructure struct) {
		DocTexWriter content;
		Parameter[] params;
		DocFile ds;

		content = new DocTexWriter(file);
		params = this.getParameters(Parameter.class);
		for (Parameter param : params) {
			ds = param.getDocFile("Datasheet");
			if (ds != null) {
				content.add("\\hypertarget{");
				content.addName("DS_" + param.getIdentifier());
				content.add("}{}");
				content.addTex(ds);
			}

			// path = texFile.getParent().replace(File.separatorChar, '/');
			// if ((path + "/").startsWith(home + "/")) {
			// path = "\\MoHomePath" + path.substring(home.length());
			// }
			// content.add("\\graphicspath{{" + path + "/}}\n");
			// content.add("\\renewcommand{\\InputPath}{\"" + path
			// + "/\"}\n");
			// content.add("\\input{" + path + "/" + texFile.getName()
			// + "}\n");
		}

		return content;
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tDataset-Comment & ");
		content.addString(this.getDSComment());
		content.add("\\\\\n");

		content.add("\tIndex & ");
		content.add(this.getIndex() + "\\\\\n");

		content.add("\tLength & ");
		content.add(this.getLength() + "\\\\\n");

		content.add("\tFile & \\mopath{");
		content.addString(
				("." + File.separator + (new File(this.getFile())).getName()).replace(File.separatorChar, '/'));
		content.add("}\\\\\n");

	}

	/**
	 * Generate documentationsfiles of current index
	 */
	public void generateDocumentationFiles() {
		Parameter[] params;
		File docDir;

		docDir = new File(this.getFilesDirectory(), "docFiles");

		if (!docDir.exists()) {
			docDir.mkdirs();
		}

		// docFiles
		params = this.getParameters(Parameter.class);
		for (Parameter p : params) {
			this.generateDocumentationFiles(p, docDir);
		}

		// component docFiles
		params = this.getParameters(ParamContainer.class);
		for (int k = params.length - 1; k >= 0; k--) {
			if (((ParamContainer) params[k]).getComponent() != null) {
				System.out.println("Generate component " + params[k].getName());
				((ParamContainer) params[k]).getComponent().generateDocFiles(docDir);
			}
		}

		try {
			this.createDocFilesContent(docDir);
		} catch (IOException e) {
		}
		System.out.println("Generation finished");
	}

	/**
	 * Generate documentation-files of current index
	 * 
	 * @param param  Parameter to generate doc-files
	 * @param docDir Directory for export-files
	 */
	protected void generateDocumentationFiles(Parameter param, File docDir) {

		try {
			if (!param.checkDocFiles()) {
				System.out.println("Generate " + param.getName());
				// param.deleteDocFiles();
				param.generateDocFiles(docDir);
			}
		} catch (Exception e) {
			this.error(param.getIdentifier(), "Error by creating documentation-files \n" + e.getMessage(),
					e.getStackTrace());
		}

	}

	/**
	 * Generate documentation-files of current index
	 *
	 * @param param Parameter to generate doc-files
	 */
	public void generateDocumentationFiles(Parameter param) {
		File docDir;

		docDir = new File(this.getFilesDirectory(), "docFiles");

		if (!docDir.exists()) {
			docDir.mkdirs();
		}

		this.generateDocumentationFiles(param, docDir);
		if ((param instanceof ParamContainer) && (((ParamContainer) param).getComponent() != null)) {
			System.out.println("Generate component " + param.getName());
			((ParamContainer) param).getComponent().generateDocFiles(docDir);
		}
	}

	/**
	 * Reset documentation files of all parameters
	 */
	public void resetDocumentationFiles() {
		this.execStructure(new ExecStructure() {

			@Override
			public void execute(Parameter param) {
				param.resetDocFiles();
			}
		});
	}

	/**
	 * Reset documentation files of parameter
	 * 
	 * @param param Parameter to reset
	 */
	public void resetDocumentationFiles(Parameter param) {
		param.resetDocFiles();
	}

	/**
	 * Delete documentation files of all parameters
	 */
	public void deleteDocumentationFiles() {
		this.execStructure(new ExecStructure() {

			@Override
			public void execute(Parameter param) {
				if (param instanceof Project) {
					((Project) param).deleteDocFilesOpt(param.getIndex());
				} else {
					param.deleteDocFiles();
				}
			}
		});
	}

	/**
	 * Delete documentation files of parameter
	 * 
	 * @param param Parameter
	 */
	public void deleteDocumentationFiles(Parameter param) {
		if (param instanceof Project) {
			((Project) param).deleteDocFilesOpt(this.getIndex());
		} else {
			param.deleteDocFiles();
		}
	}

	@Override
	protected void generateDocFiles(File directory) throws Exception {
		DocTexWriter content;
		File file;
		DocFile[] docs;
		DocFile doc;

		// recalculate checksum
		docs = this.getDocFiles();
		for (DocFile d : docs) {
			file = new File(FileUtils.trimPath(((Project) this.getRoot()).getFilesDirectory()), d.getFile());
			d.setChecksum(FileUtils.checkSum(file));
		}

		// generate template
		file = this.getUniqueDocFileName(new File(directory, "Template.tex"));
		content = new DocTexWriter(file);

		if (this.checkDocFile("FigOptRed_withSelection")) {
			doc = this.getDocFile("FigOptRed_withSelection");
			content.add("\\begin{figure}[H]");
			content.add("\\centering");
			content.addGraphic(doc, "angle=90,width=15cm");
			content.add("\\caption{");
			content.addString(doc.getDescription());
			content.add(".}");
			content.add("\\end{figure} ");
		}

		if (this.checkDocFile("TableVariedParameters_withSelection")) {
			content.addTable(this.getDocFile("TableVariedParameters_withSelection"), "Varied parameters",
					"TAB_Variedparameters");
		}

		if (this.checkDocFile("TableObjectives_withSelection")) {
			content.addTable(this.getDocFile("TableObjectives_withSelection"), "Objectives", "TAB_Objective");
		}

		if (this.checkDocFile("TableConstraints_withSelection")) {
			content.addTable(this.getDocFile("TableConstraints_withSelection"), "Contstraints", "TAB_Constraints");
		}

		if (content.length() > 0) {
			content.write();
			this.registerDocFile("Template", file.getPath(), "Template");
		}

		// generate template to fill
		for (int k = 0; k < this.getNumTemplates(); k++) {
			File template = this.getTemplate(k);
			String name = template.getName().substring(0, template.getName().length() - 4);

			FillTexReport fill = new FillTexReport(template, this);
			File tex = fill.getFile();
			if (tex.exists()) {
				int ret = -1;
				try {
					ret = fill.compile(false);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				if (ret == 0) {
					File report = new File(tex.getPath().substring(0, tex.getPath().length() - 4) + ".pdf");
					File docfile = this.getUniqueDocFileName(new File(directory, report.getName()));
					try {
						FileUtils.move(report, docfile);
						this.registerDocFile(name, docfile.getPath(), "Docu template " + name);
					} catch (Exception e) {
						this.warning(e.getMessage(), e.getStackTrace());
					}
				} else {
					this.warning("Unable to compile template DocFile'" + name + "'");
				}
				tex.delete();
			} else {
				this.warning("Unable to create template DocFile'" + template.getName() + "'");
			}

		}
	}

	@Override
	public void deleteDocFiles(int index) {
		DocFile[] docs;
		File file;
		File[] fileList;
		String[] fileParts, parts;

		docs = this.getDocFiles();
		for (DocFile d : docs) {
			if (d.getName().compareTo("Template") == 0) {
				file = new File(FileUtils.trimPath(((Project) this.getRoot()).getFilesDirectory()), d.getFile());
				fileParts = FileUtils.splitName(file.getName());
				fileList = file.getParentFile().listFiles();
				if (fileList != null) {
					for (File f : fileList) {
						parts = FileUtils.splitName(f.getName());
						if (fileParts[0].compareToIgnoreCase(parts[0]) == 0) {
							f.delete();
						}
					}
				}
				this.unregisterDocFile(d.getName());
			}
		}

	}

	/**
	 * Delete doc-files included the optimizer doc-files
	 * 
	 * @param index Index of dataset
	 */
	public void deleteDocFilesOpt(int index) {
		super.deleteDocFiles(index);
	}

	@Override
	public ParameterPath getParamByUniqueIdentifier(String uIdName, boolean allAccess) {
		String name = this.getName();

		if ((uIdName != null) && uIdName.startsWith(".")) {
			if (uIdName.length() == 1) {
				uIdName = name;
			} else {
				uIdName = name + uIdName;
			}
		}

		return super.getParamByUniqueIdentifier(uIdName, allAccess);
	}

	@Override
	protected AbstractList<ParameterPath> getParamByIdentifier(String idName, boolean thisPublic) {
		String name = this.getName();

		if ((idName != null) && idName.startsWith(".")) {
			if (idName.length() == 1) {
				idName = name;
			} else {
				idName = name + idName;
			}
		}

		return super.getParamByIdentifier(idName, thisPublic);
	}

	@Override
	public ReportParameter getReportParameter() {
		return this.reportParam;
	}

	@Override
	protected ParamDataStruct getParameterStruct() {
		return this.paramDataStruct;
	}

	// /**
	// * Set parameter-data-structure
	// *
	// * @param struct
	// * Parameter-data-structure
	// */
	// protected void setParameterStruct(ParamDataStruct struct) {
	// this.paramDataStruct = struct;
	// }

	/**
	 * Move dataset in project
	 * 
	 * @param from Move from index
	 * @param to   Move to index
	 */
	public void moveDataset(int from, int to) {
		this.getParameterStruct().move(from, to);
	}

	@Override
	public void setIndex(int index) {
		ParamDataStruct pds;
		int from;

		if ((index >= 0) && (index < this.getLength())) {
			pds = this.getParameterStruct();
			if (pds != null) {
				if (pds.getIndex() != index) {
					from = pds.getIndex();
					pds.setIndex(index);
					this.fireDatasetIndexChanged(this, from, index);
				}
			}
		} else {
			this.warning("Index " + index + " out of Range!");
		}
	}

	@Override
	public File getUniqueDocFileName(File file) {
		File path;
		String name;
		String[] fileparts, parts;
		int i = 0;
		boolean exists = true;
		LinkedList<String> files;

		// list all docfiles
		files = new LinkedList<String>();
		for (Parameter param : this.getParameters(Parameter.class)) {
			for (DocFile doc : param.getDocFiles()) {
				files.add(FileUtils.splitName(new File(doc.getFile()).getName())[0].toLowerCase());
			}
		}

		// check existing file
		path = file.getParentFile();
		fileparts = FileUtils.splitName(file.getName());
		name = fileparts[0];
		if (path.exists()) {
			while (exists) {
				exists = files.contains(name.toLowerCase());
				if (!exists) {
					for (File f : path.listFiles()) {
						parts = FileUtils.splitName(f.getName());
						exists |= (name.compareToIgnoreCase(parts[0]) == 0);
					}
				}
				if (exists) {
					i++;
					name = fileparts[0] + "(" + i + ")";
				}
			}
		}

		return new File(path, name + fileparts[1]);
	}

	/**
	 * Get help-file
	 * 
	 * @return Help
	 */
	public static File getHelpFile() {
		return FileUtils.getFile(Project.getInstallPath(), "Help", "_build", "html", "index.html");
	}

	/**
	 * Update help
	 */
	public static void updateHelp() {
		Execute exec;
		File help;
		LinkedList<String> cmd;

		cmd = new LinkedList<String>();
		cmd.add("python.exe");
		cmd.add("createHelp.py");

		help = FileUtils.getFile(Project.getInstallPath(), "Help");
		exec = new Execute(help, cmd.toArray(new String[0]));
		exec.setRedirectStreams(true);
		exec.start();

		if (exec.exitValue() != 0) {
			System.err.println("Update help doesn't finished!");
		}
	}

	/**
	 * Open help file
	 */
	public static void openHelp() {
		File help;

		help = Project.getHelpFile();
		if (help.exists()) {
			if (Desktop.isDesktopSupported()) {
				try {
					Desktop.getDesktop().open(help);
				} catch (IOException e) {
					// FIXME winopen
					System.err.println("Unable to open help (Exception while open)");
				}
			} else {
				System.err.println("Unable to open help (Unsupported)");
			}
		} else {
			System.err.println("Unable to open help (file doesn't exist)");
		}
	}

	/**
	 * Get documentation file
	 * 
	 * @return Documentation
	 */
	public static File getDocuFile() {
		return FileUtils.getFile(Project.getInstallPath(), "Documentation", "MagOptDoku.pdf");
	}

	/**
	 * Update documentation
	 */
	public static void updateDocu() {
		Execute exec;
		TexWriter tex;
		int ret;

//		exec = new Execute(Project.getDocuFile().getParentFile(), "texify", "-s", "-p", "-b", "-c", "MagOptDoku.tex");
//		exec.setRedirectStreams(true);
//		exec.start();
//		ret = exec.exitValue();

		tex = new TexWriter(new File(Project.getDocuFile().getParentFile(), "MagOptDoku.tex"));
		try {
			ret = tex.compile(false);
		} catch (IOException e) {
			ret = -1;
		}

		if (ret != 0) {
			System.err.println("Update documentation doesn't finished successfully");
			System.err.println(
					"Make sure MiKTeX is installed and <MiKTeX-Path>\\miktex\\bin\\ is in the path environment variable path");
		}
	}

	/**
	 * Open documentation file
	 */
	public static void openDocu() {
		File docu;

		docu = Project.getDocuFile();
		if (docu.exists()) {
			if (Desktop.isDesktopSupported()) {
				try {
					Desktop.getDesktop().open(docu);
				} catch (IOException e) {
					// FIXME winopen
					System.err.println("Unable to open documentation (Exception while open)");
				}
			} else {
				System.err.println("Unable to open documentation (Unsupported)");
			}
		} else {
			System.err.println("Unable to open documtation (file doesn't exist)");
		}
	}

	@Override
	public void resetAllFunctions() {
		this.execStructure(new ExecStructure() {

			@Override
			public void execute(Parameter param) {
				if (param instanceof IFunction) {
					try {
						((IFunction) param).reset();
					} catch (RemoteException e) {
						// TODO RMI
						e.printStackTrace();
					}
				}
			}
		});
	}

	@Override
	public IProject getRMIImpl() throws RemoteException {
		return new RProject(this);
	}

	@Override
	public IProject getClientImpl() {
		return new CProject(this);
	}

	/**
	 * Get list of listeners
	 * 
	 * @return <code>DatasetListener</code> list
	 */
	private LinkedList<DatasetListener> getDatasetListenersList() {
		if (this.datasetListeners == null) {
			this.datasetListeners = new LinkedList<DatasetListener>();
		}
		return this.datasetListeners;
	}

	/**
	 * Add dataset listener
	 * 
	 * @param listener Listener to add
	 */
	public void addDatasetListener(DatasetListener listener) {
		LinkedList<DatasetListener> list;

		list = this.getDatasetListenersList();
		synchronized (list) {
			if (!list.contains(listener)) {
				list.add(listener);
			}
		}
	}

	/**
	 * Remove dataset listener
	 * 
	 * @param listener Listener to remove
	 */
	public void removeDatasetListener(DatasetListener listener) {
		LinkedList<DatasetListener> list;

		list = this.getDatasetListenersList();
		synchronized (list) {
			list.remove(listener);
		}
	}

	/**
	 * Get list of dataset listener
	 * 
	 * @return List of listener
	 */
	public DatasetListener[] getDatasetListeners() {
		LinkedList<DatasetListener> list;

		list = this.getDatasetListenersList();
		synchronized (list) {
			return list.toArray(new DatasetListener[0]);
		}
	}

	/**
	 * Call to fire the event
	 * 
	 * @param event Event that occurs
	 */
	protected void fireDatasetChanged(DatasetEvent event) {
		LinkedList<DatasetListener> list;

		list = this.getDatasetListenersList();
		synchronized (list) {
			for (DatasetListener listener : list) {
				listener.datasetChanged(event);
			}
		}
	}

	/**
	 * Call after dataset-index has been changed
	 * 
	 * @param source The <code>Project</code> that changed
	 * @param from   Old dataset-index
	 * @param to     New dataset-index
	 */
	protected void fireDatasetIndexChanged(Object source, int from, int to) {
		this.fireDatasetChanged(new DatasetEvent(source, DatasetType.DATASET_INDEX_CHANGED, from, to));
	}

	/**
	 * Call after one or more dataset has been added
	 * 
	 * @param source The <code>Project</code> that changed
	 * @param from   Lower index of range
	 * @param to     Upper index of range
	 */
	protected void fireDatasetAdded(Object source, int from, int to) {
		this.fireDatasetChanged(new DatasetEvent(source, DatasetType.DATASET_ADDED, from, to));
	}

	/**
	 * Call after one or more dataset has been removed
	 * 
	 * @param source The <code>Project</code> that changed
	 * @param from   Lower index of range
	 * @param to     Upper index of range
	 */
	protected void fireDatasetRemoved(Object source, int from, int to) {
		this.fireDatasetChanged(new DatasetEvent(source, DatasetType.DATASET_REMOVED, from, to));
	}

	@Override
	void setIncomplete(int index, boolean bool) {
		if (this.getIncomplete(index) != bool) {
			super.setIncomplete(index, bool);
			this.fireDatasetIndexChanged(this, index, index);
		}
	}

	@Override
	protected void setWarningStructure(int index, boolean error) {
		if (this.getWarning(index) != error) {
			super.setWarningStructure(index, error);
			this.fireDatasetIndexChanged(this, index, index);
		}
	}

	@Override
	protected void setErrorStructure(int index, boolean error) {
		if (this.getError(index) != error) {
			super.setErrorStructure(index, error);
			this.fireDatasetIndexChanged(this, index, index);
		}
	}

	@Override
	public Project extractIndex(int index, File file) throws Exception {
		Project proj;

		proj = new Project("Project", "", file);
		this.cloneParam(proj, index);
		proj.refresh(true);

		return proj;
	}

	/**
	 * Get python connection
	 * 
	 * @return Jep-proxy
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	public JepProxy getJep() throws InterruptedException, ExecutionException {
		if (this.jep == null) {
			this.restartJep();
		}

		return this.jep;
	}

	@Override
	public void restartJep() throws InterruptedException, ExecutionException {
		this.jep = new JepProxy();
	}
}
