/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_6.
 * 
 * - find "TStruct"-sections containing periodic data and replace with
 * "TPeriodic"-Type
 * 
 * @author Werner Koppelsttter, Gnther Weidenholzer
 */
class Converter_6 {

	// /** Number of data-sets */
	// private static int len = 1;

	/**
	 * Convert the given node
	 * 
	 * @param node
	 *            Data to convert
	 * @return Changed data node
	 * @throws Exception
	 */
	static XMLNode convert(XMLNode node) throws Exception {

		// find "TStruct"-sections containing periodic data and replace with
		// "TPeriodic"-Type
		Converter_6.convertStructToTPeriodic(node);

		return node;
	}

	/**
	 * find "TStruct"-sections containing periodic data and replace with
	 * "TPeriodic"-Type
	 * 
	 * @param node
	 *            Data to convert
	 * @throws Exception
	 */
	private static void convertStructToTPeriodic(XMLNode node) throws Exception {
		XMLNode item, newnode;

		// scan subnodes
		for (int i = 0; i < node.getNodeSize(); i++) {
			item = node.getNode(i);
			if (item.getName().compareTo("Struct") == 0) {
				newnode = Converter_6.Struct2Periodic(item);
				if (newnode != null) {
					node.setNode(i, newnode);
				}
			}
		}

		// scan next
		for (XMLNode n : node.getNodes()) {
			Converter_6.convertStructToTPeriodic(n);
		}

	}

	/**
	 * 
	 * @param node
	 *            Data to convert
	 * @return Periodic representation of node data
	 * @throws Exception
	 */
	private static XMLNode Struct2Periodic(XMLNode node) throws Exception {

		XMLNode periodic = null;

		boolean pp_ok = false;
		boolean rms_ok = false;
		boolean thd_ok = false;

		double mean = Double.NaN;
		String fourier_a = null;
		String fourier_b = null;
		XMLNode item;

		for (XMLNode n : node.getNodes()) {
			if (n.getAttribute("name").compareTo("pp") == 0) {
				pp_ok = true;
			} else if (n.getAttribute("name").compareTo("mean") == 0) {
				mean = Double.parseDouble(n.getNode(0).getText());
			} else if (n.getAttribute("name").compareTo("fourier_a") == 0) {
				fourier_a = n.getNode(0).getText();
			} else if (n.getAttribute("name").compareTo("fourier_b") == 0) {
				fourier_b = n.getNode(0).getText();
			} else if (n.getAttribute("name").compareTo("rms") == 0) {
				rms_ok = true;
			} else if (n.getAttribute("name").compareTo("thd") == 0) {
				thd_ok = true;
			}
		}

		if ((node.getNodeSize() == 6) && (mean != Double.NaN)
				&& (fourier_a != null) && (fourier_b != null) && pp_ok
				&& thd_ok && rms_ok) {

			TNumeric numA = new TNumeric(fourier_a);
			TNumeric numB = new TNumeric(fourier_b);

			StringBuffer coeffStr = new StringBuffer();

			coeffStr.append("[");

			int m = numA.size()[0];
			for (int k = 0; k < m; k++) {

				// COS-coeff
				coeffStr.append(Integer.toString(k + 1));
				coeffStr.append(",0,");
				coeffStr.append(Double.toString(numA.getReal(k, 0)));
				coeffStr.append(";");

				// SIN-coeff
				coeffStr.append(Integer.toString(k + 1));
				coeffStr.append(",1,");
				coeffStr.append(Double.toString(numB.getReal(k, 0)));
				coeffStr.append(";");

			}

			coeffStr.setCharAt(coeffStr.length() - 1, ']');

			periodic = new XMLNode("Periodic");

			// add DCOffset
			item = new XMLNode("DCOffset");
			item.setText(Double.toString(mean));
			periodic.addNode(item);

			// add Coeffs
			item = new XMLNode("Coeffs");
			item.setText(coeffStr.toString());
			periodic.addNode(item);
		}

		return periodic;
	}

}
