/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.general.utils.XMLNode;
import magopt.utils.MopAccess;
import magopt.utils.MopFile;

/**
 * Conversion of xml-files to a desired version Check the version of file and
 * convert to new version
 * 
 * @author Werner Koppelstaetter
 */
public class XMLConverter implements MopAccess {

	/**
	 * Structure with XML-data
	 */
	XMLNode node;

	/**
	 * File for conversion
	 */
	private MopFile file;

	/**
	 * Version of the load file
	 */
	private Integer oldVer = 0;

	/**
	 * Version of the program
	 */
	private Integer newVer = 0;

	/**
	 * Constructor of the class
	 * 
	 * @param file
	 *            Name of file to convert
	 * @param newVersion
	 *            Desired Version of the file
	 * @throws Exception
	 */
	public XMLConverter(String file, String newVersion) throws Exception {
		this(new MopFile(file), newVersion);
	}

	/**
	 * Constructor of the class
	 * 
	 * @param file
	 *            File to convert
	 * @param newVersion
	 *            Desired Version of the file
	 * @throws Exception
	 */
	public XMLConverter(File file, String newVersion) throws Exception {
		this(new MopFile(file.getPath()), newVersion);
	}

	/**
	 * Constructor of the class
	 * 
	 * @param file
	 *            MopFile to convert
	 * @param newVersion
	 *            Desired Version of the file
	 * @throws Exception
	 */
	public XMLConverter(MopFile file, String newVersion) throws Exception {

		Integer value = XMLConverter.getVersionCode(newVersion);
		if (value != null) {
			this.newVer = value;
		}

		if (file.exists()) {
			this.file = file;
			this.loadVersion(this.file);
		} else {
			throw new FileNotFoundException("File does not exist!");
		}
	}

	/**
	 * Convert version-string to a version-number
	 * 
	 * @param strVersion
	 *            String representation of version
	 * @return Version-string
	 */
	public static Integer getVersionCode(String strVersion) {
		String[] tmp;
		Integer value = null;
		if (strVersion.matches("V\\.\\d+\\.\\d+")) {
			tmp = strVersion.split("\\.");
			value = (1000000 * Integer.valueOf(tmp[1]))
					+ Integer.valueOf(tmp[2]);
		}
		return value;
	}

	/**
	 * Convert version-number to a version-string
	 * 
	 * @param intVersion
	 *            Integer representation of version
	 * @return Version-number
	 */
	public static String getVersionString(Integer intVersion) {
		return new String(
				"V." + (intVersion / 1000000) + "." + (intVersion % 1000000));
	}

	/**
	 * Compare the desired-version with the file-version
	 * 
	 * @return the value 0 if the desired-version is equal file-version; a value
	 *         less than 0 if the desired-version is less than file-version; and
	 *         a value greater than 0 if the desired-version is greater
	 *         file-version;
	 */
	public int compareVersion() {
		return this.newVer - this.oldVer;
	}

	/**
	 * Get the version of the file
	 * 
	 * @return Version-string of the file
	 */
	public String getOldVersion() {
		return XMLConverter.getVersionString(this.oldVer);
	}

	/**
	 * Get the desired-version
	 * 
	 * @return Version-string of the desired-version
	 */
	public String getNewVersion() {
		return XMLConverter.getVersionString(this.newVer);
	}

	/**
	 * Convert xml-structure from file-version to desired-version
	 * 
	 * @throws Exception
	 */
	public void convert() throws Exception {
		File backupFile;
		String strFile;

		// convert file
		if (this.oldVer > this.newVer) {
			throw new IOException(
					"File version is greater than program version!");
		} else if (this.oldVer < this.newVer) {

			// load file
			this.loadXML(this.file);

			// create backup
			strFile = this.file.getName() + "_"
					+ this.getOldVersion().replace('.', '_');
			backupFile = new File(this.file.getParentFile(), strFile);
			if (backupFile.exists()) {
				int num = 0;
				do {
					backupFile = new File(this.file.getParentFile(),
							strFile + "(" + num++ + ")");
				} while (backupFile.exists());
			}
			FileUtils.copy(this.file, backupFile);

			// run conversions
			if ((this.oldVer < 1) && (1 <= this.newVer)) {
				this.node = Converter_1.convert(this.node);
			}
			if ((this.oldVer < 2) && (2 <= this.newVer)) {
				this.node = Converter_2.convert(this.node);
			}
			if ((this.oldVer < 3) && (3 <= this.newVer)) {
				this.node = Converter_3.convert(this.node);
			}
			if ((this.oldVer < 4) && (4 <= this.newVer)) {
				this.node = Converter_4.convert(this.node);
			}
			if ((this.oldVer < 5) && (5 <= this.newVer)) {
				this.node = Converter_5.convert(this.node);
			}
			if ((this.oldVer < 6) && (6 <= this.newVer)) {
				this.node = Converter_6.convert(this.node);
			}
			if ((this.oldVer < 7) && (7 <= this.newVer)) {
				this.node = Converter_7.convert(this.node);
			}
			if ((this.oldVer < 8) && (8 <= this.newVer)) {
				this.node = Converter_8.convert(this.node);
			}
			if ((this.oldVer < 9) && (9 <= this.newVer)) {
				this.node = Converter_9.convert(this.node,
						this.file.getParentFile());
			}
			if ((this.oldVer < 10) && (10 <= this.newVer)) {
				this.node = Converter_10.convert(this.node);
			}
			if ((this.oldVer < 11) && (11 <= this.newVer)) {
				this.node = Converter_11.convert(this.node);
			}
			if ((this.oldVer < 12) && (12 <= this.newVer)) {
				this.node = Converter_12.convert(this.node);
			}
			if ((this.oldVer < 13) && (13 <= this.newVer)) {
				this.node = Converter_13.convert(this.node);
			}
			if ((this.oldVer < 14) && (14 <= this.newVer)) {
				this.node = Converter_14.convert(this.node);
			}
			if ((this.oldVer < 15) && (15 <= this.newVer)) {
				this.node = Converter_15.convert(this.node);
			}
			if ((this.oldVer < 16) && (16 <= this.newVer)) {
				this.node = Converter_16.convert(this.node);
			}
			if ((this.oldVer < 17) && (17 <= this.newVer)) {
				this.node = Converter_17.convert(this.node);
			}
			if ((this.oldVer < 18) && (18 <= this.newVer)) {
				this.node = Converter_18.convert(this.node);
			}
			if ((this.oldVer < 19) && (19 <= this.newVer)) {
				this.node = Converter_19.convert(this.node);
			}
			if ((this.oldVer < 20) && (20 <= this.newVer)) {
				this.node = Converter_20.convert(this.node);
			}
			if ((this.oldVer < 21) && (21 <= this.newVer)) {
				this.node = Converter_21.convert(this.node);
			}
			if ((this.oldVer < 22) && (22 <= this.newVer)) {
				this.node = Converter_22.convert(this.node);
			}
			if ((this.oldVer < 23) && (23 <= this.newVer)) {
				this.node = Converter_23.convert(this.node);
			}
			if ((this.oldVer < 24) && (24 == this.newVer)) {
				this.node = Converter_24.convert(this.node, this.file);
			}
			if ((this.oldVer < 25) && (25 <= this.newVer)) {
				this.node = Converter_25.convert(this.node, this.file);
			}
			if ((this.oldVer < 26) && (26 <= this.newVer)) {
				this.node = Converter_26.convert(this.node);
			}
			if ((this.oldVer < 27) && (27 <= this.newVer)) {
				this.node = Converter_27.convert(this.node);
			}
			if ((this.oldVer < 28) && (28 <= this.newVer)) {
				this.node = Converter_28.convert(this.node);
			}
			if ((this.oldVer < 29) && (29 <= this.newVer)) {
				this.node = Converter_29.convert(this.node);
			}
			if ((this.oldVer < 30) && (30 <= this.newVer)) {
				Converter_30.convertOptimizer(this.file);
			}
			if ((this.oldVer < 31) && (31 <= this.newVer)) {
				Converter_31.convertOptimizer(this.file);
			}
			if ((this.oldVer < 32) && (32 <= this.newVer)) {
				Converter_32.convertOptimizer(this.file);
			}
			if ((this.oldVer < 32) && (32 <= this.newVer)) {
				Converter_32.convertOptimizer(this.file);
			}
			if ((this.oldVer < 34) && (34 <= this.newVer)) {
				Converter_34.convertOptimizer(this.file);
			}
			if ((this.oldVer < 35) && (35 <= this.newVer)) {
				Converter_35.convert(this.node);
			}
			if ((this.oldVer < 36) && (36 <= this.newVer)) {
				Converter_36.convert(this.node);
			}
			if ((this.oldVer < 37) && (37 <= this.newVer)) {
				Converter_37.convert(this.node);
			}
			if ((this.oldVer < 40) && (40 <= this.newVer)) {
				Converter_40.convert(this.node);
			}
			if ((this.oldVer < 41) && (41 <= this.newVer)) {
				Converter_41.convert(this.node);
			}
			if ((this.oldVer < 42) && (42 <= this.newVer)) {
				Converter_42.convert(this.node);
			}

			// TODO Insert new Converter here

			// save
			this.saveXML(this.file);
		}
	}

	/**
	 * Load from xml-file version-information
	 * 
	 * @param file
	 *            Project-file
	 * @throws Exception
	 */
	private void loadVersion(MopFile file) throws Exception {

		file.readStream(new MopAccess() {
			@Override
			public void readStream(InputStream is) throws Exception {
				Integer tmp;

				XMLStreamReader xmlReader = XMLInputFactory.newInstance()
						.createXMLStreamReader(is, "utf-8");
				while (xmlReader
						.getEventType() != XMLStreamConstants.START_ELEMENT) {
					if (xmlReader
							.getEventType() == XMLStreamConstants.COMMENT) {
						tmp = XMLConverter.getVersionCode(xmlReader.getText());
						if (tmp != null) {
							XMLConverter.this.oldVer = tmp;
						}
					}
					xmlReader.next();
				}

				xmlReader.close();
			}

			@Override
			public void writeStream(OutputStream os) throws Exception {
			}

		}, "Project.xml");

	}

	/**
	 * Load xml-file into inner structure for editing
	 * 
	 * @param file
	 *            Project-file
	 * @throws Exception
	 */
	private void loadXML(MopFile file) throws Exception {

		file.readStream(this, "Project.xml");

	}

	/**
	 * Save inner structure into xml-file
	 * 
	 * @param file
	 *            Project-file
	 * @throws Exception
	 */
	private void saveXML(MopFile file) throws Exception {

		file.writeStream(this, "Project.xml");

	}

	/**
	 * Read content of mop-file
	 * 
	 * @param is
	 *            Inputstream for reading file-content
	 * @throws Exception
	 */
	@Override
	public void readStream(InputStream is) throws Exception {

		XMLStreamReader xmlReader = XMLInputFactory.newInstance()
				.createXMLStreamReader(is, "utf-8");
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		// read objects
		this.node = new XMLNode(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new XMLStreamException("Undefined error by reading file!");
		}
		xmlReader.close();
	}

	/**
	 * Write content of mop-file
	 * 
	 * @param os
	 *            Outputstream for writing file-content
	 * @throws Exception
	 */
	@Override
	public void writeStream(OutputStream os) throws Exception {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance()
				.createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");
		xmlWriter.writeComment(XMLConverter.getVersionString(this.newVer));

		// write object to File
		this.node.writeXML(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}

	/**
	 * Load structure of file to debugging
	 * 
	 * @return Structure of file
	 * @throws Exception
	 */
	public XMLNode load() throws Exception {
		this.loadXML(this.file);
		return this.node;
	}
}
