/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.event;

import java.awt.BorderLayout;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;

import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.utils.MatlabControl;
import matlabcontrol.MatlabInvocationException;

/**
 * List event detailsof an event
 * 
 * @author Werner Koppelsttter
 */
public class EventDialog extends JDialog {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;
	/** Parent of dialog */
	protected EventView parent;
	/** Event to display */
	protected Event event;
	/** JTable to display stack-trace of event */
	protected JTable table;
	/** Stack-trace-model */
	protected StackTraceModel model;

	/**
	 * Dialog with detail of event
	 * 
	 * @param parent
	 *            Parent of dialog
	 * @param event
	 *            Event to display
	 */
	public EventDialog(EventView parent, Event event) {
		super();

		this.parent = parent;
		this.event = event;

		this.setTitle("Event detail");
		this.setLayout(new BorderLayout());

		JPanel panel = new JPanel();
		panel.setBorder(new EmptyBorder(5, 5, 5, 5));
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));

		// add descriptor
		JPanel panelName = new JPanel();
		panelName.setLayout(new BoxLayout(panelName, BoxLayout.Y_AXIS));
		panelName.add(new JLabel("Type"));
		panelName.add(new JLabel("Date"));
		panelName.add(new JLabel("Identifier"));
		panelName.add(new JLabel("Description"));
		panel.add(panelName);
		panel.add(Box.createHorizontalStrut(10));

		// add values
		JLabel label;
		JPanel panelValue = new JPanel();
		panelValue.setLayout(new BoxLayout(panelValue, BoxLayout.Y_AXIS));

		label = new JLabel(event.getType().toString());
		label.setToolTipText(event.getType().toString());
		panelValue.add(label);

		label = new JLabel(event.getDateString());
		label.setToolTipText(event.getDateString());
		panelValue.add(label);

		label = new JLabel(event.getIdentifier());
		label.setToolTipText(event.getIdentifier());
		panelValue.add(label);

		label = new JLabel(event.getDescription().replace("\n", "; "));
		label.setToolTipText("<html>" + event.getDescription().replace("\n", "<br>") + "</html>");
		panelValue.add(label);

		panel.add(panelValue);
		panel.add(Box.createHorizontalGlue());
		this.add(panel, BorderLayout.NORTH);

		// add stack-trace-list
		this.table = new JTable();
		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.setShowVerticalLines(false);
		this.model = new StackTraceModel();
		this.table.setModel(this.model);

		// add scroll-pane
		JScrollPane pane = new JScrollPane(this.table);
		pane.setBorder(new EmptyBorder(5, 5, 5, 5));
		this.add(pane, BorderLayout.CENTER);

		// Cell renderer
		this.table.setDefaultRenderer(String.class, new StackTraceRenderer());

		// set max width for checkbox column
		this.table.getColumnModel().getColumn(StackTraceColumn.METHODE.getValue()).setPreferredWidth(80);
		this.table.getColumnModel().getColumn(StackTraceColumn.LINE.getValue()).setPreferredWidth(40);
		this.table.getColumnModel().getColumn(StackTraceColumn.LINE.getValue()).setMaxWidth(50);
		this.table.getColumnModel().getColumn(StackTraceColumn.FILE.getValue()).setPreferredWidth(50);
		this.table.getColumnModel().getColumn(StackTraceColumn.CLASS.getValue()).setPreferredWidth(50);

		// mouse-listener
		this.table.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent event) {
				if (event.getClickCount() == 2) {
					EventDialog.this.openStackTrace();
				}
			}

			@Override
			public void mouseReleased(MouseEvent event) {
				this.showPopup(event);
			}

			@Override
			public void mousePressed(MouseEvent event) {
				this.showPopup(event);
			}

			private void showPopup(MouseEvent event) {

				if (event.isPopupTrigger()) {
					// select row
					int row = EventDialog.this.table.rowAtPoint(event.getPoint());
					EventDialog.this.table.getSelectionModel().setSelectionInterval(row, row);

					// menu
					JPopupMenu popup = new JPopupMenu();

					JMenuItem item = new JMenuItem("Open");
					item.addActionListener(new ActionListener() {
						@Override
						public void actionPerformed(ActionEvent event) {
							EventDialog.this.openStackTrace();
						}
					});
					popup.add(item);

					String name = EventDialog.this.event.getStack()[row].getFileName();
					item.setEnabled(name.toLowerCase().endsWith(".m") || name.toLowerCase().endsWith(".py"));

					popup.show(event.getComponent(), event.getX(), event.getY());
				}
			}
		});

		// key-action
		this.table.getInputMap().put(KeyStroke.getKeyStroke("ENTER"), "enter");
		this.table.getActionMap().put("enter", new AbstractAction() {
			private static final long serialVersionUID = 1L;

			@Override
			public void actionPerformed(ActionEvent e) {
				EventDialog.this.openStackTrace();
			}
		});

		this.setPreferredSize(new Dimension(800, 350));
		this.setLocationByPlatform(true);
		this.pack();
		this.model.setEvent(this.event);
		this.setModal(true);
		this.setVisible(true);

	}

	/**
	 * Open stack-trace element
	 */
	private void openStackTrace() {
		final StackTraceElement trace;

		trace = EventDialog.this.event.getStack()[EventDialog.this.table.getSelectedRow()];

		if ((new File(trace.getFileName())).exists()) {
			if (trace.getFileName().toLowerCase().endsWith(".m")) {
				new Thread() {
					@Override
					public void run() {
						try {
							MatlabControl.getProxy().feval("opentoline",
									new Object[] { trace.getFileName(), trace.getLineNumber() });
						} catch (MatlabInvocationException e) {
							e.printStackTrace();
						}
					}
				}.start();
			} else if (trace.getFileName().toLowerCase().endsWith(".py")) {
				if (Desktop.isDesktopSupported()) {
					File file = new File(trace.getFileName());
					if (file.exists()) {
						try {
							Desktop.getDesktop().open(file);
						} catch (IOException e) {
							// XXX winopen ?
						}
					}
				}
			}
		}

	}
}
