/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.tree;

import java.awt.BorderLayout;
import java.rmi.RemoteException;
import java.util.LinkedList;

import javax.swing.DropMode;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ToolTipManager;
import javax.swing.border.EmptyBorder;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;

/**
 * Tree view panel for project
 * 
 * @author Werner Koppelsttter, Markus Priglinger
 */
public class ProjectTreeView extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Project to display */
	protected IProject project = null;

	/** CTree object of the table */
	public ProjectDnDTree tree;

	/** Tree model */
	private ProjectTreeModel model;

	/**
	 * Create a panel to display a project tree
	 */
	public ProjectTreeView() {
		super();

		this.setLayout(new BorderLayout());
		this.tree = new ProjectDnDTree();
		this.model = new ProjectTreeModel();
		this.tree.setModel(this.model);
		// this.model.addTreeModelListener(this.tree);
		this.tree.setBorder(new EmptyBorder(5, 5, 5, 5));
		this.tree.setCellRenderer(new ProjectTreeCellRenderer());

		this.tree.setDragEnabled(true);
		this.tree.setAutoscrolls(true);
		this.tree.getSelectionModel().setSelectionMode(
				TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
		this.tree.setDropMode(DropMode.ON_OR_INSERT);
		this.tree.setTransferHandler(new ProjectTreeTransferHandler());
		this.tree.expandPath(new TreePath(this.model));

		JScrollPane scrollPane = new JScrollPane(this.tree);
		// scrollPane.setPreferredSize(new Dimension(600, 800));
		this.add(scrollPane, BorderLayout.CENTER);
		// enable tool tips
		ToolTipManager.sharedInstance().registerComponent(this.tree);
	}

	/**
	 * Set the project to display
	 * 
	 * @param project
	 *            Project to display
	 */
	public void setProject(IProject project) {

		IParameter selected;

		selected = this.getSelectedParameter();
		this.project = project;
		this.update();

		if (this.project != null) {
			try {
				if ((selected == null)
						|| (this.project.getUniqueObject(selected
								.getIdentifier()) == null)) {
					this.tree.setSelectionRow(0);
				} else {
					this.selectNode(selected.getIdentifier());
				}
			} catch (RemoteException e) {
				// TODO RMI
				this.tree.setSelectionRow(0);
			}
		}
	}

	/**
	 * Update tree
	 */
	protected void update() {

		this.model.setProject(this.project);
	}

	/**
	 * Get selected Parameter
	 * 
	 * @return Selected Parameter
	 */
	public IParameter getSelectedParameter() {

		IParameter param = null;
		if (this.tree.getSelectionCount() > 0) {
			param = (IParameter) this.tree.getSelectionPath()
					.getLastPathComponent();
		}
		return param;
	}

	/**
	 * Get selected Parameters
	 * 
	 * @return Selected Parameters
	 */
	public IParameter[] getSelectedParameters() {

		TreePath[] paths;
		IParameter[] params = null;

		if (this.tree.getSelectionCount() > 0) {
			paths = this.tree.getSelectionPaths();
			params = new IParameter[paths.length];

			for (int i = 0; i < paths.length; i++) {
				params[i] = (IParameter) paths[i].getLastPathComponent();
			}
		}
		return params;
	}

	/**
	 * Select node with given identifier
	 * 
	 * @param identifier
	 *            Identifier of node
	 */
	public void selectNode(String identifier) {
		IParameter param;
		LinkedList<IParameter> pathList = new LinkedList<IParameter>();
		TreePath path;

		if (this.project != null) {
			try {
				param = this.project.getUniqueObject(identifier);
				while (param != null) {
					pathList.addFirst(param);
					param = param.getParentParameter();
				}
			} catch (RemoteException e) {
				// TODO RMI
			}

			if (pathList.size() > 0) {
				path = new TreePath(pathList.toArray());
				this.tree.setSelectionPath(path);
				this.tree.expandPath(path);
			}
		}
	}
}
