/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.structure.database;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.LinkedList;

import at.lcm.bu21.magopt.structure.event.EventLog;
import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Manager the connections to database
 * 
 * @author Werner Koppelsttter
 */
public class DatabaseConnection implements MaterialConnection {

	/** Database connection string */
	private String database = null;

	/** Database version to link */
	private String version = "1.5";

	/** User for connection */
	private String user;

	/** Password for connection */
	private String pwd;

	/** Connection to database */
	private Connection conn;

	/**
	 * Create conneciton to database with standard-user
	 * 
	 * @throws Exception
	 */
	public DatabaseConnection() throws Exception {
		this("magopt_basic", "magopt");
	}

	/**
	 * Create conneciton to database
	 * 
	 * @param user
	 *            User-name
	 * @param password
	 *            Password
	 * @throws Exception
	 */
	public DatabaseConnection(String user, String password) throws Exception {
		ResultSet rs = null;

		this.user = user;
		this.pwd = password;
		this.database = "jdbc:mariadb://"
				+ MagOptConfig.getConfig("DatabaseServer") + "/magopt_db";

		try {
			Class.forName("org.mariadb.jdbc.Driver");
			this.conn = DriverManager.getConnection(this.database + "?user="
					+ this.user + "&password=" + this.pwd);

			rs = this.conn.createStatement().executeQuery(
					"SELECT version.version FROM version ORDER By version.id");
			while (rs.next()) {
				if (this.version
						.compareTo(rs.getString("version.version")) != 0) {
					throw new Exception(
							"Database and Magopt version are different!\n"
									+ "Please update magopt");
				}
			}
		} catch (ClassNotFoundException e) {
			throw new Exception("Database-driver not found!");
			// } catch (SQLException e) {
			// throw new Exception("Error by connecting to database!");
		} finally {
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				}
				rs = null;
			}
		}
	}

	/**
	 * Check if the dataset is available for user
	 * 
	 * @param datasetID
	 *            Dataset-ID
	 * @return Is dataset available
	 */
	protected boolean checkDataset(long datasetID) {
		boolean valid = false;
		ResultSet rs = null;
		String query;

		try {
			query = "SELECT dataset.dataset_id "
					+ "FROM (((dataset INNER JOIN material "
					+ "ON dataset.fk_material=material.material_id) "
					+ "INNER JOIN magoptuser_has_material "
					+ "ON material.material_id=magoptuser_has_material.fk_material) "
					+ "INNER JOIN magoptuser "
					+ "ON magoptuser_has_material.fk_magoptuser=magoptuser.id) "
					+ "WHERE (dataset.active=True) AND (material.active=True) "
					+ "AND (dataset.dataset_id='" + datasetID + "') "
					+ "AND (magoptuser.name='" + this.user + "')";
			rs = this.conn.createStatement().executeQuery(query);

			if (rs.next()) {
				valid = true;
			}
		} catch (Exception ex) {
		} finally {
			// release resources
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				} // ignore

				rs = null;
			}
		}

		return valid;
	}

	@Override
	public DBItem[] getMaterialItems(long datasetID) {
		DBItem[] items = new DBItem[] { new DBItem(), new DBItem(),
				new DBItem() };
		ResultSet rs = null;
		String query;

		if (this.checkDataset(datasetID)) {
			try {
				query = "SELECT dataset.dataset_id, dataset.description, material.material_id, "
						+ "material.name, material_class.id, material_class.name "
						+ "FROM ((dataset INNER JOIN material "
						+ "ON dataset.fk_material=material.material_id) "
						+ "INNER JOIN material_class "
						+ "ON material.fk_material_class=material_class.id) "
						+ "WHERE (dataset.active=True) AND (material.active=True) "
						+ "AND (dataset.dataset_id='" + datasetID + "')";
				rs = this.conn.createStatement().executeQuery(query);

				while (rs.next()) {
					items[0] = new DBItem(rs.getLong("material_class.id"),
							rs.getString("material_class.name"));
					items[1] = new DBItem(rs.getLong("material.material_id"),
							rs.getString("material.name"));
					items[2] = new DBItem(rs.getLong("dataset.dataset_id"),
							rs.getString("dataset.description"));
				}
			} catch (Exception ex) {
				ex.printStackTrace();
			} finally {
				// release resources
				if (rs != null) {
					try {
						rs.close();
					} catch (SQLException sqlEx) {
					} // ignore

					rs = null;
				}
			}
		}

		return items;

	}

	@Override
	public DBItem[] getMaterialClassList() {
		ResultSet rs = null;
		LinkedList<DBItem> list = new LinkedList<DBItem>();

		list.add(new DBItem());
		try {
			rs = this.conn.createStatement().executeQuery(
					"SELECT material_class.id, material_class.name "
							+ "FROM material_class");
			while (rs.next()) {
				list.add(new DBItem(rs.getLong("material_class.id"),
						rs.getString("material_class.name")));
			}
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			// release resources
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				} // ignore

				rs = null;
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public DBItem[] getMaterialList(DBItem materialClass) {
		ResultSet rs = null;
		LinkedList<DBItem> list = new LinkedList<DBItem>();
		String query;

		list.add(new DBItem());
		try {

			query = "SELECT material.material_id, material.name "
					+ "FROM ((material INNER JOIN magoptuser_has_material ON "
					+ "material.material_id=magoptuser_has_material.fk_material) "
					+ "INNER JOIN magoptuser ON "
					+ "magoptuser_has_material.fk_magoptuser=magoptuser.id) "
					+ "WHERE (material.active=True) ";
			if ((materialClass != null) && (materialClass.getID() >= 0)) {
				query += "AND (material.fk_material_class='"
						+ materialClass.getID() + "') ";
			}
			query += "AND (magoptuser.name='" + this.user + "') "
					+ "ORDER BY material.name";
			rs = this.conn.createStatement().executeQuery(query);

			while (rs.next()) {
				list.add(new DBItem(rs.getLong("material.material_id"),
						rs.getString("material.name")));
			}
		} catch (Exception ex) {
			ex.printStackTrace();
		} finally {
			// release resources
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				} // ignore

				rs = null;
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public DBItem[] getDatasetList(DBItem material) {
		ResultSet rs = null;
		LinkedList<DBItem> list = new LinkedList<DBItem>();
		String query;

		list.add(new DBItem());
		if ((material != null) && (material.getID() >= 0)) {
			try {
				query = "SELECT dataset.dataset_id, dataset.description "
						+ "FROM (((dataset INNER JOIN material "
						+ "ON dataset.fk_material=material.material_id) "
						+ "INNER JOIN magoptuser_has_material "
						+ "ON material.material_id=magoptuser_has_material.fk_material) "
						+ "INNER JOIN magoptuser "
						+ "ON magoptuser_has_material.fk_magoptuser=magoptuser.id) "
						+ "WHERE (dataset.active=True) AND (material.active=True)"
						+ "AND (dataset.fk_material='" + material.getID()
						+ "') " + "AND (magoptuser.name='" + this.user + "')"
						+ "ORDER BY dataset.description";
				rs = this.conn.createStatement().executeQuery(query);

				while (rs.next()) {
					list.add(new DBItem(rs.getLong("dataset.dataset_id"),
							rs.getString("dataset.description")));
				}
			} catch (Exception ex) {
				ex.printStackTrace();
			} finally {
				// release resources
				if (rs != null) {
					try {
						rs.close();
					} catch (SQLException sqlEx) {
					} // ignore

					rs = null;
				}
			}
		}

		return list.toArray(new DBItem[0]);
	}

	@Override
	public LinkedList<MaterialProperty<Double>> getProperties(long datasetID,
			EventLog log) {
		ResultSet rs = null;
		LinkedList<MaterialProperty<Double>> list = new LinkedList<MaterialProperty<Double>>();
		String query;

		if (this.checkDataset(datasetID)) {
			try {
				query = "SELECT value.value, property.symbol, property.name, "
						+ "property.tex_string, property.locked, unit.symbol, "
						+ "property_group.name "
						+ "FROM ((((dataset INNER JOIN value "
						+ "ON dataset.dataset_id=value.fk_dataset) "
						+ "INNER JOIN property ON value.fk_property=property.id) "
						+ "INNER JOIN unit ON value.fk_unit=unit.id) "
						+ "INNER JOIN property_group "
						+ "ON property.fk_property_group=property_group.id) "
						+ "WHERE ((dataset.dataset_id='" + datasetID
						+ "') AND (dataset.active=True) AND (value.active=True))";
				rs = this.conn.createStatement().executeQuery(query);

				while (rs.next()) {
					try {
						list.add(new MaterialProperty<Double>(
								rs.getString("property.symbol"),
								rs.getString("property.name"),
								rs.getString("property_group.name"),
								rs.getDouble("value.value"),
								rs.getString("unit.symbol"),
								rs.getString("property.tex_string"),
								rs.getBoolean("property.locked"), false));
					} catch (Exception ex) {
						log.error("", ex.getMessage(), ex.getStackTrace());
					}

				}
			} catch (SQLException ex) {
				log.error("", ex.getMessage(), ex.getStackTrace());
			} finally {
				// release resources
				if (rs != null) {
					try {
						rs.close();
					} catch (SQLException sqlEx) {
					} // ignore

					rs = null;
				}
			}
		}

		return list;
	}

	@Override
	public LinkedList<MaterialDiagram> getDiagrams(long datasetID,
			EventLog log) {
		ResultSet rs = null;
		LinkedList<MaterialDiagram> list = new LinkedList<MaterialDiagram>();
		MaterialDiagram diagram;
		String query, name;
		LinkedList<MaterialProperty<Double[]>> axis;

		if (this.checkDataset(datasetID)) {
			try {
				query = "SELECT diagram.description, diagram.diagram_id "
						+ "FROM (dataset INNER JOIN diagram "
						+ "ON dataset.dataset_id=diagram.fk_dataset) "
						+ "WHERE ((dataset.dataset_id='" + datasetID
						+ "') AND (dataset.active=True) AND (diagram.active=True))";
				rs = this.conn.createStatement().executeQuery(query);

				while (rs.next()) {
					try {
						axis = this.getDiagramAxis(log,
								rs.getLong("diagram.diagram_id"));
						name = "";
						for (MaterialProperty<Double[]> axes : axis) {
							if (axes.getFixed()) {
								name += "at";
							} else {
								if (name.length() != 0) {
									name += "vs";
								}
							}
							name += axes.getName();

						}
						diagram = new MaterialDiagram(name,
								rs.getString("diagram.description"), axis);

						list.add(diagram);
					} catch (Exception ex) {
						log.error("", ex.getMessage(), ex.getStackTrace());
					}

				}
			} catch (SQLException ex) {
				log.error("", ex.getMessage(), ex.getStackTrace());
			} finally {
				// release resources
				if (rs != null) {
					try {
						rs.close();
					} catch (SQLException sqlEx) {
					} // ignore

					rs = null;
				}
			}
		}

		return list;
	}

	/**
	 * Get diagram axis and values
	 * 
	 * @param log
	 *            Logging for exceptions
	 * @param diagram
	 *            Database identifier of diagram
	 * @return List of axis
	 */
	protected LinkedList<MaterialProperty<Double[]>> getDiagramAxis(
			EventLog log, long diagram) {
		ResultSet rs = null;
		String query;
		LinkedList<MaterialProperty<Double[]>> axis = new LinkedList<MaterialProperty<Double[]>>();
		Double[] values;

		try {
			query = "SELECT diagram_axis.id, property.symbol, property.name, "
					+ "property.tex_string, property.locked, unit.symbol, "
					+ "diagram_axis.axis_fixed "
					+ "FROM ((diagram_axis INNER JOIN property "
					+ "ON diagram_axis.fk_property=property.id) "
					+ "INNER JOIN unit ON diagram_axis.fk_unit=unit.id) "
					+ "WHERE (((diagram_axis.fk_diagram)='" + diagram + "') "
					+ "AND ((diagram_axis.active)=True)) "
					+ "ORDER BY diagram_axis.axis_index";
			rs = this.conn.createStatement().executeQuery(query);

			while (rs.next()) {
				try {
					values = this.getDiagramValues(log,
							rs.getLong("diagram_axis.id"));
					axis.add(new MaterialProperty<Double[]>(
							rs.getString("property.symbol"),
							rs.getString("property.name"), "diagram", values,
							rs.getString("unit.symbol"),
							rs.getString("property.tex_string"),
							rs.getBoolean("property.locked"),
							rs.getBoolean("diagram_axis.axis_fixed")));
				} catch (Exception ex) {
					log.error("", ex.getMessage(), ex.getStackTrace());
				}
			}

			// } catch (SQLException ex) {
			// // XXX handle any errors
			// System.out.println("SQLException: " + ex.getMessage());
			// System.out.println("SQLState: " + ex.getSQLState());
			// System.out.println("VendorError: " + ex.getErrorCode());
		} catch (Exception ex) {
			log.error("", ex.getMessage(), ex.getStackTrace());
		} finally {
			// release resources
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				} // ignore

				rs = null;
			}
		}
		return axis;
	}

	/**
	 * Get diagram values
	 * 
	 * @param log
	 *            Logging for exceptions
	 * @param axis
	 *            Axis of diagram
	 * @return List of values
	 */
	protected Double[] getDiagramValues(EventLog log, long axis) {
		ResultSet rs = null;
		String query;
		LinkedList<Double> value;

		value = new LinkedList<Double>();
		try {
			query = "SELECT diagram_point.value " + "FROM diagram_point "
					+ "WHERE (((diagram_point.fk_diagram_axis)='" + axis
					+ "') AND ((diagram_point.active)=True)) "
					+ "ORDER BY diagram_point.point_index";
			rs = this.conn.createStatement().executeQuery(query);

			while (rs.next()) {
				value.add(rs.getDouble("diagram_point.value"));
			}

			// } catch (SQLException ex) {
			// // XXX handle any errors
			// System.out.println("SQLException: " + ex.getMessage());
			// System.out.println("SQLState: " + ex.getSQLState());
			// System.out.println("VendorError: " + ex.getErrorCode());
		} catch (Exception ex) {
			log.error("", ex.getMessage(), ex.getStackTrace());
		} finally {
			// release resources
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				} // ignore

				rs = null;
			}
		}

		return value.toArray(new Double[0]);
	}

	/**
	 * Get database-identifier for parameter-name
	 * 
	 * @param name
	 *            Name of parameter
	 * @return Identifier of parameter
	 * @throws SQLException
	 */
	public long getDBParameterId(String name) throws SQLException {
		long id = -1;
		ResultSet rs = null;

		try {
			rs = this.conn.createStatement()
					.executeQuery("SELECT parameter.id FROM parameter "
							+ "WHERE (parameter.symbol='" + name + "')");
			while (rs.next()) {
				id = rs.getLong("parameter.id");
			}
		} catch (SQLException e) {
			throw e;
		} finally {
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				}
				rs = null;
			}
		}

		return id;
	}

	/**
	 * Get parameter-name for identifier
	 * 
	 * @param id
	 *            Identifier of parameter
	 * @return Name of parameter
	 * @throws SQLException
	 */
	public String getDBParameterName(long id) throws SQLException {
		String name = null;
		ResultSet rs = null;

		try {
			rs = this.conn.createStatement()
					.executeQuery("SELECT parameter.symbol FROM parameter "
							+ "WHERE (parameter.id=" + id + ")");
			while (rs.next()) {
				name = rs.getString("parameter.symbol");
			}
		} catch (SQLException e) {
			throw e;
		} finally {
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				}
				rs = null;
			}
		}

		return name;
	}

	/**
	 * Get parameter-values
	 * 
	 * @param parameterID
	 *            Paramter-id
	 * 
	 * @return Parameter-values
	 * @throws Exception
	 */
	public Hashtable<String, String> getDBParameter(long parameterID)
			throws Exception {
		ResultSet rs = null;
		Hashtable<String, String> table = new Hashtable<String, String>();

		try {
			rs = this.conn.createStatement().executeQuery(
					"SELECT parameter.id, parameter.symbol, parameter.name, parameter.tex_string, unit.symbol "
							+ "FROM parameter LEFT JOIN unit ON parameter.fk_unit = unit.id "
							+ "WHERE (((parameter.id)=" + parameterID + "))");
			while (rs.next()) {
				table.put("Name", rs.getString("parameter.symbol"));
				table.put("Description", rs.getString("parameter.name"));
				table.put("Tex", rs.getString("parameter.tex_string"));
				table.put("Unit", rs.getString("unit.symbol"));
			}
		} catch (Exception e) {
			throw e;
		} finally {
			if (rs != null) {
				try {
					rs.close();
				} catch (SQLException sqlEx) {
				}
				rs = null;
			}
		}

		return table;
	}

	@Override
	public void close() {
		try {
			this.conn.close();
		} catch (SQLException e) {
		}
	}
}
