/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.magopt.structure.xsd.XMLValidator;

/**
 * Manage access to archive-files of a mop-file
 * 
 * @author Werner Koppelsttter
 */
public class OptimizationArchive {

	/** Temporary field for reading and writing archives */
	private TStruct data;

	/** Access-class for reading/writing of archive data */
	private MopAccess access = new MopAccess() {

		/**
		 * Read structure from archive
		 * 
		 * @param is
		 *            InputStream of archive
		 */
		@Override
		public void readStream(InputStream is) throws Exception {

			XMLInputFactory inputFactory = XMLInputFactory.newInstance();
			XMLStreamReader reader = inputFactory.createXMLStreamReader(is);

			while (reader.getEventType() != XMLStreamConstants.START_ELEMENT) {
				reader.next();
			}

			OptimizationArchive.this.data = (TStruct) TData.xml2obj(reader);

			if (reader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
				throw new IOException("Undefined error while reading file!");
			}

			reader.close();
		}

		/**
		 * Write structure to archive
		 * 
		 * @param os
		 *            OutputStream of archive
		 */
		@Override
		public void writeStream(OutputStream os) throws Exception {

			XMLOutputFactory outputFactory = XMLOutputFactory.newInstance();
			XMLStreamWriter writer = outputFactory.createXMLStreamWriter(os);

			writer.writeStartDocument();
			OptimizationArchive.this.data.obj2xml(writer);
			writer.writeEndDocument();
			writer.close();

		}
	};

	/** Mop-file with the archives */
	protected MopFile mop;

	/**
	 * Constructor of OptimizerArchive
	 * 
	 * @param mop
	 *            mop-file
	 */
	public OptimizationArchive(MopFile mop) {
		this.mop = mop;
	}

	/**
	 * Get list of archive names
	 * 
	 * @return Archive-file list
	 * @throws Exception
	 */
	public String[] getArchives() throws Exception {
		LinkedList<String> list = new LinkedList<String>();
		String[] filelist;

		filelist = this.mop.getFileNames();
		for (String name : filelist) {
			if (name.startsWith(MopFile.optimizationTag)
					&& name.endsWith(".xml")) {
				list.add(name.substring(MopFile.optimizationTag.length(),
						name.length() - 4));
			}
		}

		return list.toArray(new String[0]);
	}

	/**
	 * Read archive with the given name
	 * 
	 * @param name
	 *            Name of archive
	 * @return Archive data structure
	 * @throws Exception
	 */
	public TStruct read(String name) throws Exception {
		TStruct data;

		// validation of File
		new XMLValidator().validateMop(this.mop, MopFile.optimizationTag + name
				+ ".xml", "TStruct.xsd");

		this.mop.readStream(this.access, MopFile.optimizationTag + name
				+ ".xml");

		data = this.data;
		this.data = null;
		return data;
	}

	/**
	 * Write archive with the given name
	 * 
	 * @param name
	 *            Name of archive
	 * @param data
	 *            Structure with archive data
	 * @throws Exception
	 */
	public void write(String name, TStruct data) throws Exception {
		this.data = data;

		this.mop.writeStream(this.access, MopFile.optimizationTag + name
				+ ".xml");

		this.data = null;
	}

	/**
	 * Remove archive from file
	 * 
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	public void remove(String name) throws Exception {
		this.mop.remove(MopFile.optimizationTag + name + ".xml");
	}
}
