/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.File;
import java.util.LinkedList;

import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.structure.event.EventLog;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.bu21.optim.core.SavePoint.SavePointType;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.event.OptimizerListener;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.project.ComputeProject;
import at.lcm.bu21.optim.core.project.SelectOptimum;
import at.lcm.bu21.optim.core.project.XLSWrite;
import magopt.utils.MopFile;
import main.ocvwrap.java.OpenCVDynamicBinder;

/**
 * Controller of optimization
 * 
 * @author Werner Koppelsttter
 */
public class Optimization {

	/** List of listeners */
	private LinkedList<OptimizerListener> listeners = new LinkedList<OptimizerListener>();

	/** Event-logger */
	private EventLog logger = new EventLog(1000);

	/** Optimization problem */
	private Problem problem;

	/** Project optimization object */
	private ComputeProject calcProject;

	/** Project mapping management object */
	private MappingManager calcMapping;

	/** Write to excel is active */
	private boolean writeExcel = false;

	/** Select pareto is active */
	private boolean selectPareto = false;

	/** Is test-mode enabled */
	private boolean testMode = false;

	/**
	 * Create controller of optimization
	 * 
	 * @param problem
	 *            Problem to optimize
	 * @throws Exception
	 */
	protected Optimization(Problem problem) throws Exception {
		this.problem = problem;
		this.testMode = Boolean.parseBoolean(MagOptConfig.getConfig("OptimizerTestMode"));
		this.calcProject = new ComputeProject(this);

		try {
			OpenCVDynamicBinder.loadLibrary();
		} catch (Throwable e) {
			this.addLogError("Could not load opencv library!");
			// throw e;
		}
		this.calcMapping = new MappingManager(this);
	}

	/**
	 * Check if test-mode is enabled
	 * 
	 * @return Is test-mode
	 */
	public boolean isTestMode() {
		return this.testMode;
	}

	/**
	 * Get current problem
	 * 
	 * @return Problem
	 */
	public Problem getProblem() {
		return this.problem;
	}

	/**
	 * Fire optimizer event
	 * 
	 * @param event
	 *            Optimizer event
	 */
	public void fireOptmizerPerformed(OptimizerEvent event) {
		synchronized (this.listeners) {
			for (OptimizerListener listener : this.listeners) {
				listener.optimizerPerformed(event);
			}
		}
	}

	/**
	 * Add optimizer listener
	 * 
	 * @param listener
	 *            Optimizier-listener to add
	 */
	public void addOptimizerListener(OptimizerListener listener) {
		synchronized (this.listeners) {
			if (!this.listeners.contains(listener)) {
				this.listeners.add(listener);
			}
		}
	}

	/**
	 * Remove optimizer listener
	 * 
	 * @param listener
	 *            Listener to remove
	 */
	public void removeOptimizerListener(OptimizerListener listener) {
		synchronized (this.listeners) {
			this.listeners.remove(listener);
		}
	}

	/**
	 * Get list of optimizer listener
	 * 
	 * @return List of listener
	 */
	public OptimizerListener[] getListners() {
		synchronized (this.listeners) {
			return this.listeners.toArray(new OptimizerListener[0]);
		}
	}

	/**
	 * Add log-message
	 * 
	 * @param log
	 *            Message
	 */
	public void addLog(String log) {
		// FIXME logging
		System.out.println(log);
		this.logger.info("", log, new StackTraceElement[0]);
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.LOG));
	}

	/**
	 * Add log-message
	 * 
	 * @param log
	 *            Message
	 */
	public void addLogError(String log) {
		this.addLogError(log, new StackTraceElement[0]);
	}

	/**
	 * Add log-message
	 * 
	 * @param log
	 *            Message
	 * @param stack
	 *            List of trace-elements
	 */
	public void addLogError(String log, StackTraceElement[] stack) {
		// FIXME logging
		System.out.println(log);
		this.logger.error("", log, stack);
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.LOG));
	}

	/**
	 * Clear logger
	 */
	public void clearLog() {
		this.logger.clear();
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.LOG));
	}

	/**
	 * Get event-logger
	 * 
	 * @return Event-log
	 */
	public EventLog getEventLog() {
		return this.logger;
	}

	/**
	 * Get control of project optimization
	 * 
	 * @return Project-control
	 */
	public ComputeProject getProjectControl() {
		return this.calcProject;
	}

	/**
	 * Get control of mapping optimization
	 * 
	 * @return Mapping-control
	 */
	public MappingManager getMappingControl() {
		return this.calcMapping;
	}

	/**
	 * Shutdown optimizer
	 */
	public void shutdown() {
		this.calcProject.shutdown();
		this.calcMapping.shutdown();

		if (!this.isTestMode()) {
			try {
				this.calcProject.getProblem().getArchive().save();
			} catch (Exception ex) {
				ex.printStackTrace();
			}
		}
	}

	/**
	 * Init project buffer
	 * 
	 * @param num
	 *            Number of individual for initialize
	 */
	public void initProjectBuffer(int num) {
		this.calcProject.initBuffer(num);
	}

	/**
	 * Clear project-buffer
	 */
	public void clearProjectBuffer() {
		this.calcProject.clearBuffer();
	}

	/**
	 * Add to project-buffer
	 * 
	 * @param inds
	 *            Individuals
	 */
	public void add2Buffer(Individual[] inds) {
		this.calcProject.add2Buffer(inds);
	}

	/**
	 * Start project optimization
	 */
	public void runProject() {
		this.calcProject.runOptimizer();
	}

	/**
	 * Finish project optimization
	 */
	public void finishProject() {
		this.calcProject.finishOptimizer();
	}

	/**
	 * Stop project optimization
	 */
	public void stopProject() {
		this.calcProject.stopOptimizer();
	}

	/**
	 * Get status of project calculation
	 * 
	 * @return Status
	 */
	public String getProjectStatus() {
		return this.calcProject.getStatus();
	}

	/**
	 * Set status of recompute mapping based
	 * 
	 * @param recomp
	 *            Recompute
	 */
	public void setRecompute(boolean recomp) {
		this.calcProject.setRecompute(recomp);
	}

	/**
	 * Get status of recompute
	 * 
	 * @return Recompute
	 */
	public boolean isRecompute() {
		return this.calcProject.isRecompute();
	}

	/**
	 * Create mapping optimization
	 */
	public void createMapping() {
		this.calcMapping.runGeneralMappingCreation();
	}

	/**
	 * Create mapping optimization
	 * 
	 * @param chosenInd
	 *            Individual
	 */
	public void initializeSensitivityAnalysis(Individual chosenInd) {
		this.calcMapping.initializeSensitivityAnalysis(chosenInd);
	}

	/**
	 * Run mapping optimization
	 */
	public void runMapping() {
		this.calcMapping.runMappingBasedOptimization();
	}

	/**
	 * Write pareto to excel
	 * 
	 * @param excel
	 *            Excel-file
	 * @return File
	 * @throws Exception
	 */
	public File writeProjectToExcel(File excel) throws Exception {
		File file;

		this.writeExcel = true;
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

		file = XLSWrite.writeXLSReport(this.getProblem(), excel);

		this.writeExcel = false;
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

		return file;
	}

	/**
	 * Select pareto
	 * 
	 * @param mop
	 *            New mop-file
	 * @param list
	 *            List of individuals
	 * @param selected
	 *            Index of selected
	 * @param suppressed
	 *            Index of suppressed
	 * @param doc
	 *            Create doc-files
	 * @return Created project
	 * @throws Exception
	 */
	public Project selectPareto(MopFile mop, Individual[] list, int[] selected, int[] suppressed, boolean doc)
			throws Exception {
		Project project;

		this.selectPareto = true;
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

		project = new SelectOptimum().createProject(mop, this.getProblem(), list, selected, suppressed, doc);

		this.selectPareto = false;
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));

		return project;
	}

	/**
	 * Get write to excel is active
	 * 
	 * @return Is active
	 */
	public boolean getWriteExcel() {
		return this.writeExcel;
	}

	/**
	 * Get select paretp is active
	 * 
	 * @return Is active
	 */
	public boolean getSelectPareto() {
		return this.selectPareto;
	}

	/**
	 * Refresh display
	 * 
	 * @throws Exception
	 */
	public void refreshDisplay() throws Exception {
		this.calcProject.getProblem().createSavePoint(SavePointType.USER, true, this.calcProject.getNumberValid(),
				this.calcProject.getNumberInvalid(), this.calcProject.getNumberError());
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
		this.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
	}

	/**
	 * Add user plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void addUserPlot(String name) {
		this.getProblem().addUserPlot(name);
	}

	/**
	 * Remove user plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void rmUserPlot(String name) {
		this.getProblem().rmUserPlot(name);
	}

	/**
	 * Is individual (variables) in progress (in computation or buffer)
	 * 
	 * @param ind
	 *            Individual to check (variables)
	 * @return Is in progress
	 */
	public boolean isInProgress(Individual ind) {
		return this.calcProject.isInProgress(ind);
	}
}
