/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import at.lcm.bu21.optim.core.mapping.jobs.MappingJobDispatcher;
import at.lcm.bu21.optim.core.mapping.utils.comparators.MappingTaskEstimatedExecutionTimeComparator;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.TExperimentType;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTask;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingTaskBatch;
import main.ocvwrap.java.OpenCVDataSource;
import main.ocvwrap.java.OpenCVMLPPredictor;
import main.ocvwrap.java.OpenCVUtils;

public class MappingCreationThread extends Thread {

	private final Map<Integer, TMappingTaskBatch> taskBatches = new HashMap<Integer, TMappingTaskBatch>();

	private final Map<String, Map<Integer, OpenCVDataSource>> cvDatasets = new HashMap<String, Map<Integer, OpenCVDataSource>>();

	private final List<TMappingTask> uncompletedTasks = new ArrayList<TMappingTask>();
	private final ConcurrentLinkedQueue<TMappingTask> completedTasks;
	private final Set<String> targetNames = new TreeSet<String>();
	private final Map<String, String> jobIds = new HashMap<String, String>();
	private final Map<String, File> inputFiles = new HashMap<String, File>();
	private final Map<String, File> outputFiles = new HashMap<String, File>();
	private final Map<String, File> modelFiles = new HashMap<String, File>();
	private final MappingJobDispatcher jobDispatcher = new MappingJobDispatcher();
	private final MappingHolder parentMappingHolder;
	private OpenCVDataSource baseDataset;
	private int totalNrOfTasks;
	private int nrOfTaskBatches;
	private int trainingSetSize;
	private int nrOfCrossValidationFolds;
	private int nrOfDatasetOutputs;
	private OpenCVMLPPredictor trainedPredictionModel;
	private TMappingTask trainingTask;
	private final String experimentFolder;
	private TExperimentType experimentType;
	private String baseDatasetFile;
	private final AtomicBoolean stopMappingCreation;
	private final AtomicBoolean hasUnremovedJobs;
	private final Integer maxDuration;

	public MappingCreationThread(String experimentFolder, ConcurrentLinkedQueue<TMappingTask> completedTasksQueue,
			MappingHolder mappingHolder, AtomicBoolean stopMappingCreation, AtomicBoolean hasUnremovedJobs,
			Integer maxDuration) {
		super();
		this.experimentFolder = experimentFolder;
		this.completedTasks = completedTasksQueue;
		this.parentMappingHolder = mappingHolder;
		this.stopMappingCreation = stopMappingCreation;
		this.hasUnremovedJobs = hasUnremovedJobs;
		this.maxDuration = maxDuration;
	}

	private void createAllPossibleTasksForExperiment(List<List<String>> generalParamList) {

		if (generalParamList.size() == 0) {
			return;
		}

		int totalNrOfTasks = 1;
		for (List<String> parameterVector : generalParamList) {
			totalNrOfTasks *= (parameterVector.size() - 1);
		}
		List<List<String>> combinationsArray = new ArrayList<List<String>>();
		for (int i = 0; i < totalNrOfTasks; i++) {
			combinationsArray.add(new ArrayList<String>());
		}

		int iterationFactor = 1;
		int repeatFactor = totalNrOfTasks;
		int discretizationFactor = 1;
		for (List<String> parameterVector : generalParamList) {
			discretizationFactor = parameterVector.size() - 1;
			repeatFactor = repeatFactor / discretizationFactor;
			int insertPos = 0;
			for (int i = 0; i < iterationFactor; i++) {
				for (int j = 1; j <= discretizationFactor; j++) {
					for (int k = 0; k < repeatFactor; k++) {
						combinationsArray.get(insertPos++).add(parameterVector.get(j));
					}
				}
			}
			iterationFactor *= discretizationFactor;
		}

		for (int i = 0; i < combinationsArray.size(); i++) {
			TMappingTask opencvTask = new TMappingTask();
			int j = 0;
			for (List<String> parameterVector : generalParamList) {
				opencvTask.getParameters().put(parameterVector.get(0), combinationsArray.get(i).get(j++));
			}
			this.uncompletedTasks.add(opencvTask);
		}
	}

	private void createTargetDataset(OpenCVDataSource dataset, String targetName) {
		dataset.setTargetStatusOfLastAttributes(this.nrOfDatasetOutputs, true);
		dataset.setUsableStatusOfLastAttributes(this.nrOfDatasetOutputs, false);
		dataset.unsetAsUsable(TMappingTask.TParameter.IND_ID.getValue().toUpperCase());
		dataset.setAsUsable(targetName.toUpperCase());
	}

	private void createCrossValidationDatasets() {

		OpenCVDataSource randomizedTrainingSet = new OpenCVDataSource(this.baseDataset, true);
		Random rndGen = new Random(new Random(12345678L).nextLong());
		randomizedTrainingSet.randomizeDataSamples(rndGen);

		for (String targetName : this.targetNames) {
			OpenCVDataSource targetDataset = new OpenCVDataSource(randomizedTrainingSet, true);
			this.createTargetDataset(targetDataset, targetName);

			int foldSize = targetDataset.getNrOfOpenCVSamples() / this.nrOfCrossValidationFolds;
			int largerFoldsCount = targetDataset.getNrOfOpenCVSamples() % this.nrOfCrossValidationFolds;
			int currentCopyIndex = 0;
			Map<Integer, OpenCVDataSource> testFolds = new TreeMap<Integer, OpenCVDataSource>();
			Map<Integer, OpenCVDataSource> trainingFolds = new TreeMap<Integer, OpenCVDataSource>();
			for (int i = 0; i < this.nrOfCrossValidationFolds; i++) {
				if (largerFoldsCount > 0) {
					largerFoldsCount--;
					OpenCVDataSource foldTestSet = new OpenCVDataSource(targetDataset, false);
					foldTestSet.addSamplesFromSource(targetDataset, currentCopyIndex, currentCopyIndex + foldSize);
					testFolds.put(i, foldTestSet);
					currentCopyIndex += foldSize + 1;
				} else {
					OpenCVDataSource foldTestSet = new OpenCVDataSource(targetDataset, false);
					foldTestSet.addSamplesFromSource(targetDataset, currentCopyIndex,
							(currentCopyIndex + foldSize) - 1);
					testFolds.put(i, foldTestSet);
					currentCopyIndex += foldSize;
				}
			}

			for (int i = 0; i < this.nrOfCrossValidationFolds; i++) {
				OpenCVDataSource foldTrainingSet = new OpenCVDataSource(targetDataset, false);
				for (int j = 0; j < this.nrOfCrossValidationFolds; j++) {
					if (i != j) {
						foldTrainingSet.addSamplesFromSource(testFolds.get(j), 0,
								testFolds.get(j).getNrOfOpenCVSamples() - 1);
					}
				}
				trainingFolds.put(i, foldTrainingSet);
			}

			this.cvDatasets.put(TMappingConstants.TRAINING_PREFIX + targetName.toUpperCase(), trainingFolds);
			this.cvDatasets.put(TMappingConstants.TEST_PREFIX + targetName.toUpperCase(), testFolds);
		}
	}

	private void createTaskBatches() {
		for (int i = 0; i < this.nrOfTaskBatches; i++) {
			this.taskBatches.put(i, new TMappingTaskBatch());
		}
		int currTaskBatchIndex = 0;
		for (TMappingTask task : this.uncompletedTasks) {
			this.taskBatches.get(currTaskBatchIndex++).getTaskList().add(new TMappingTask(task));
			if (currTaskBatchIndex == this.nrOfTaskBatches) {
				currTaskBatchIndex = 0;
			}
		}
	}

	private void createOpenCVJobs(boolean requireModelTraining) {

		this.hasUnremovedJobs.set(true);

		this.inputFiles.clear();
		this.outputFiles.clear();
		this.modelFiles.clear();

		for (int i = 0; i < this.nrOfTaskBatches; i++) {
			this.inputFiles.put("" + i, new File(this.experimentFolder + "/IN_taskBatch_" + i + ".tbf"));
			this.outputFiles.put("" + i, new File(this.experimentFolder + "/OUT_taskBatch_" + i + ".tbf"));
			if (requireModelTraining) {
				/** Retrieval of the regression model is necessary */
				this.modelFiles.put("" + i,
						new File(this.experimentFolder + "/" + TMappingConstants.REMOTE_MODEL_FILE_NAME));
			}

			this.jobIds.put("" + i, this.jobDispatcher.createJob(this.experimentFolder, this.inputFiles.get("" + i),
					"" + this.nrOfCrossValidationFolds, i));

			// int[] jobCounts = this.jobDispatcher
			// .getJobStatusCounts(this.experimentFolder);
			// System.out.println("Created job #" + i + " : I" + jobCounts[0]
			// + " R" + jobCounts[1] + " C" + jobCounts[2] + " H"
			// + jobCounts[3]);
		}
	}

	private void deleteCurrentResultFiles() {
		try {
			File f = new File(this.experimentFolder + "/" + TMappingConstants.REMOTE_MODEL_FILE_NAME);
			if (f.exists()) {
				f.delete();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		try {
			File f = new File(this.experimentFolder + "/OUT_taskBatch_0.tbf");
			if (f.exists()) {
				f.delete();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void monitorCrossValidationJobCompletion() {
		List<Boolean> taskBatchCompleted = new ArrayList<Boolean>();
		for (int i = 0; i < this.nrOfTaskBatches; i++) {
			taskBatchCompleted.add(false);
		}

		// int time = 0;
		boolean allJobsFinished = false;
		long startTime = System.currentTimeMillis();
		long timePassed = 0;

		while ((!allJobsFinished) && (timePassed <= (this.maxDuration * 60L)) && (!this.stopMappingCreation.get())) {

			// time += 5;
			try {
				Thread.sleep(5000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			timePassed = (System.currentTimeMillis() - startTime) / 1000;

			// System.out.println("Time passed: " + timePassed + " of "
			// + (this.maxDuration * 60));

			// int[] jobCounts = this.jobDispatcher
			// .getJobStatusCounts(this.experimentFolder);
			// System.out.println("Job status after " + time + " seconds: I"
			// + jobCounts[0] + " R" + jobCounts[1] + " C" + jobCounts[2]
			// + " H" + jobCounts[3]);

			allJobsFinished = this.jobDispatcher.checkAllJobsFinished(this.experimentFolder);

			for (int i = 0; i < this.nrOfTaskBatches; i++) {
				if (taskBatchCompleted.get(i) == false) {
					if (this.jobDispatcher.isJobCompleted(this.jobIds.get("" + i))) {
						// System.out.println("Job #" + i + " is completed!");
						taskBatchCompleted.set(i, true);

						try {
							this.jobDispatcher.getJobTestSetResults(this.jobIds.get("" + i),
									this.outputFiles.get("" + i));
							if (this.outputFiles.get("" + i).exists()) {
								TMappingTaskBatch taskBatch = new TMappingTaskBatch(
										this.outputFiles.get("" + i).getAbsolutePath());
								this.completedTasks.addAll(taskBatch.getTaskList());
							} else {
								// TODO mark somehow that the retrival of the
								// task batch file was unsuccessful
							}
						} catch (Exception e) {
							this.parentMappingHolder.addLog("ERROR: could not fetch the results of job " + i + "!");
							e.printStackTrace();
						}

					}
				}
			}
		}

		this.jobDispatcher.removeAllRemainingJobs(this.experimentFolder);
		this.hasUnremovedJobs.set(false);
	}

	private void monitorModelTrainingJobCompletion(int trialCount) {

		this.parentMappingHolder.addLog("Training regression model ... (trial " + trialCount + " of "
				+ TMappingConstants.MAX_NR_OF_TRAINING_ATTEMPTS + ")");

		boolean jobIsFinished = false;
		boolean jobIsSuccessful = false;
		// int time = 0;
		long startTime = System.currentTimeMillis();
		long timePassed = 0;

		while (!jobIsFinished && (timePassed <= (this.maxDuration * 60L)) && (!this.stopMappingCreation.get())) {

			// time += 5;
			try {
				Thread.sleep(5000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			timePassed = (System.currentTimeMillis() - startTime) / 1000;

			jobIsFinished = this.jobDispatcher.isJobFinished(this.jobIds.get("0"));

			if (this.jobDispatcher.isJobCompleted(this.jobIds.get("0"))) {
				// System.out.println("Model creation job is completed!");

				try {
					this.jobDispatcher.getAllJobResults(this.jobIds.get("0"), this.outputFiles.get("0"),
							this.modelFiles.get("0"));
					if (this.outputFiles.get("0").exists() && (this.modelFiles.get("0").exists())) {

						File f = new File(this.experimentFolder + "/" + TMappingConstants.REMOTE_MODEL_FILE_NAME);

						// read file
						StringBuilder sb = new StringBuilder();
						int charsRead;
						char[] buff = new char[4096];
						InputStreamReader read = new InputStreamReader(new FileInputStream(f), "ISO-8859-1");
						while ((charsRead = read.read(buff)) != -1) {
							sb.append(buff, 0, charsRead);
						}
						read.close();
						String persistedPredictor = sb.toString();

						this.trainedPredictionModel = OpenCVUtils.loadPredictorFromString(persistedPredictor);
						this.trainedPredictionModel.performPostDeserializationActions();

						TMappingTaskBatch taskBatch = new TMappingTaskBatch(
								this.outputFiles.get("0").getAbsolutePath());
						this.completedTasks.addAll(taskBatch.getTaskList());
						this.trainingTask = taskBatch.getTaskList().get(0);
						jobIsSuccessful = true;
					} else {
						// TODO mark somehow that the retrieval of the
						// task batch file was unsuccessful
						jobIsSuccessful = false;
					}
				} catch (Exception e) {
					this.parentMappingHolder.addLog("ERROR: could not fetch the results of the model creation job!");
					e.printStackTrace();
					jobIsSuccessful = false;
				} // end try

			} // end if

		} // end while

		if ((!jobIsSuccessful) && (trialCount < TMappingConstants.MAX_NR_OF_TRAINING_ATTEMPTS)) {
			this.deleteCurrentResultFiles();
			this.createOpenCVJobs(true);
			this.monitorModelTrainingJobCompletion(trialCount + 1);
		}

		this.jobDispatcher.removeAllRemainingJobs(this.experimentFolder);
		this.hasUnremovedJobs.set(false);
	}

	private boolean prepareDataSets() {
		boolean outcome = true;

		try {
			/** Read the base dataset file. */
			this.readBaseDataset(this.baseDatasetFile);

			/** Create the cross-validation datasets */
			this.createCrossValidationDatasets();

			/** SAve the required datasets to the local experiment folder */
			this.saveDatasetsToExperimentFolder();

		} catch (Exception e) {
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}

	private boolean prepareOpenCVTaskBatches() {
		boolean outcome = true;

		try {

			/** Sort the tasks by their estimated execution times */
			Collections.sort(this.uncompletedTasks, new MappingTaskEstimatedExecutionTimeComparator());

			/** Group the required tasks into execution batches **/
			this.createTaskBatches();

			/** Save the execution batches to the local experiment folder */
			this.saveTaskBatchesToExperimentFolder();

		} catch (Exception e) {
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}

	private boolean readExperimentMasterFile() {
		String masterFileName = this.experimentFolder + "/" + TMappingConstants.EXPERIMENT_MASTER_FILE;

		boolean outcome = true;

		/**
		 * This general parameter list will contain all the parameter sublists that must
		 * be combined in order to obtain all the possible tests required by the current
		 * experiment. Each contained parameter sublist has the NAME of the parameter on
		 * the first position and the possible parameter values on the next positions.
		 */
		List<List<String>> generalParamList = new ArrayList<List<String>>();

		try {
			BufferedReader br = new BufferedReader(new FileReader(masterFileName));
			String line;

			/**
			 * Add the experiment type parameter sublist to the general parameter list.
			 */
			line = br.readLine();
			String[] lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			generalParamList.add(new ArrayList<String>(Arrays.asList(lineTokens)));
			if (TMappingConstants.TExperimentType.CV_EVALUATION_REG.getValue().equals((lineTokens[1]).toUpperCase())) {
				this.experimentType = TMappingConstants.TExperimentType.CV_EVALUATION_REG;
			} else if (TMappingConstants.TExperimentType.MODEL_TRAINING_REG.getValue()
					.equals((lineTokens[1]).toUpperCase())) {
				this.experimentType = TMappingConstants.TExperimentType.MODEL_TRAINING_REG;
			} else if (TMappingConstants.TExperimentType.CV_EVALUATION_CLS.getValue()
					.equals((lineTokens[1]).toUpperCase())) {
				this.experimentType = TMappingConstants.TExperimentType.CV_EVALUATION_CLS;
			} else {
				this.experimentType = TMappingConstants.TExperimentType.MODEL_TRAINING_CLS;
			}

			/**
			 * The number. of cross validation folds that will be used to evaluate/predict
			 * the quality of a regression model. <- this is just an execution control
			 * parameter and will NOT be added to the general parameter list
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			this.nrOfCrossValidationFolds = Integer.parseInt(lineTokens[1]);

			/**
			 * The total number of test evaluation batches that will be created. All the
			 * tests that will be performed for the current experiment will be grouped into
			 * these batches. Each batch is a single Condor\Local task. <- this is just an
			 * execution control parameter will NOT be added to the general parameter list
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			this.nrOfTaskBatches = Integer.parseInt(lineTokens[1]);

			/**
			 * The number of samples from the provided dataset will be used for model
			 * evaluation \training. <- this is just an execution control parameter will NOT
			 * be added to the general parameter list
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			this.trainingSetSize = Integer.parseInt(lineTokens[1]);

			/**
			 * The relative path to the .arff file that contains the dataset to be used for
			 * model evaluation \ training. <- this is just an execution control parameter
			 * will NOT be added to the general parameter list
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			this.baseDatasetFile = lineTokens[1];

			/**
			 * Read the names of the target variables. Create a sublist with these values
			 * and add it to the general parameter list.
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_ARRAY_SPLIT);
			for (int i = 0; i < lineTokens.length; i++) {
				if (i != 0) {
					lineTokens[i] = lineTokens[i].toUpperCase();
					this.targetNames.add(lineTokens[i]);
				}
			}
			generalParamList.add(new ArrayList<String>(Arrays.asList(lineTokens)));

			/**
			 * Add the regression model type parameter sublist to the general parameter
			 * list.
			 */
			line = br.readLine();
			lineTokens = line.split(TMappingConstants.KEY_VALUE_SPLIT);
			generalParamList.add(new ArrayList<String>(Arrays.asList(lineTokens)));

			/**
			 * Add the regression model configuration parameter sublists to the general
			 * parameter list.
			 */
			while ((line = br.readLine()) != null) {
				lineTokens = line.split(TMappingConstants.KEY_ARRAY_SPLIT);
				generalParamList.add(new ArrayList<String>(Arrays.asList(lineTokens)));
			}

			/**
			 * Use the general parameter list to create all the possible model evaluation \
			 * training tasks for this experiment.
			 */
			this.createAllPossibleTasksForExperiment(generalParamList);

		} catch (Exception e) {
			/** There was a problem reading the experiment master file */
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}

	@Override
	public void run() {
		this.readExperimentMasterFile();

		this.prepareDataSets();

		this.prepareOpenCVTaskBatches();

		this.createOpenCVJobs((this.experimentType.equals(TMappingConstants.TExperimentType.MODEL_TRAINING_REG))
				|| (this.experimentType.equals(TMappingConstants.TExperimentType.MODEL_TRAINING_CLS)));

		if ((this.experimentType.equals(TMappingConstants.TExperimentType.CV_EVALUATION_REG))
				|| (this.experimentType.equals(TMappingConstants.TExperimentType.CV_EVALUATION_CLS))) {
			this.monitorCrossValidationJobCompletion();
		} else {
			this.monitorModelTrainingJobCompletion(1);
		}

		if ((this.experimentType.equals(TMappingConstants.TExperimentType.CV_EVALUATION_REG))
				|| (this.experimentType.equals(TMappingConstants.TExperimentType.CV_EVALUATION_CLS))) {
			this.parentMappingHolder.handleEndOfCrossValidationExperiment();
		} else {
			this.parentMappingHolder.handleEndOfModelTrainingExperiment(this.trainedPredictionModel, this.trainingTask);
		}
	}

	private boolean readBaseDataset(String fileName) {
		boolean outcome = true;

		try {
			this.baseDataset = new OpenCVDataSource(fileName);

			String[] nameTokens = this.baseDataset.getDataSetName().split("_");
			this.nrOfDatasetOutputs = Integer.parseInt(nameTokens[1]);

			// System.out.println("BASE DATASET SIZE: "
			// + this.baseDataset.numInstances());
			// System.out.println("BASE DATASET OUTPUTS: "
			// + this.nrOfDatasetOutputs);
		} catch (Exception e) {
			this.parentMappingHolder.addLog("ERROR: could not read base dataset at: " + fileName);
			e.printStackTrace();
			outcome = false;
		}
		return outcome;
	}

	private boolean saveDatasetsToExperimentFolder() {
		boolean outcome = true;

		try {
			File dir = new File(this.experimentFolder);
			if (!dir.exists() && !dir.mkdirs()) {
				throw new IOException("ERROR: could not create directory path: " + this.experimentFolder);
			}

			/**
			 * Save the training and test dataset to the local experiment folder
			 */
			for (int i = 0; i < this.nrOfCrossValidationFolds; i++) {
				for (String target : this.targetNames) {

					OpenCVDataSource trainData = this.cvDatasets.get(TMappingConstants.TRAINING_PREFIX + target).get(i);
					trainData.saveToFile(
							this.experimentFolder + "/" + target + TMappingConstants.TRAINING_PREFIX + i + ".arff",
							true);

					OpenCVDataSource testData = this.cvDatasets.get(TMappingConstants.TEST_PREFIX + target).get(i);
					testData.saveToFile(
							this.experimentFolder + "/" + target + TMappingConstants.TEST_PREFIX + i + ".arff", true);
				}
			}
		} catch (Exception e) {
			this.parentMappingHolder.addLog("ERROR: could not create .arff files at: " + this.experimentFolder
					+ " ;requested nr. of cross-validation folds: " + this.nrOfCrossValidationFolds);
			e.printStackTrace();
			outcome = false;
		}

		return outcome;
	}

	private boolean saveTaskBatchesToExperimentFolder() {
		boolean outcome = true;

		try {
			// check if directory exists or try to create it
			File dir = new File(this.experimentFolder);
			if (!dir.exists() && !dir.mkdirs()) {
				throw new IOException("ERROR: could not create directory path: " + this.experimentFolder);
			}

			// save all test sets to file
			for (int i = 0; i < this.nrOfTaskBatches; i++) {
				this.taskBatches.get(i).saveSetToFile(this.experimentFolder + "\\" + "IN_taskBatch_" + i + ".tbf",
						true);
			}

		} catch (Exception e) {
			this.parentMappingHolder.addLog("ERROR: could not create task batch files at: " + this.experimentFolder);
			e.printStackTrace();

			outcome = false;
		}

		return outcome;
	}

}
