#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9188 $
$LastChangedDate:: 2019-06-11 11:10:29 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import datetime
import sphinx.ext.apidoc


class ApiDoc:

    def __init__(self, output_dir: str):
        self.output_dir = output_dir

    def run(self, packages_path):
        sphinx.ext.apidoc.main(['-o', self.output_dir, '-d', '1', packages_path])


class AutoDoc:

    def __init__(self, output_dir: str):
        self.output_dir = output_dir

    def run(self, packages_path, excludeModules=None):
        for package_name in os.listdir(packages_path):
            path = os.path.join(packages_path, package_name)
            if self.isPackage(path):
                try:
                    self.runPackage(package_name, path, excludeModules)
                except Exception as e:
                    print(e)

    def runPackage(self, packageName, packPath, excludeModules=None):
        """
        Function to generate a .rst file with autosummary directives
        which can be imported into the documentation.
        Used Filename is the same as the given package
        name.

        Parameters
        ----------
        packageName : string
            Name of the package
        packPath : string
            Full path to the package
        excludeModules:
            List of strings with full module names to exclude from build
        """
        import pyclbr  # might be problematic, hence imported inside the method

        if excludeModules is None:
            excludeModules = []
        '''Full path to directory to put the file into'''
        destPath = self.output_dir
        # get modules in package
        moduleNames = []
        modulePaths = []

        self.getModules(packPath, packageName, moduleNames, modulePaths)
        moduleNames, modulePaths = self.sortModules(moduleNames, modulePaths)
        importStatements = self.getImportStatements(moduleNames, modulePaths, packageName)

        if not os.path.isdir(destPath):
            os.makedirs(destPath)

        # import package and modules and get module members
        with open(os.path.join(destPath, packageName + '.rst'), 'w') as file:

            file.write(
                '.. File auto generated during build on {}\n   by {}\n\n\n'.format(datetime.datetime.now().ctime(),
                                                                                   os.getlogin()))
            header = 'Documentation of {} package'.format(packageName)
            file.write(self.underline(header))
            file.write('\n')
            for i, importStatement in enumerate(importStatements):
                if importStatement not in excludeModules:
                    module = pyclbr.readmodule_ex(importStatements[i])
                    file.write(self.underline(importStatements[i], '^'))
                    file.write('\n')
                    file.write('.. currentmodule:: {}\n'.format(importStatements[i]))
                    file.write('\n')
                    file.write('.. autosummary::\n')
                    file.write('   :nosignatures:\n')
                    file.write('   :toctree: _autosummary\n')
                    file.write('\n')
                    members = module.keys()
                    for _, memberName in enumerate(members):
                        file.write('   {}\n'.format(memberName))
                    file.write('\n')

    @staticmethod
    def getImportStatements(moduleNames, modulePaths, packageName):
        """
        Function for getting valid module import statements from the module names
        and their paths.

        Parameters
        ----------
        moduleNames : list of string
            The names of the modules
        modulePaths : list of strings
            Absolute paths of modules
        packageName : name of the package

        Returns
        -------
        importStrings : list of strings
            List of import statements

        """
        importStatements = []
        for i, (name, path) in enumerate(zip(moduleNames, modulePaths)):
            if name != '__init__':
                importStatements.append(path[path.find(packageName):].replace('\\', '.') + '.' + name)

        return importStatements

    def getModules(self, dirPath, packageName, moduleNames, modulePaths):
        """
        Recursively get modules from a package.
        Function works in place.

        Parameters
        ----------
        dirPath : string
            Full path to first package level
        packageName : string
            Name of the package
        moduleNames : list
            List to append found packages
        modulePaths: list

        """
        if self.isPackage(dirPath):
            files = os.listdir(dirPath)
            for file in files:
                extension = os.path.splitext(file)[1]
                if extension == '.py':
                    moduleNames.append(os.path.splitext(file)[0])
                    modulePaths.append(dirPath)
                else:
                    if self.isPackage(os.path.join(dirPath, file)):
                        self.getModules(os.path.join(dirPath, file), packageName, moduleNames, modulePaths)
        else:
            raise ValueError('No package')

    @staticmethod
    def sortModules(moduleNames, modulePaths):
        """
        Function for sorting moduleNames and modulePaths according to their path length.
        Module names and paths are then arranged in an hierarchical order.

        Parameters
        ----------
        moduleNames : list of strings
           List of module names
        modulePaths : list of path strings
           List of module path strings

        """
        zipped = zip(modulePaths, moduleNames)
        zipped_sorted = sorted(zipped, key=lambda t: (len(t[0]), t[1]))
        modulePathsSorted = [x[0] for x in zipped_sorted]
        moduleNamesSorted = [x[1] for x in zipped_sorted]
        return moduleNamesSorted, modulePathsSorted

    @staticmethod
    def isPackage(path):
        """
        Checks if given path is a package.

        Parameters
        ----------
        path : string

        Returns
        -------
        ret : bool
        """
        if not os.path.isdir(path):
            return False

        files = os.listdir(path)
        if '__init__.py' in files:
            return True
        else:
            return False

    @staticmethod
    def underline(header, underlineChar='-'):
        """
        Function for creating a underline for a sphinx header section.

        Parameters
        ----------
        header : string
        underlineChar : string

        Returns
        -------
        underlinedHeader : string
        """
        return header + '\n' + len(header) * underlineChar + '\n'
