# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10211 $
$LastChangedDate:: 2021-05-20 14:18:02 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

from abc import abstractmethod
from builtins import print

import os
import sys
import time
import shutil
import pickle
import tempfile
import traceback
from symspace.jobinterface.portDescWrapper import SymJobDesc

try:
    import jep
    import numpy as np
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
    # noinspection PyUnresolvedReferences
    from at.lcm.bu21.magopt.structure import PortDesc
    # noinspection PyUnresolvedReferences
    from java.lang import StackTraceElement
    # noinspection PyUnresolvedReferences
    from java.io import File as jFile
    # noinspection PyUnresolvedReferences
    from jep import jarray
    # noinspection PyUnresolvedReferences
    from jobmanager import JobDesc
    # noinspection PyUnresolvedReferences
    from jobmanager import JobType
    from symspace.wrapper.symwrapper import SymPythonFunction
except Exception as e:
    print(e)


class ParamPythonFnc(object):
    """
    Base class for |sym|'s Python functions which are chained
    into the tree structure.
    """

    def __init__(self, structure):
        """
        Initialize.
        """
        self._structure = structure
        self.debug_mode = False

        # Todo: in future, do not copy the complete _Scripts directory but only the necessary files it contains (?)
        self.job_add_scripts_dir = True  # add all python modules and packages of the _Scripts directory to the job
        # Todo: test automatic collection of job files
        self.job_add_component_files = False  # collect and add component dependencies automatically to the job
        self.max_file_size_mb = 100.0  # maximum file size in MB

    @classmethod
    def getFunctionID(cls):
        return __name__ + '_' + cls.__name__

    @classmethod
    def getDefaultDescription(cls):
        return cls.__doc__.splitlines()[0].strip()

    @abstractmethod
    def importVars(self):
        """Description of the import variables

        Returns a list of tuples, which defines the import variables.
        Each tuple consists of multiple entries in the following order:

        1. Variable Name : string

        2. Variable Description : string

        3. Variable Unit : string

        4. Optional : bool

            The boolean value 'Optional' defines if the variable has to be set (False) or can be left unset (True).

        5. Type / Dimension : string

            The string 'Type / Dimension' defines the type and the dimension / size of the variable:

                * 'Null / 1'

                * 'Bool / 1'

                * 'String / 1'

                * 'Numeric / [n_dim1 x n_dim2]'

                * 'Periodic / [n_dim1 x n_dim2]'

                * 'Array / [n_dim1 x n_dim2]'

                * 'Struct / n_fields'

            If n_dim1, n_dim2 or n_fields is unknown in advance use ? instead. E.g. '[1 x ?]'

        6. Struct : tuple

            The tuple 'Struct' defines the inner structure of the variable.

                * No entry or empty list '[]' for all types except of 'Struct'

                *  If 'Type' is 'Struct' all variables, which are known in advance to be contained need to be described.
                The description of sub-level variables is done in the same way as the top layer description.


        return [('ImportVar1', 'Description of import variable 1', 'Unit1', Optional1, 'Type1 / Dimension1', Struct 1),
                ('ImportVar2', 'Description of import variable 2', 'Unit2', Optional2, 'Type2 / Dimension2', Struct 2)]

        Returns
        -------
        list
            List of import variables

        """
        pass

    @abstractmethod
    def exportVars(self):
        """Description of the export variables

           Same as :func:`importVars`
        """
        pass

    def getJobRequirement(self) -> str:
        """
        Get the Job requirements

        Examples
        --------

        '(PYTHON >= \"3.5.3\")' requires Python >= 3.5.3 at remote instance
        '(PYTHON >= \"3.5.3\") && (JEP >= \"3.6.3\")') requires Python >= 3.5.3 and JEP >= 3.6.3 at remote instance

        Returns
        -------
        requirements: str
        """
        return ""

    def getVariation(self) -> list:
        """
        Get the list of variation import parameter keys

        Returns
        -------
        variation : list
                e.g. ['a', 'b']
                The selected import parameters need to be of type np.ndarray!
        """
        return None

    def getAdditionalFiles(self) -> list:
        """Define additional file and folder dependencies for a job

        These files and folders will be copied into the jobs working directory

        Examples
        -------
        >>> symProjectPath = self.getProjectPath()
        >>> my_file = os.path.join(symProjectPath, 'resources', 'my_file.txt')
        >>> lst = [my_file, ]
        >>> lst.append(os.path.join(symProjectPath, 'my_dir'))
        >>> # import my_module
        >>> # lst.append(my_module.__file__) # adding a python module, by dynamic path
        >>> # import my_pkg
        >>> # lst.append(os.path.dirname(my_pkg.__file__))  # adding a python package, by dynamic path

        Returns
        -------
        paths: list
        """
        return []

    def getAdditionalModules(self) -> list:
        """Define additional module dependencies for a job

        Same as :func:`getAdditionalFiles`

        Todo: Currently there is no difference to getAdditionalFiles!

        Examples
        -------
        >>> symProjectPath = self.getProjectPath()
        >>> my_module = os.path.join(symProjectPath, 'src', 'my_module.py')
        >>> lst_modules = [my_module, ]

        Returns
        -------
        module_paths: list
        """
        return []

    def getAdditionalPackages(self) -> list:
        """Define additional package dependencies for a job

        Same as :func:`getAdditionalFiles`

        Todo: Currently there is no difference to getAdditionalFiles!

        Examples
        -------
        >>> symProjectPath = self.getProjectPath()
        >>> my_package = os.path.join(symProjectPath, 'src', 'my_package')
        >>> lst_pkgs = [my_package, ]

        Returns
        -------
        package_paths: list
        """
        return []

    def getResultFileTransfer(self, d_job: dict) -> list:
        """Define additional file and folder result keys of a job

        These files and folders will be transferred from the job directory into the parameter directory

        Todo: define a standardized way for the definition of file transfer e.g.
              This method needs to be available also from the
              Dummy ParamPythonFunction in symspace.jobinterface.starter.dummy_parampythonfnc
        def addResultFileTransfer(self, d, paths: list):
          for index, path in enumerate(paths):
              d['__resultFileTransfer_{}'.format(index)] = path

        Returns
        -------
        paths: list
        """
        return [key for key in d_job.keys() if key.startswith('__resultFileTransfer_')]

    def isDistributable(self) -> bool:
        """This method is used to define if the function is distributable"""
        return False

    def isComponent(self) -> bool:
        """
        Return if the Function is a Component

        Returns
        -------
        bool
        """
        param = SymPythonFunction(javaObj=self._structure)
        return param.isComponent()

    def getFile(self) -> str:
        """Get the full file path of the function"""
        return ""

    def getUniverse(self) -> str:
        return 'JAVA'

    def getDockerImage(self) -> str:
        raise NotImplementedError

    def getProjectPath(self) -> str:
        """Get path to the |sym| project folder"""
        proj = self._structure.getRoot()
        moProjectPath = os.path.dirname(os.path.normpath(proj.getFile()))
        return moProjectPath

    def runJob(self, d: dict, run_args: dict) -> int:
        """
        This method can be overwritten by a subclass.

        If this method is implemented in a subclass it, may be called instead of the subclass' run method
        e.g. when using a remote execution wrapper class

        Parameters
        ----------
        d : dict
        run_args : dict
            See run method

        Returns
        -------
        status: int
            If status == 0 the function regenerated successfully
            Else the function update is not completed OR failed
        """
        return 1

    def run(self, run_args: dict) -> int:
        """Run method of the ParamPythonFunction

        This method will be overwritten by a subclass.
        This superclass implementation is used to handle remote calculation

        Parameters
        ----------
        run_args : dict
                Arguments of caller (Control flags)
                {RemoteCompute=false, RunSim=false, Optimize=false}

        Returns
        -------
        status : int
            If status == 0 the function regenerated successfully
            Else the function update is not completed OR failed

        """
        infileName = 'ParamInput.p'
        outfileName = 'ParamOutput.p'

        d = self.importValues()

        remoteCompute = run_args['RemoteCompute']
        runSim = run_args['RunSim']
        # optimize = run_args['Optimize']

        timeout = sys.maxsize
        if run_args.get('Timeout') is not None:
            timeout = run_args['Timeout']

        if self.isDistributable() and not runSim:
            self.warning('Stopped before sending Jobs!')
            return 1

        if self._structure.getTmpDSData("JobID") is None:
            self.preJob(d, run_args)

            if self.isDistributable():
                '''function is distributable submit via jobmanager'''

                proj = self._structure.getRoot()

                '''Add run_args to the input dictionary
                On remote machine python starter will pass this value to run Job'''
                d['__run_args'] = dict(PyJDataInterface.convert_J2Py(run_args))

                rows = 1
                cols = 1
                if self.getVariation() is not None:
                    shape = None
                    for var in self.getVariation():
                        tmp = d[var]
                        if isinstance(tmp, np.ndarray):
                            if shape is None:
                                shape = tmp.shape
                                rows = shape[0]
                                cols = shape[1]
                            else:
                                if shape != tmp.shape:
                                    raise Exception('Shape for variation parameters are not equal ({})'.format(var))
                        else:
                            raise Exception('Type of variation parameter is not an array ({})'.format(var))

                '''universal tasks for all variations'''

                '''create job'''
                if 'jep' in self.getJobRequirement().lower():
                    job = 'jobmanager.python.CLILocalJepJob'
                else:
                    job = 'jobmanager.python.CLILocalJob'

                add_files, link_files = self._getJobFiles()

                jobIDs = []
                for row in range(0, rows):
                    jobIDs.append([])
                    for col in range(0, cols):
                        '''create job dictionary'''
                        d_job = {}
                        for key in d.keys():
                            d_job[key] = d[key]
                        if self.getVariation() is not None:
                            for var in self.getVariation():
                                d_job[var] = d_job[var][row, col]

                        max_file_size_mb_str = '{:g}'.format(self.max_file_size_mb)

                        if self.getUniverse().upper() == 'DOCKER':
                            jobID = SymJobDesc.createDockerJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                                                               self._structure.getClassName() + "_{}_{}".format(row,
                                                                                                                col),
                                                               self.getDockerImage(), 'python3',
                                                               ['symstarter.py',
                                                                self._structure.getModuleName(),
                                                                self._structure.getClassName(),
                                                                infileName, outfileName, max_file_size_mb_str])
                        # jobID = SymJobDesc.createDockerJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                        #                                    self._structure.getClassName() + "_{}_{}".format(row,
                        #                                                                                     col),
                        #                                    self.getDockerImage(), 'java',
                        #                                    ['jobmanager.server.JobStarter', job, 'symstarter.py',
                        #                                     self._structure.getModuleName(),
                        #                                     self._structure.getClassName(),
                        #                                     infileName, outfileName])
                        else:
                            # elif self.getUniverse().upper() == 'JAVA':
                            jobID = SymJobDesc.createJavaJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                                                             self._structure.getClassName() + "_{}_{}".format(row, col),
                                                             ['jobmanager.server.JobStarter', job, 'symstarter.py',
                                                              self._structure.getModuleName(),
                                                              self._structure.getClassName(), infileName, outfileName,
                                                              max_file_size_mb_str])
                        # else:
                        #     jobID = SymJobDesc.createDefaultJob(proj.getFile(), proj.getDsId(), self._structure.getPId(),
                        #                                         self._structure.getClassName() + "_{}_{}".format(row, col),
                        #                                         'python',
                        #                                         ['symstarter.py',
                        #                                          self._structure.getModuleName(),
                        #                                          self._structure.getClassName(),
                        #                                          infileName, outfileName])

                        jobIDs[row].append(jobID)

                        jobPath = SymJobDesc.getPath(jobID)

                        '''add import files'''
                        imports = self.getImports()
                        for imp in imports:
                            name = imp[0]
                            if name in d.keys():
                                path = d_job[name]
                                if isinstance(path, str) and os.path.isfile(path):
                                    if path not in add_files:
                                        add_files.append(path)
                                    '''change the file path to the job working directory'''
                                    d_job[name] = os.path.basename(path)
                                elif isinstance(path, str) and os.path.isdir(path):
                                    zip_file = self._zipDirectory(path)
                                    if zip_file not in add_files:
                                        add_files.append(zip_file)
                                    '''change the directory path to the job working directory'''
                                    d_job[name] = os.path.basename(path)

                        '''create parameter file'''
                        infilePath = os.path.normpath(os.path.join(jobPath, infileName))
                        link_files.append(infilePath)
                        with open(infilePath, 'wb') as fh:
                            pickle.dump(d_job, fh)

                        for file in add_files:
                            if not os.path.isfile(file):
                                raise FileNotFoundError('File does not exist: "{}"'.format(file))
                            SymJobDesc.addFile(jobID, file)

                        for file in link_files:
                            if not os.path.isfile(file):
                                raise FileNotFoundError('File does not exist: "{}"'.format(file))
                            SymJobDesc.linkFile(jobID, file)

                        '''Set timeouts and requirements'''
                        SymJobDesc.setTimeout(jobID, timeout)
                        SymJobDesc.setRequirement(jobID, self.getJobRequirement())

                        if remoteCompute:
                            SymJobDesc.submit(jobID, destination='cluster')
                        else:
                            SymJobDesc.submit(jobID, destination='local')

                self._structure.info('Jobs sent to JobManager ({})'.format(rows * cols))
                self._structure.setTmpDSData("JobID", jobIDs)
                status = 1
            else:
                '''function is not distributable -> run without job manager'''
                status = self.runJob(d, run_args)

        else:
            '''process single job'''
            jobIDs = self._structure.getTmpDSData("JobID")
            rows = len(jobIDs)
            cols = len(jobIDs[0])

            '''check all finished'''
            status = 0
            error = 0
            for row in range(0, rows):
                for col in range(0, cols):
                    job_error = 0
                    jobID = jobIDs[row][col]
                    if SymJobDesc.isFinished(jobID):
                        fileList = SymJobDesc.getFileList(jobID)
                        fileListBasenames = [os.path.basename(file) for file in fileList]

                        if SymJobDesc.getStatus(jobID) == 'COMPLETED':

                            if 'symstarter_error.log' in fileListBasenames and status == 0:
                                '''Fail, if sym starter error log was created'''
                                index = fileListBasenames.index('symstarter_error.log')
                                stack = []
                                with open(fileList[index], 'r') as f:
                                    _type = None
                                    msg = None
                                    for line in f:
                                        line = line.strip()
                                        if _type is None:
                                            _type = line
                                        elif msg is None:
                                            msg = line
                                        else:
                                            exc = line.split(',')
                                            stack.insert(0, StackTraceElement('', exc[0], exc[1], int(exc[2])))
                                self._structure.errorWithStack('Job: {}'.format(msg), stack)
                                job_error += 1

                            elif 'starter_error.log' in fileListBasenames and status == 0:
                                '''Fail, if java starter error log was created'''
                                # index = fileListBasenames.index('starter_error.log')
                                # with open(fileList[index], 'r') as f:
                                #     msg = f.read()
                                self.error('Error file starter_error.log created by java starter')
                                job_error += 1

                            if 'symstarter_debug.log' in fileListBasenames and status == 0:
                                '''Read runJob status'''
                                run_job_status = None
                                index = fileListBasenames.index('symstarter_debug.log')
                                with open(fileList[index], 'r') as f:
                                    for line in f:
                                        if 'status runJob:' in line:
                                            run_job_status = line.split('status runJob:')[-1].strip()
                                            break
                                if run_job_status != '0':
                                    '''Fail, if job status is not 0'''
                                    self.error('runJob returned status: {}'.format(run_job_status))
                                    job_error += 1

                            if 'starter_debug.log' in fileListBasenames and status == 0:
                                '''Read cli status'''
                                cliStatus = None
                                index = fileListBasenames.index('starter_debug.log')
                                with open(fileList[index], 'r') as f:
                                    for line in f:
                                        line = line.strip()
                                        if line.startswith('status'):
                                            cliStatus = line.split(':')[1].strip()
                                            break
                                if cliStatus != '0':
                                    '''Fail, if cli status is not 0'''
                                    self.error('Local commandline interface returned with status: {}'.format(cliStatus))
                                    job_error += 1

                            # elif 'starter_debug.log' not in fileListBasenames and status == 0:
                            #     '''Fail, if file is missing'''
                            #     self.error('Missing Java starter debug logfile')
                            #     error += 1

                            if outfileName not in fileListBasenames and status == 0:
                                '''Fail, if file is missing'''
                                self.error('Missing outfile of python subprocess')
                                job_error += 1

                        elif SymJobDesc.getStatus(jobID) == 'ERROR':
                            if 'symstarter_error.log' in fileListBasenames:
                                index = fileListBasenames.index('symstarter_error.log')
                                stack = []
                                with open(fileList[index], 'r') as f:
                                    _type = None
                                    msg = None
                                    for line in f:
                                        line = line.strip()
                                        if _type is None:
                                            _type = line
                                        elif msg is None:
                                            msg = line
                                        else:
                                            exc = line.split(',')
                                            stack.insert(0, StackTraceElement('', exc[0], exc[1], int(exc[2])))
                                self._structure.errorWithStack('Job: {}'.format(msg), stack)
                            job_error += 1
                        elif SymJobDesc.getStatus(jobID) == 'HELD':
                            job_error += 1
                        elif SymJobDesc.getStatus(jobID) == 'TIMEOUT':
                            job_error += 1
                    else:
                        status += 1

                    if job_error > 0:
                        error += 1

            if status == 0:
                if error == 0:
                    '''jobs finished successfully'''
                    self._transferResultFiles(d, jobIDs, outfileName)

                    '''reset statemachine'''
                    self._structure.rmTmpDSData("JobID")
                else:
                    raise Exception('{} of {} finished with error'.format(error, rows * cols))
            else:
                self.info("Simulation in progress ({})".format(status))

        if status == 0:
            '''run finished successfully'''
            self.postJob(d, run_args)
            self.exportValues(d)

        return status

    def preJob(self, d: dict, run_args: dict):
        """
        Run this method before running the main method of the subclass (run, runJob)

        e.g. to set temporary values for the run methods

        Parameters
        ----------
        run_args : dict
        d : dict
            Contains importVars
            Additional keys, value pairs may be added,
            this additional values can be in the (run, runJob) methods
        """
        pass

    def postJob(self, d: dict, run_args: dict):
        """
        Run this method after running the main method of the subclass (run, runJob)

        e.g. to change the structure, value or type of export variables before exporting them

        Parameters
        ----------
        run_args : dict
        d : dict
            Contains importVars and exportVars
        """
        pass

    def convertVariantsType(self, d: dict):
        """
        Convert variants result

        Try to convert object arrays (TArray)
        into numeric arrays (TNumeric) or periodic matrix (TPeriodicMatrix)

        Parameters
        ----------
        d :dict
        """
        if self.getVariation() is not None:
            for var in self.exportVars():
                var_name = var[0]
                var_type = var[4].lower()
                try:
                    if 'numeric' in var_type:
                        '''variable type is defined as numeric in the port description'''
                        d[var_name] = d[var_name].astype(np.number)
                    elif 'periodic' in var_type:
                        '''variable type is defined as periodic in the port description'''
                        pass  # Todo: implement conversion from Object NDArray to PeriodicMatrix
                except Exception as ex:
                    print(ex)

    def packageJob(self):
        # Todo: add documentation
        pass

    def submitJob(self):
        # Todo: add documentation
        pass

    def refreshDependency(self):
        self.setImportVars(self.importVars())
        self.setExportVars(self.exportVars())

    def generateDocFiles(self, path: str):
        # Todo: add documentation
        pass

    def reportTex(self, template: str) -> str:
        # Todo: add documentation
        return template

    def reset(self):
        """remove job from jobmanager"""
        jobIDs = self._structure.getTmpDSData("JobID")
        if jobIDs is not None:
            for jobIDs_row in jobIDs:
                for jobID in jobIDs_row:
                    SymJobDesc.remove(str(jobID))

        self._structure.rmTmpDSData("JobID")

    def getImports(self):
        # Todo: add documentation
        imp = []
        depend = self._structure.getDepend().getImports()
        for dep in depend:
            imp.append((dep.getVarName(), dep.getParam(), dep.getHandle()))
        return imp

    def getExports(self):
        # Todo: add documentation
        exp = []
        depend = self._structure.getDepend().getExports()
        for dep in depend:
            exp.append((dep.getVarName(), dep.getParam(), dep.getHandle()))
        return exp

    def getUpdateRequired(self):
        # Todo: add documentation
        return self._structure.getUpdateRequired()

    def __getPortDesc(self, var):
        # Todo: add documentation
        item = None
        length = len(var)
        if length == 1:
            item = PortDesc(var[0], var[0], "", False)
        elif length == 2:
            item = PortDesc(var[0], var[1], "", False)
        elif length == 3:
            item = PortDesc(var[0], var[1], var[2], False)
        elif length == 4:
            item = PortDesc(var[0], var[1], var[2], var[3])
        elif length == 5:
            item = PortDesc(var[0], var[1], var[2], var[3], var[4])
        elif length == 6:
            # print(type(var[5]))
            if type(var[5]) is list:
                subItem = []
                for sub in var[5]:
                    subItem.append(self.__getPortDesc(sub))
                item = PortDesc(var[0], var[1], var[2], var[3], var[4])
                item.setStructureList(subItem)
            else:
                subItem = self.__getPortDesc(var[5])
                item = PortDesc(var[0], var[1], var[2], var[3], var[4], subItem)
        else:
            self.warning("Undefined port-description")

        return item

    def setImportVars(self, varList):
        # Todo: add documentation
        items = []
        for var in varList:
            item = self.__getPortDesc(var)
            if item is not None:
                items.append(item)

        jarr = jarray(len(items), PortDesc)
        for i in range(len(items)):
            jarr[i] = items[i]

        self._structure.getDepend().setImportVars(jarr)

    def setExportVars(self, varList):
        # Todo: add documentation
        items = []
        for var in varList:
            item = self.__getPortDesc(var)
            if item is not None:
                items.append(item)

        jarr = jarray(len(items), PortDesc)
        for i in range(len(items)):
            jarr[i] = items[i]

        self._structure.getDepend().setExportVars(jarr)

    def setImportAssignments(self, **kwargs):
        # Todo: add documentation
        self._structure.getDepend().setImportAssignmentsMap(kwargs)

    def setExportAssignments(self, **kwargs):
        # Todo: add documentation
        self._structure.getDepend().setExportAssignmentsMap(kwargs)

    def setUpdateRequired(self, val):
        # Todo: add documentation
        self._structure.setUpdateRequired(val)

    def info(self, message):
        # Todo: add documentation
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.infoWithStack(message, stack)

    def warning(self, message):
        # Todo: add documentation
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.warningWithStack(message, stack)

    def error(self, message):
        # Todo: add documentation
        stack = []
        for trace in traceback.extract_stack():
            stack.insert(0, StackTraceElement('', trace.name, trace.filename, trace.lineno))
        self._structure.errorWithStack(message, stack)

    def importValues(self) -> dict:
        # Todo: add documentation
        d = {}
        imp = self._structure.getDepend().getImports()

        for im in imp:
            name = im.getVarName()
            data = self._structure.getImportValue(name)

            if data is not None:
                d[name] = PyJDataInterface.convert_J2Py(data)

        return d

    def importRawValues(self) -> dict:
        # Todo: add documentation
        d = {}
        imp = self._structure.getDepend().getImports()

        for im in imp:
            name = im.getVarName()
            data = self._structure.getImportValue(name)

            if data is not None:
                d[name] = data

        return d

    def exportValues(self, d: dict):
        # Todo: add documentation
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            if name in d:
                self._structure.setExportValue(name, PyJDataInterface.convert_Py2J(d[name]))
            else:
                raise Exception("Export value for '" + name + "' doesn't exist")

    def revertExportValues(self, d: dict):
        # Todo: add documentation
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            data = self._structure.getExportValue(name)

            if data is not None:
                d[name] = PyJDataInterface.convert_J2Py(data)

    def revertExportRawValues(self, d: dict):
        # Todo: add documentation
        exp = self._structure.getDepend().getExports()

        for ex in exp:
            name = ex.getVarName()
            data = self._structure.getExportValue(name)

            if data is not None:
                d[name] = data

    def getImportUnit(self, name):
        # Todo: add documentation
        return self._structure.getImportUnit(name)

    def getExportUnit(self, name):
        # Todo: add documentation
        return self._structure.getExportUnit(name)

    def registerDocFile(self, name, path, desc):
        # Todo: add documentation
        return self._structure.registerDocFile(name, path, desc)

    def getUniqueDocFileName(self, directory):
        # Todo: add documentation
        return self._structure.getUniqueDocFileName(directory)

    def getDocFiles(self):
        # Todo: add documentation
        docs = []
        docFiles = self._structure.getDocFiles()
        for doc in docFiles:
            docs.append((doc.getName(), doc.getDescription(), doc.getFile()))
        return docs

    def getDocFile(self, name):
        # Todo: add documentation
        doc = self._structure.getDocFile(name)
        if doc is not None:
            return doc.getName(), doc.getDescription(), doc.getFile()
        else:
            return None

    def getTmpData(self, key):
        # Todo: add documentation
        return self._structure.getTmpData(key)

    def setTmpData(self, key, data):
        # Todo: add documentation
        self._structure.setTmpData(key, data)

    def rmTmpData(self, key):
        # Todo: add documentation
        self._structure.rmTmpData(key)

    def getTmpDSData(self, key):
        # Todo: add documentation
        return self._structure.getTmpDSData(key)

    def setTmpDSData(self, key, data):
        # Todo: add documentation
        self._structure.setTmpDSData(key, data)

    def rmTmpDSData(self, key):
        # Todo: add documentation
        self._structure.rmTmpDSData(key)

    def getParamDirectory(self):
        # Todo: add documentation
        return self._structure.getParamDirectory().getPath()

    def hasPreview(self) -> bool:
        """Return true if the preview method is implemented"""
        return hasattr(self, 'preview') and callable(getattr(self, 'preview'))

    def hasShow(self) -> bool:
        """Return true if the preview method is implemented"""
        return hasattr(self, 'show') and callable(getattr(self, 'show'))

    def getModule(self) -> str:
        # Todo: add documentation
        import inspect
        return inspect.getfile(self.__class__)

    def _getJobFiles(self) -> tuple:
        """
        Get the files with are added and linked to a job

        Returns
        -------
        tuple:
            unique_add_files : list
                Files to add to the job
            unique_link_files: list
                Files to link to the job
        """
        from symspace.osutils.pathutils import Paths

        symspace_home = Paths.getSyMSpaceRoot()

        fnc_file = self.getFile()

        scripts_dir = os.path.join(self._structure.getRoot().getAllFilesDirectory().getAbsolutePath(), '_Script')

        unsorted_files = self.getAdditionalFiles()
        unsorted_modules = self.getAdditionalModules()
        unsorted_pkgs = self.getAdditionalPackages()

        add_files = [fnc_file,
                     os.path.join(symspace_home, 'Tools', 'JobManager', 'Java', 'jobmanager.jar'),
                     os.path.join(symspace_home, 'System', 'Python', 'src', 'symspace',
                                  'jobinterface', 'starter', 'symstarter.py')
                     ]

        link_files = [os.path.join(symspace_home, 'System', 'Java', 'magopt.jar'),
                      os.path.join(symspace_home, 'System', 'General', 'Java', 'general.jar'),
                      os.path.join(symspace_home, 'System', 'General', 'External', 'commons',
                                   'commons-math3-3.0.jar'),
                      os.path.join(symspace_home, 'External', 'javasysmon', 'javasysmon-0.3.3.jar'),
                      ]

        '''archive python packages'''
        pkgs_paths = [os.path.join(symspace_home, 'System', 'Python', 'src', 'symspace'),
                      os.path.join(symspace_home, 'System', 'General', 'Python', 'src', 'lcm')
                      ]

        unsorted_paths = []
        unsorted_paths.extend(unsorted_files)
        unsorted_paths.extend(unsorted_modules)
        unsorted_paths.extend(unsorted_pkgs)
        unsorted_paths.extend(pkgs_paths)

        if self.job_add_scripts_dir and os.path.isdir(scripts_dir):
            '''collect paths in the _Scripts directory'''
            unsorted_paths.extend([os.path.join(scripts_dir, path) for path in os.listdir(scripts_dir)])

        if self.job_add_component_files:
            '''collect modules, packages and other files of the component'''
            param = SymPythonFunction(javaObj=self._structure)
            if param.isComponent():
                '''check if the function is a component'''
                component = param.getComponent()
                component_files = component.listLinkedFiles()
                unsorted_paths.extend(component_files)

        unique_unsorted_paths = set(unsorted_paths)

        for path in unique_unsorted_paths:
            if os.path.isfile(path) and path not in add_files:
                '''path is a file, which can be added'''
                add_files.append(path)
            elif os.path.isdir(path):
                '''path is a directory, which may be zipped and added'''
                zip_file = self._zipDirectory(path)
                if zip_file not in add_files:
                    add_files.append(zip_file)

        unique_add_files = list(set(add_files))
        unique_link_files = list(set(link_files))

        return unique_add_files, unique_link_files

    def _zipDirectory(self, path: str) -> str:
        """
        Create a zip file from an existing folder.

        The zip file is created in a temporary directory
        and labeled with the timestamp of the last time the folder (and its content) was last modified.
        This way it does only need to be re-created if the content of folder changed.

        If the zip file already exists, just return the path.

        Parameters
        ----------
        path : str
            Source directory path

        Returns
        -------
        zip_file :str
            Zip file path
        """
        if os.path.getsize(path) < self.max_file_size_mb * 1024 ** 2:
            '''check if the directory does not exceed the maximum defined sized'''

            '''create timestamp from the last modified tag (including subdirectories)'''
            last_modified = time.localtime(max(os.path.getmtime(root) for root, _, _ in os.walk(path)))
            time_stamp = time.strftime('%Y%m%d_%H%M%S', last_modified)

            folder_name = os.path.basename(path)
            target_path = os.path.join(tempfile.gettempdir(), '{}_{}'.format(folder_name, time_stamp))
            zip_file = '{}.zip'.format(target_path)

            if not os.path.exists(zip_file):
                '''create zip file from folder'''
                parent_folder = os.path.dirname(path)
                shutil.make_archive(target_path, format='zip', root_dir=parent_folder, base_dir=folder_name)

                if not os.path.isfile(zip_file):
                    raise FileNotFoundError('File does not exist: {}'.format(zip_file))
                if not os.path.getsize(zip_file) > 0:
                    raise FileExistsError('File is empty: {}'.format(zip_file))

                print('File created:', zip_file)

        else:
            '''folder size is to big'''
            raise Exception('Folder size is bigger than {} MB'.format(self.max_file_size_mb))

        return zip_file

    def _transferResultFiles(self, d: dict, jobIDs: list, outfileName: str):
        """
        Transfer the result files from the job directory to the ParamDirectory
        Parameters
        ----------
        d: dict
        jobIDs: list
        outfileName : str
        """
        rows = len(jobIDs)
        cols = len(jobIDs[0])

        folder = self.getParamDirectory()
        ''' clear folder'''
        for f in os.listdir(folder):
            path = os.path.join(folder, f)
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)

        '''init results'''
        exports = self.getExports()
        for exp in exports:
            name = exp[0]
            d[name] = np.ndarray((rows, cols), object)

        '''read results'''
        for row in range(0, rows):
            for col in range(0, cols):
                jobID = jobIDs[row][col]
                job_dir = SymJobDesc.getPath(jobID)
                fileList = SymJobDesc.getFileList(jobID)  # java code returns files and folders!
                fileList = [file for file in fileList if os.path.isfile(file)]  # only files
                fileListBasenames = [os.path.basename(file) for file in fileList]

                '''Succeed and read results'''
                index = fileListBasenames.index(outfileName)
                with open(fileList[index], 'rb') as f:
                    d_job = pickle.load(f)

                ''' transfer exports and copy export-files'''
                # Todo: this only considers files and folders in the job's top-level directory,
                #  but no files/folders in the subfolders etc.

                '''get manually defined files/folders to transfer'''
                job_export_names = self.getResultFileTransfer(d_job)

                '''get files/folders defined by export to transfer'''
                job_export_names.extend([exp[0] for exp in exports])

                for name in job_export_names:
                    val = d_job[name]

                    if isinstance(val, str):
                        '''only the basename is considered 
                           - hence it does not matter if val is a relative or absolute path 
                           - but files/folders must be in the toplevel directory
                        '''
                        base_name = os.path.basename(val)
                        '''zip-file is named according to the dictionary key'''
                        zip_file_bn = '{}.zip'.format(name)
                        if base_name in fileListBasenames or zip_file_bn in fileListBasenames:
                            if base_name in fileListBasenames:
                                '''the file exist in the job directory'''
                                old_path = fileList[fileListBasenames.index(base_name)]
                            else:
                                '''the zip-file exist in the job directory'''
                                old_path = os.path.join(job_dir, base_name)
                                if not os.path.isdir(old_path):
                                    '''unpack if the folder does not already exist in the toplevel dir'''
                                    zip_file = os.path.join(job_dir, zip_file_bn)
                                    shutil.unpack_archive(zip_file, extract_dir=job_dir)
                                    if not os.path.isdir(old_path):
                                        raise NotADirectoryError('Unpacking failed: {}'.format(zip_file))

                            if self.getVariation() is None:
                                new_path = os.path.join(folder, base_name)
                            else:
                                var_dir = os.path.join(folder, "{}_{}".format(row, col))
                                if not os.path.exists(var_dir):
                                    os.mkdir(var_dir)
                                new_path = os.path.join(var_dir, base_name)

                            '''check if the variable is a file or folder'''
                            if os.path.isfile(old_path):
                                '''copy file'''
                                shutil.copyfile(old_path, new_path)
                            elif os.path.isdir(old_path):
                                '''copy directory tree'''
                                shutil.copytree(old_path, new_path)
                            else:
                                raise Exception('Error due to unexpected case!')

                            val = new_path

                    if name in d.keys():
                        '''insert in dictionary'''
                        if self.getVariation() is None:
                            d[name] = val
                        else:
                            d[name][row, col] = val

                '''remove job from jobmanager'''
                SymJobDesc.remove(jobID)
