# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10184 $
$LastChangedDate:: 2021-05-05 14:10:10 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import io
import compileall
import symspace.osutils.processutils
from symspace.pyutils.zippkg import ZipPackage
from symspace.pyutils.version_util import VersionUtil


class EggPackage(ZipPackage):
    file_extension = 'egg'

    @classmethod
    def createFromPkgDir(cls, pkg_dir: str, target_dir: str = None, pkg_version: str = None,
                         compile_bytecode: bool = True, dry_run: bool = False) -> str:
        """
        Create a distribution from a package directory

        Parameters
        ----------
        pkg_dir: str
            Package directory
        target_dir: str
            Target directory for the created distribution file
        pkg_version: str
            Package version
        compile_bytecode: bool
            Compile py to pyc
        dry_run: bool
            Don't create a file (but return the file path, where it would be if it was created)

        Examples
        --------
        >>> egg_file = EggPackage.createFromPkgDir('x:\\my_pkg')

        Returns
        -------
        file: str
            Egg-Distribution file
        """
        try:
            '''imports could result in errors'''
            import setuptools
        except Exception:
            raise ImportError('Unable to import setuptools!')

        version_tag = 'py{}'.format(VersionUtil.getPythonVersion())
        pkg_name = os.path.basename(pkg_dir)
        parent_folder = os.path.dirname(pkg_dir)

        if pkg_version is None:
            pkg_version = '0.0.0'
        if target_dir is None:
            target_dir = os.path.join(parent_folder, 'dist')

        target_path = os.path.join(target_dir, '{}-{}-{}'.format(pkg_name, pkg_version, version_tag))
        target_file = '{}.{}'.format(target_path, cls.file_extension)

        if dry_run:
            file = target_file
        else:
            if compile_bytecode:
                '''compile py to pyc
                Python will not attempt to modify the .zip archive by adding the corresponding .pyc file,
                meaning that if a ZIP archive doesn’t contain .pyc files, importing may be rather slow.
                '''
                compileall.compile_dir(pkg_dir, quiet=True)

            script_args = ['bdist_egg', '--quiet',
                           '--dist-dir', target_dir,
                           # '--exclude-source-files'
                           ]

            wdir = os.getcwd()
            os.chdir(parent_folder)

            if symspace.osutils.processutils.getInterpreter() != 'python':
                '''workaround: redirect stream to avoid errors, if not run in python-interpreter'''
                sys.stdout = io.StringIO()
                sys.stderr = io.StringIO()

            dist = setuptools.setup(script_args=script_args,
                                    name=pkg_name,
                                    version=pkg_version,
                                    packages=setuptools.find_packages(where=parent_folder,
                                                                      include=[pkg_name, '{}.*'.format(pkg_name)]),
                                    )

            if symspace.osutils.processutils.getInterpreter() != 'python':
                '''reset stream to default behaviour'''
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__

            os.chdir(wdir)
            file = dist.dist_files[0][2]
            if target_file != file:
                raise Exception('Unexpected error: {} != {}'.format(target_file, file))
            if not os.path.isfile(file):
                raise FileNotFoundError('File does not exist: {}'.format(file))
            print('File created:', file)

        return file
