#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 9639 $
$LastChangedDate:: 2020-02-12 13:38:34 +0100#$

This file is part of SyMSpace.

Copyright (c) 2019, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

try:
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
    from at.lcm.bu21.magopt.structure import CollectorStructure
except Exception as e:
    print(e)


class CollectStruct():
    """
    Base class for collect container and material container
    """
    @staticmethod
    def _convertObject(obj, name=''):
        if isinstance(obj, dict):
            if 'unit__' in obj:
                res = StructParam(obj, name)
            else:
                res = StructContainer(obj, name)
        else:
            res = None

        return res

    def __init__(self, par, name=''):
        """
        Initialize Collect Struct

        Parameters
        ----------
        par : dict
        name : str

        Examples
        --------

        **Magnet material**

        .. code-block:: python

            mat_magnet = StructContainer(MaterialMagnet)
            Br = mat_magnet['Br'].getValue('T')
            mur = mat_magnet['mur'].getValue()

        """
        if not isinstance(par, dict):
            raise ValueError('parameter has to be of type dictionary')

        self.param = par
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __gt__(self, other):
        ret = self.name.__gt__(other.name)
        # print(ret)
        return ret

    def getName(self):
        return self.name

    def toDict(self):
        return self.param


class StructParam(CollectStruct):
    """
    Parameter is definde as dict with entries 'value__', 'unit__'
    """
    def __init__(self, *args):
        """
        Init method

        Parameters
        ----------
        args : dict, str
            param, name
        args : str, any, str, str
            name, value, unit, description

        Examples
        --------

        **Create new**

        .. code-block:: python

            volume = StructParam('volume', 5000, 'mm3', 'Volume of part')
        """

        cnt = len(args)

        if isinstance(args[0], dict):
            super().__init__(args[0], args[1])
        else:
            name = args[0]
            value = args[1]
            unit = ''
            if cnt >= 3:
                unit = args[2]
            desc = ''
            if cnt == 4:
                desc = args[3]
            self.name = name
            self.param = {'value__': value, 'unit__': unit, 'description__': desc}

    def __repr__(self):
        str = '''{}
    value: {}
    unit: {}
    description: {}'''.format(self.name, self.getData(), self.getUnit(), self.getDescription())
        return str

    def __str__(self):
        return self.__repr__()

    def getData(self):
        return self.param['value__']

    def getUnit(self):
        return self.param['unit__']

    def getDescription(self):
        if 'description__' in self.param:
            ret = self.param['description__']
        else:
            ret = ''
        return ret

    def getValue(self, unit=None):

        if unit is None or unit == self.getUnit():
            data = self.getData()
        else:
            data = PyJDataInterface.convert_Py2J(self.param)
            data = CollectorStructure(data).getValue(None, unit)
            data = PyJDataInterface.convert_J2Py(data)

        return data

    def setData(self, value):
        self.param['value__'] = value

    def setUnit(self, unit : str):
        self.param['unit__'] = unit

    def changeUnit(self, unit : str):
        data = self.getValue(unit)
        self.setData(data)
        self.setUnit(unit)

    def setDescription(self, desc):
        self.param['description__'] = desc

    def setValue(self, value, unit=None):
        if unit is None or unit == self.getUnit():
            if isinstance(value, StructParam):
                self.param = value.param
            else:
                self.setData(value)
        else:
            data = CollectorStructure()
            data.setUnit(unit)
            value = PyJDataInterface.convert_Py2J(value)
            data.setValue(None, value)
            value = data.getValue(None, self.getUnit())
            value = PyJDataInterface.convert_J2Py(value)
            self.setData(value)



class StructContainer(CollectStruct):
    """
    Struct container is a dict with StructContainer and StructParam entries
    """
    def __init__(self, *args):
        """
        Init method

        Parameters
        ----------
        args : dict, str

        Examples
        --------

        **Magnet material**

        .. code-block:: python

            mat_magnet = StructContainer(MaterialMagnet)
            Br = mat_magnet['Br'].getValue('T')
            mur = mat_magnet['mur'].getValue()


        **Create new**

        .. code-block:: python

            mass = StructContainer('Mass')

        """
        if isinstance(args[0], dict):
            if len(args) == 1:
                super().__init__(args[0], '')
            else:
                super().__init__(args[0], args[1])
        else:
            self.name = args[0]
            self.param = dict()

    def __getitem__(self, key):
        """
        Indexing of parameter
        """
        if not isinstance(key, str):
            raise ValueError('key has to be of type string')
        return self.getObject(key)

    def getParam(self):
        """
        Get List of child-parameters.

        Returns
        -------
        ret : list
            List of child parameters

        """
        res = []
        for par in self.param:
            obj = CollectStruct._convertObject(self.param[par], par)
            if obj is not None:
                res.append(obj)
        return res

    def getObject(self, identString):
        obj = self.param
        for ident in identString.split('.'):
            if ident in obj:
                obj = obj[ident]
            else:
                obj = None
                break
        if obj is not None:
            ret = CollectStruct._convertObject(obj, ident)
        else:
            ret = None
        return ret

    def addParam(self, par):
        if isinstance(par, str):
            self.param[par] = dict()
        else:
            self.param[par.name] = par.param

    def removeParam(self, name):
        if name in self.param:
            self.param.pop(name)

    def hasParam(self, identString):
        return self.getObject(identString) is not None

    def setName(self, name):
        self.name = name
